@@ -21,7 +21,6 @@
 
 #define _BSD_SOURCE
 
-#include "jsocket.h"
 
 #include <stdio.h>
 #include <sys/types.h>
@@ -32,46 +31,45 @@
 #include <strings.h>
 #include <string.h>
 #include <unistd.h>
-#include "jassert.h"
 #include <errno.h>
 #include <algorithm>
 #include <set>
+#include <typeinfo>
 
-#ifndef DMTCP
-#  define DECORATE_FN(fn) ::fn
-#else
-#  include "syscallwrappers.h"
-#  define DECORATE_FN(fn) ::_real_ ## fn
-#endif
-
-#ifdef DMTCP
-#  ifdef RECORD_REPLAY
-#    define select _real_select
-#  endif
-#endif
+#include "jsocket.h"
+#include "jalib.h"
+#include "jassert.h"
 
 const jalib::JSockAddr jalib::JSockAddr::ANY ( NULL );
 
 jalib::JSockAddr::JSockAddr ( const char* hostname /* == NULL*/,
                               int port /* == -1*/ )
 {
-  memset ( &_addr, 0, sizeof ( _addr ) );
-  _addr.sin_family = AF_INET;
+  // Initialization for any case
+  memset( (void*)&_addr, 0, sizeof( _addr ) );
+  for(unsigned int i=0; i < (max_count + 1); i++){
+    _addr[i].sin_family = AF_INET;
+  }
+  _count = 0;
+
   if ( hostname == NULL ) {
-    _addr.sin_addr.s_addr = INADDR_ANY;
-    if (port != -1) _addr.sin_port = htons (port);
+    _count = 1;
+    _addr[0].sin_addr.s_addr = INADDR_ANY;
+    if (port != -1)
+      _addr[0].sin_port = htons (port);
     return;
   }
-#if 1
+
+#if 0
   struct hostent ret, *result;
   char buf[1024];
   int h_errnop;
-  
+
   int res = gethostbyname_r(hostname, &ret, buf, sizeof buf, &result, &h_errnop);
 
   // Fall back to gethostbyname on error
   if (res != 0) {
-    JWARNING (false) (hostname) (hstrerror)
+    JWARNING (false) (hostname) (hstrerror(h_errno))
       .Text("gethostbyname_r failed, calling gethostbyname");
     result = gethostbyname ( hostname );
   }
@@ -84,6 +82,9 @@ jalib::JSockAddr::JSockAddr ( const char* hostname /* == NULL*/,
 
     memcpy ( &_addr.sin_addr.s_addr, result->h_addr, result->h_length );
     if (port != -1) _addr.sin_port = htons (port);
+  } else { // else (hostname, port) not valid; poison the port number
+    JASSERT( typeid(_addr.sin_port) == typeid(in_port_t) );
+    _addr.sin_port = (in_port_t)-2;
   }
 #else
   struct addrinfo hints;
@@ -96,17 +97,47 @@ jalib::JSockAddr::JSockAddr ( const char* hostname /* == NULL*/,
   /* FIXME: Ulrich Drepper states the following about getaddrinfo():
    *   The most important thing when using getaddrinfo is to make sure that all
    *   results are used in order. To stress the important words again: all and
-   *   order. Too many (incorrect) programs only use the first result. 
+   *   order. Too many (incorrect) programs only use the first result.
    * Source: http://www.akkadia.org/drepper/userapi-ipv6.html
    *
    * This would require some sort of redesign of JSockAddr and JSocket classes.
    */
   int e = getaddrinfo(hostname, NULL, &hints, &res);
+  if( e == EAI_NONAME ){
+    /*
+     * If the AI_ADDRCONFIG flag is passed to getaddrinfo() with AF_INET or
+     * AF_INET6 address families,
+     * addresses in /etc/hosts are not resolved properly.
+     * According to following bug reports:
+     * https://bugs.launchpad.net/ubuntu/+source/glibc/+bug/583278
+     * https://bugzilla.mozilla.org/show_bug.cgi?id=467497#c10
+     *
+     * As a temporary workaround, just call it without AI_ADDRCONFIG again
+     */
+    hints.ai_flags = 0;
+    e = getaddrinfo( hostname, NULL, &hints, &res);
+  }
+
   JWARNING (e == 0) (e) (gai_strerror(e)) (hostname) .Text ("No such host");
   if (e == 0) {
-    JASSERT(sizeof _addr >= res->ai_addrlen) (sizeof _addr) (res->ai_addrlen);
-    memcpy(&_addr, res->ai_addr, res->ai_addrlen);
-    if (port != -1) _addr.sin_port = htons (port);
+    JASSERT(sizeof(*_addr) >= res->ai_addrlen)(sizeof(*_addr))(res->ai_addrlen);
+
+    // 1. count number of addresses returned
+    struct addrinfo *r;
+    for(r = res, _count = 0; r != NULL; r = r->ai_next, _count++);
+    if( _count > max_count)
+      _count = max_count;
+
+    // 2. array for storing all nesessary addresses
+    int i;
+    for(r = res, i = 0; r != NULL; r = r->ai_next, i++) {
+      memcpy(_addr + i, r->ai_addr, r->ai_addrlen);
+      if (port != -1)
+        _addr[i].sin_port = htons (port);
+    }
+
+  } else { // else (hostname, port) not valid; poison the port number
+    _addr[0].sin_port = (unsigned short)-2;
   }
 
   freeaddrinfo(res);
@@ -115,50 +146,70 @@ jalib::JSockAddr::JSockAddr ( const char* hostname /* == NULL*/,
 
 jalib::JSocket::JSocket()
 {
-  _sockfd = DECORATE_FN ( socket ) ( AF_INET, SOCK_STREAM, 0 );
+  _sockfd = jalib::socket ( AF_INET, SOCK_STREAM, 0 );
 }
 
 
 bool jalib::JSocket::connect ( const JSockAddr& addr, int port )
 {
-  return JSocket::connect ( ( sockaddr* ) &addr._addr, sizeof ( addr._addr ), port );
+  bool ret = false;
+  // jalib::JSockAddr::JSockAddr used -2 to poison port (invalid host)
+  if (addr._addr->sin_port == (unsigned short)-2)
+    return false;
+  for(unsigned int i=0; i< addr._count; i++){
+    if( ret = JSocket::connect( (sockaddr*)(addr._addr + i),
+                                sizeof(addr._addr[0]), port ) ){
+      break;
+    }else{
+      if( errno != ECONNREFUSED)
+        break;
+    }
+  }
+  return ret;
 }
 
 
-bool jalib::JSocket::connect ( const  struct  sockaddr  *addr,  socklen_t addrlen, int port )
+bool jalib::JSocket::connect ( const  struct  sockaddr  *addr,
+                               socklen_t addrlen, int port )
 {
   struct sockaddr_storage addrbuf;
   memset ( &addrbuf,0,sizeof ( addrbuf ) );
   JASSERT ( addrlen <= sizeof ( addrbuf ) ) ( addrlen ) ( sizeof ( addrbuf ) );
   memcpy ( &addrbuf,addr,addrlen );
-  JWARNING ( addrlen == sizeof ( sockaddr_in ) ) ( addrlen ) ( sizeof ( sockaddr_in ) ).Text ( "may not be correct socket type" );
-  ( ( sockaddr_in* ) &addrbuf )->sin_port = htons ( port );
-  return DECORATE_FN ( connect ) ( _sockfd, ( sockaddr* ) &addrbuf, addrlen ) == 0;
+  JWARNING ( addrlen == sizeof ( sockaddr_in ) ) ( addrlen )
+          ( sizeof ( sockaddr_in ) ).Text ( "may not be correct socket type" );
+  ( (sockaddr_in*)&addrbuf )->sin_port = htons ( port );
+  return jalib::connect( _sockfd, (sockaddr*)&addrbuf, addrlen ) == 0;
 }
 
 bool jalib::JSocket::bind ( const JSockAddr& addr, int port )
 {
-  struct sockaddr_in addrbuf = addr._addr;
-  addrbuf.sin_port = htons ( port );
-  return bind ( ( sockaddr* ) &addrbuf, sizeof ( addrbuf ) );
+  bool ret = false;
+  for( unsigned int i=0; i<addr._count; i++){
+    struct sockaddr_in addrbuf = addr._addr[i];
+    addrbuf.sin_port = htons ( port );
+    int retval = bind( (sockaddr*)&addrbuf, sizeof(addrbuf) );
+    ret = ret || retval;
+  }
+  return ret;
 }
 
 bool jalib::JSocket::bind ( const  struct  sockaddr  *addr,  socklen_t addrlen )
 {
-  return DECORATE_FN ( bind ) ( _sockfd, addr, addrlen ) == 0;
+  return jalib::bind ( _sockfd, addr, addrlen ) == 0;
 }
 
 bool jalib::JSocket::listen ( int backlog/* = 32*/ )
 {
-  return DECORATE_FN ( listen ) ( _sockfd, backlog ) == 0;
+  return jalib::listen  ( _sockfd, backlog ) == 0;
 }
 
 jalib::JSocket jalib::JSocket::accept ( struct sockaddr_storage* remoteAddr,socklen_t* remoteLen )
 {
   if ( remoteAddr == NULL || remoteLen == NULL )
-    return JSocket ( DECORATE_FN ( accept ) ( _sockfd,NULL,NULL ) );
+    return JSocket ( jalib::accept ( _sockfd,NULL,NULL ) );
   else
-    return JSocket ( DECORATE_FN ( accept ) ( _sockfd, ( sockaddr* ) remoteAddr, remoteLen ) );
+    return JSocket ( jalib::accept ( _sockfd, ( sockaddr* ) remoteAddr, remoteLen ) );
 }
 
 void jalib::JSocket::enablePortReuse()
@@ -180,14 +231,14 @@ void jalib::JSocket::enablePortReuse()
 bool jalib::JSocket::close()
 {
   if ( !isValid() ) return false;
-  int ret = DECORATE_FN(close) ( _sockfd );
+  int ret = jalib::close ( _sockfd );
   _sockfd = -1;
   return ret==0;
 }
 
 ssize_t jalib::JSocket::read ( char* buf, size_t len )
 {
-  return ::read ( _sockfd,buf,len );
+  return jalib::read ( _sockfd,buf,len );
 }
 
 ssize_t jalib::JSocket::write ( const char* buf, size_t len )
@@ -216,17 +267,17 @@ ssize_t jalib::JSocket::readAll ( char* buf, size_t len )
     tv.tv_sec = 120;
     tv.tv_usec = 0;
 
-    retval = select ( tmp_sockfd+1, &rfds, NULL, NULL, &tv );
+    retval = jalib::select ( tmp_sockfd+1, &rfds, NULL, NULL, &tv );
     /* Don't rely on the value of tv now! */
 
-
     if ( retval == -1 )
     {
       if ( errno == EBADF ) {
         JWARNING (false) .Text ( "Socket already closed" );
         return -1;
-      } else if( errno != EINTR ){ 
-        JWARNING ( retval >= 0 ) ( tmp_sockfd ) ( JASSERT_ERRNO ).Text ( "select() failed" );
+      } else if( errno != EINTR ){
+        JWARNING ( retval >= 0 )
+          ( tmp_sockfd ) ( JASSERT_ERRNO ).Text ( "select() failed" );
         return -1;
       }
     }
@@ -261,56 +312,53 @@ ssize_t jalib::JSocket::readAll ( char* buf, size_t len )
 ssize_t jalib::JSocket::writeAll ( const char* buf, size_t len )
 {
   int origLen = len;
-  while ( len > 0 )
-  {
+  while (len > 0) {
     fd_set wfds;
     struct timeval tv;
     int retval;
 
     int tmp_sockfd = _sockfd;
-    if ( tmp_sockfd == -1 ) {
+    if (tmp_sockfd == -1) {
       return -1;
     }
 
     /* Watch stdin (fd 0) to see when it has input. */
-    FD_ZERO ( &wfds );
-    FD_SET ( tmp_sockfd, &wfds );
+    FD_ZERO(&wfds);
+    FD_SET(tmp_sockfd, &wfds);
 
     /* Wait up to five seconds. */
     tv.tv_sec = 30;
     tv.tv_usec = 0;
 
-    retval = select ( tmp_sockfd+1, NULL, &wfds, NULL, &tv );
+    retval = jalib::select(tmp_sockfd+1, NULL, &wfds, NULL, &tv);
     /* Don't rely on the value of tv now! */
 
 
-    if ( retval == -1 )
-    {
-      if ( errno == EBADF ) {
-        JWARNING (false) .Text ( "Socket already closed" );
-        return -1;
-      } 
-      JWARNING ( retval >= 0 ) ( tmp_sockfd ) ( JASSERT_ERRNO ).Text ( "select() failed" );
-      return -1;
-    }
-    else if ( retval )
-    {
-      errno = 0;
-      ssize_t cnt = write ( buf, len );
-      if ( cnt <= 0 && errno != EAGAIN && errno != EINTR )
-      {
-        JWARNING ( cnt > 0 ) ( cnt ) ( len ) ( JASSERT_ERRNO ).Text ( "JSocket read failure" );
+    if (retval == -1) {
+      if (errno == EBADF || errno == EPIPE) {
+        JTRACE("Socket already closed");
         return -1;
       }
-      if ( cnt > 0 )
-      {
+      JWARNING(retval >= 0) (tmp_sockfd) (JASSERT_ERRNO)
+        .Text ("select() failed");
+      return -1;
+    } else if (retval) {
+      errno = 0;
+      ssize_t cnt = write(buf, len);
+      if (cnt < 0 && (errno == EBADF || errno == EPIPE)) {
+        JTRACE("Remote peer disconnected, write failed");
+        return -1;
+      } else if (cnt <= 0 && errno != EAGAIN && errno != EINTR) {
+        JWARNING(cnt > 0) (cnt) (len) (JASSERT_ERRNO)
+          .Text("JSocket write failure");
+        return -1;
+      }
+      if (cnt > 0) {
         buf += cnt;
         len -= cnt;
       }
-    }
-    else
-    {
-      JTRACE ( "still waiting for data" ) ( tmp_sockfd ) ( len );
+    } else {
+      JTRACE("still waiting for data") (tmp_sockfd) (len);
     }
   }
   return origLen;
@@ -399,7 +447,7 @@ void jalib::JChunkReader::readAll()
  */
 jalib::JChunkReader::JChunkReader ( JSocket sock, int chunkSize )
     : JReaderInterface ( sock )
-    , _buffer ( new char[chunkSize] )
+    , _buffer ( (char*) jalib::JAllocDispatcher::malloc(chunkSize) )
     , _length ( chunkSize )
     , _read ( 0 )
     , _hadError ( false )
@@ -413,7 +461,7 @@ jalib::JChunkReader::JChunkReader ( JSocket sock, int chunkSize )
  */
 jalib::JChunkReader::JChunkReader ( const JChunkReader& that )
     : JReaderInterface ( that )
-    , _buffer ( new char[that._length] )
+    , _buffer ( (char*) jalib::JAllocDispatcher::malloc(that._length) )
     , _length ( that._length )
     , _read ( that._read )
     , _hadError ( that._hadError )
@@ -427,7 +475,7 @@ jalib::JChunkReader::JChunkReader ( const JChunkReader& that )
  */
 jalib::JChunkReader::~JChunkReader()
 {
-  delete [] _buffer;
+  jalib::JAllocDispatcher::free(_buffer);
   _buffer = 0;
 }
 
@@ -438,7 +486,7 @@ jalib::JChunkReader::~JChunkReader()
 jalib::JChunkReader& jalib::JChunkReader::operator= ( const JChunkReader& that )
 {
   if ( this == &that ) return *this;;
-  delete [] _buffer;
+  jalib::JAllocDispatcher::free(_buffer);
   _buffer = 0;
   new ( this ) JChunkReader ( that );
   return *this;
@@ -453,7 +501,7 @@ jalib::JChunkReader& jalib::JChunkReader::operator= ( const JChunkReader& that )
 void jalib::JSocket::changeFd ( int newFd )
 {
   if ( _sockfd == newFd ) return;
-  JASSERT ( newFd == DECORATE_FN(dup2) ( _sockfd, newFd ) )
+  JASSERT ( newFd == dup2 ( _sockfd, newFd ) )
       ( _sockfd ) ( newFd ).Text ( "dup2 failed" );
   close();
   _sockfd = newFd;
@@ -508,7 +556,7 @@ void jalib::JMultiSocketProgram::monitorSockets ( double dblTimeout )
 
   setTimeoutInterval ( dblTimeout );
 
-  timeoutBuf=timeoutInterval;
+  timeoutBuf = timeoutInterval;
   timeout = timeoutEnabled ? &timeoutBuf : NULL;
 
   IntSet closedFds;
@@ -523,6 +571,13 @@ void jalib::JMultiSocketProgram::monitorSockets ( double dblTimeout )
     FD_ZERO ( &rfds );
     FD_ZERO ( &wfds );
 
+    if( timeout == NULL && timeoutEnabled){
+      timeoutBuf=timeoutInterval;
+      timeout = &timeoutBuf;
+    }else if( timeout != NULL && !timeoutEnabled){
+      timeout = NULL;
+    }
+
     //collect listen fds in rfds, cleanup dead sockets
     for ( i=0; i<_listenSockets.size(); ++i )
     {
@@ -565,6 +620,7 @@ void jalib::JMultiSocketProgram::monitorSockets ( double dblTimeout )
         _dataSockets[i] = _dataSockets[_dataSockets.size()-1];
         _dataSockets.pop_back();
         i--;
+        processPostDisconnect();
       }
     }
 
@@ -596,17 +652,24 @@ void jalib::JMultiSocketProgram::monitorSockets ( double dblTimeout )
       return;
     }
 
+    //NOTE:  The top level routine of dmtcp_coordinator calls monitorSockets(),
+    //  which then waits on activity from clients connecting to the coordinator.
+    //  Depending on the result of this selection, this rountine,
+    //  monitorSockets() will in turn call onData(), onConnect(),
+    //  onTimeoutInterval(), etc.  Those routines do their work, and on
+    //  completion, they then return to monitorSockets() to wait for more
+    //  work to do.  dmtcp_coordinator.cpp also describes some of this logic.
     //this will block till we have some work to do
-    int retval = select ( maxFd+1, &rfds, &wfds, NULL, timeout );
+    int retval = jalib::select ( maxFd+1, &rfds, &wfds, NULL, timeout );
 
     if ( retval == -1 )
     {
-      JWARNING ( retval != -1 ) ( maxFd ) ( retval ) ( JASSERT_ERRNO ).Text ( "select failed" );
+      JWARNING ( retval != -1 )
+        ( maxFd ) ( retval ) ( JASSERT_ERRNO ).Text ( "select failed" );
       return;
     }
     else if ( retval > 0 )
     {
-
       //write all data
       for ( i=0; i<_writes.size(); ++i )
       {
@@ -642,9 +705,10 @@ void jalib::JMultiSocketProgram::monitorSockets ( double dblTimeout )
         if ( fd >= 0 && FD_ISSET ( fd, &rfds ) )
         {
           struct sockaddr_storage addr;
-          socklen_t               addrlen=sizeof ( addr );
+          socklen_t addrlen = sizeof ( addr );
           JSocket sk = _listenSockets[i].accept ( &addr,&addrlen );
-          JTRACE ( "accepting new connection" ) ( i ) ( sk.sockfd() ) ( _listenSockets[i].sockfd() ) ( JASSERT_ERRNO );
+          JTRACE ( "accepting new connection" ) ( i ) ( sk.sockfd() )
+            ( _listenSockets[i].sockfd() ) ( JASSERT_ERRNO );
           if ( sk.isValid() )
           {
             onConnect ( sk, ( sockaddr* ) &addr,addrlen );
@@ -655,9 +719,6 @@ void jalib::JMultiSocketProgram::monitorSockets ( double dblTimeout )
           }
         }
       }
-
-
-
     }
 
     if ( timeoutEnabled )
@@ -688,7 +749,7 @@ void jalib::JMultiSocketProgram::monitorSockets ( double dblTimeout )
  */
 jalib::JChunkWriter::JChunkWriter ( JSocket sock, const char* buf, int len )
     :JWriterInterface ( sock )
-    ,_buffer ( new char [len] )
+    , _buffer ( (char*) jalib::JAllocDispatcher::malloc(len) )
     ,_length ( len )
     ,_sent ( 0 )
     ,_hadError ( false )
@@ -702,7 +763,7 @@ jalib::JChunkWriter::JChunkWriter ( JSocket sock, const char* buf, int len )
  */
 jalib::JChunkWriter::JChunkWriter ( const JChunkWriter& that )
     :JWriterInterface ( that )
-    ,_buffer ( new char [that._length] )
+    , _buffer ( (char*) jalib::JAllocDispatcher::malloc(that._length) )
     ,_length ( that._length )
     ,_sent ( that._sent )
     ,_hadError ( false )
@@ -716,7 +777,7 @@ jalib::JChunkWriter::JChunkWriter ( const JChunkWriter& that )
  */
 jalib::JChunkWriter::~JChunkWriter()
 {
-  delete [] _buffer;
+  jalib::JAllocDispatcher::free(_buffer);
   _buffer = 0;
 }
 
@@ -727,7 +788,7 @@ jalib::JChunkWriter::~JChunkWriter()
 jalib::JChunkWriter& jalib::JChunkWriter::operator= ( const JChunkWriter& that )
 {
   if ( this == &that ) return *this;;
-  delete [] _buffer;
+  jalib::JAllocDispatcher::free(_buffer);
   _buffer = 0;
   new ( this ) JChunkWriter ( that );
   return *this;
