@@ -19,6 +19,8 @@
  *  <http://www.gnu.org/licenses/>.                                         *
  ****************************************************************************/
 
+#include "constants.h"
+#include "syscallwrappers.h"
 #include "connection.h"
 #include  "../jalib/jassert.h"
 #include  "../jalib/jfilesystem.h"
@@ -28,7 +30,6 @@
 #include "connectionrewirer.h"
 #include "connectionmanager.h"
 #include "dmtcpmessagetypes.h"
-#include "dmtcpworker.h"
 #include "virtualpidtable.h"
 #include "util.h"
 #include  "../jalib/jsocket.h"
@@ -50,9 +51,9 @@ static ssize_t ptmxReadAll(int fd, const void *origBuf, size_t maxCount);
 static ssize_t ptmxWriteAll(int fd, const void *buf, bool isPacketMode);
 
 #ifdef REALLY_VERBOSE_CONNECTION_CPP
-bool really_verbose = true;
+static bool really_verbose = true;
 #else
-bool really_verbose = false;
+static bool really_verbose = false;
 #endif
 
 static dmtcp::string _procFDPath ( int fd )
@@ -80,8 +81,9 @@ static int _makeDeadSocket()
   JASSERT ( sp[0]>=0 && sp[1]>=0 ) ( sp[0] ) ( sp[1] )
     .Text ( "socketpair() failed" );
   _real_close ( sp[1] );
-  if (really_verbose)
+  if (really_verbose) {
     JTRACE ( "Created dead socket." ) ( sp[0] );
+  }
   return sp[0];
 }
 
@@ -92,7 +94,7 @@ static bool _isVimApp ( )
   if (isVimApp == -1) {
     dmtcp::string progName = jalib::Filesystem::GetProgramName();
 
-    if (progName == "vi" || progName == "vim" || progName == "vim-normal" || 
+    if (progName == "vi" || progName == "vim" || progName == "vim-normal" ||
         progName == "vim.basic"  || progName == "vim.tiny" ||
         progName == "vim.gtk" || progName == "vim.gnome" ) {
       isVimApp = 1;
@@ -105,14 +107,42 @@ static bool _isVimApp ( )
 
 static bool _isBlacklistedFile ( dmtcp::string& path )
 {
-  if ((Util::strStartsWith(path, "/dev/") &&
-       !Util::strStartsWith(path, "/dev/shm/")) ||
-      Util::strStartsWith(path, dmtcp::UniquePid::getTmpDir().c_str())) {
+  if ((dmtcp::Util::strStartsWith(path, "/dev/") &&
+       !dmtcp::Util::strStartsWith(path, "/dev/shm/")) ||
+      dmtcp::Util::strStartsWith(path, dmtcp::UniquePid::getTmpDir().c_str())) {
     return true;
   }
   return false;
 }
 
+static bool _isBlacklistedTcp ( int sockfd,
+				const sockaddr* saddr, socklen_t len )
+{
+  JASSERT( saddr != NULL );
+  if ( len >= sizeof(saddr->sa_family) && saddr->sa_family == AF_INET ) {
+    struct sockaddr_in* addr = (sockaddr_in*)saddr;
+    // Ports 389 and 636 are the well-known ports in /etc/services that
+    // are reserved for LDAP.  Bash continues to maintain a connection to
+    // LDAP, leading to problems at restart time.  So, we discover the LDAP
+    // remote addresses, and turn them into dead sockets at restart time.
+    //However, libc.so:getpwuid() can call libnss_ldap.so which calls
+    // libldap-2.4.so to create the LDAP socket while evading our connect
+    // wrapper.
+    int blacklistedRemotePorts[] = {389, 636, -1}; /* LDAP ports */
+    int i;
+    for (i = 0; blacklistedRemotePorts[i] != -1; i++) {
+      if (ntohs(addr->sin_port) == blacklistedRemotePorts[i]) {
+        JTRACE("LDAP port found")(ntohs(addr->sin_port))
+              (blacklistedRemotePorts[0])(blacklistedRemotePorts[1]);
+        return true;
+      }
+    }
+  }
+  // FIXME:  Consider adding configure or dmtcp_checkpoint option to disable
+  //   all remote connections.  Handy as quick test for special cases.
+  return false;
+}
+
 dmtcp::Connection::Connection ( int t )
   : _id ( ConnectionIdentifier::Create() )
   , _type ( ( ConnectionType ) t )
@@ -161,7 +191,7 @@ void dmtcp::Connection::restoreOptions ( const dmtcp::vector<int>& fds )
   //  The correct fix would be to restore the fcntlowner after we have the
   // original->current pid mappings.
   errno = 0;
-  JASSERT ( fcntl ( fds[0], F_SETOWN, ORIGINAL_TO_CURRENT_PID(_fcntlOwner) ) == 0 ) 
+  JASSERT ( fcntl ( fds[0], F_SETOWN, ORIGINAL_TO_CURRENT_PID(_fcntlOwner) ) == 0 )
     ( fds[0] ) ( _fcntlOwner ) ( JASSERT_ERRNO );
 
   // This JASSERT will almost always trigger until we fix the above mentioned
@@ -175,7 +205,7 @@ void dmtcp::Connection::restoreOptions ( const dmtcp::vector<int>& fds )
 void dmtcp::Connection::doLocking ( const dmtcp::vector<int>& fds )
 {
   errno = 0;
-  JASSERT ( fcntl ( fds[0], F_SETOWN, _real_getpid() ) == 0 ) 
+  JASSERT ( fcntl ( fds[0], F_SETOWN, _real_getpid() ) == 0 )
     ( fds[0] ) ( JASSERT_ERRNO );
 }
 
@@ -206,11 +236,11 @@ dmtcp::TcpConnection& dmtcp::TcpConnection::asTcp()
   return *this;
 }
 
-void dmtcp::TcpConnection::onBind ( const struct sockaddr* addr, socklen_t len )
+void dmtcp::TcpConnection::onBind (const struct sockaddr* addr, socklen_t len)
 {
-  if (really_verbose)
+  if (really_verbose) {
     JTRACE ("Binding.") ( id() ) ( len );
-
+  }
   JASSERT ( tcpType() == TCP_CREATED ) ( tcpType() ) ( id() )
     .Text ( "Binding a socket in use????" );
   JASSERT ( len <= sizeof _bindAddr ) ( len ) ( sizeof _bindAddr )
@@ -222,9 +252,9 @@ void dmtcp::TcpConnection::onBind ( const struct sockaddr* addr, socklen_t len )
 }
 void dmtcp::TcpConnection::onListen ( int backlog )
 {
-  if (really_verbose)
+  if (really_verbose) {
     JTRACE ( "Listening." ) ( id() ) ( backlog );
-
+  }
   JASSERT ( tcpType() == TCP_BIND ) ( tcpType() ) ( id() )
     .Text ( "Listening on a non-bind()ed socket????" );
   // A -1 backlog is not an error.
@@ -234,17 +264,24 @@ void dmtcp::TcpConnection::onListen ( int backlog )
   _type = TCP_LISTEN;
   _listenBacklog = backlog;
 }
-void dmtcp::TcpConnection::onConnect( int sockfd, 
-                                      const  struct sockaddr *addr, 
+void dmtcp::TcpConnection::onConnect( int sockfd,
+                                      const struct sockaddr *addr,
                                       socklen_t len )
 {
-  if (really_verbose)
+  if (really_verbose) {
     JTRACE ( "Connecting." ) ( id() );
-
+  }
   JASSERT ( tcpType() == TCP_CREATED ) ( tcpType() ) ( id() )
     .Text ( "Connecting with an in-use socket????" );
 
-  _type = TCP_CONNECT;
+  /* socketpair wrapper calls onConnect with sockfd == -1 and addr == NULL */
+  if (addr != NULL && _isBlacklistedTcp(sockfd, addr, len)) {
+    _type = TCP_EXTERNAL_CONNECT;
+    _connectAddrlen = len;
+    memcpy ( &_connectAddr, addr, len );
+  } else {
+    _type = TCP_CONNECT;
+  }
 }
 
 /*onAccept*/
@@ -260,8 +297,9 @@ dmtcp::TcpConnection::TcpConnection ( const TcpConnection& parent, const Connect
   , _bindAddrlen ( 0 )
   , _acceptRemoteId ( remote )
 {
-  if (really_verbose)
+  if (really_verbose) {
     JTRACE ( "Accepting." ) ( id() ) ( parent.id() ) ( remote );
+  }
 
   //     JASSERT(parent.tcpType() == TCP_LISTEN)(parent.tcpType())(parent.id())
   //             .Text("Accepting from a non listening socket????");
@@ -307,8 +345,9 @@ void dmtcp::TcpConnection::preCheckpointPeerLookup ( const dmtcp::vector<int>& f
       }
       else
       {
-        if (really_verbose)
+        if (really_verbose) {
           JTRACE ( "Did not get lock.  Won't lookup." ) ( fds[0] ) ( id() );
+        }
       }
       break;
     case TCP_LISTEN:
@@ -326,8 +365,9 @@ void dmtcp::TcpConnection::preCheckpoint ( const dmtcp::vector<int>& fds
 
   if ( ( _fcntlFlags & O_ASYNC ) != 0 )
   {
-    if (really_verbose)
+    if (really_verbose) {
       JTRACE ( "Removing O_ASYNC flag during checkpoint." ) ( fds[0] ) ( id() );
+    }
     errno = 0;
     JASSERT ( fcntl ( fds[0],F_SETFL,_fcntlFlags & ~O_ASYNC ) == 0 ) ( JASSERT_ERRNO ) ( fds[0] ) ( id() );
   }
@@ -345,8 +385,9 @@ void dmtcp::TcpConnection::preCheckpoint ( const dmtcp::vector<int>& fds
       }
       else
       {
-        if (really_verbose)
+        if (really_verbose) {
           JTRACE ( "Did not get lock.  Won't drain" ) ( fds[0] ) ( id() );
+        }
       }
       break;
     case TCP_LISTEN:
@@ -377,9 +418,10 @@ void dmtcp::TcpConnection::doSendHandshakes( const dmtcp::vector<int>& fds, cons
         }
         else
         {
-          if (really_verbose)
+          if (really_verbose) {
             JTRACE("Skipping handshake send (shared socket, not owner).")
 		(id()) (fds[0]);
+          }
         }
         break;
       case TCP_EXTERNAL_CONNECT:
@@ -397,14 +439,16 @@ void dmtcp::TcpConnection::doSendHandshakes( const dmtcp::vector<int>& fds, cons
           JTRACE ("Receiving handshake ...") (id()) (fds[0]);
           jalib::JSocket sock(fds[0]);
           recvHandshake( sock, coordinator );
-          if (really_verbose)
+          if (really_verbose) {
             JTRACE ("Received handshake.") (getRemoteId()) (fds[0]);
+          }
         }
         else
         {
-          if (really_verbose)
+          if (really_verbose) {
             JTRACE ("Skipping handshake recv (shared socket, not owner).")
               (id()) (fds[0]);
+          }
         }
         break;
       case TCP_EXTERNAL_CONNECT:
@@ -417,8 +461,9 @@ void dmtcp::TcpConnection::postCheckpoint ( const dmtcp::vector<int>& fds, bool
 {
   if ( ( _fcntlFlags & O_ASYNC ) != 0 )
   {
-    if (really_verbose)
+    if (really_verbose) {
       JTRACE ( "Re-adding O_ASYNC flag." ) ( fds[0] ) ( id() );
+    }
     restoreOptions ( fds );
   }
 }
@@ -453,8 +498,9 @@ void dmtcp::TcpConnection::restore ( const dmtcp::vector<int>& fds, ConnectionRe
         ( _sockProtocol )
         .Text ( "Socket type not yet [fully] supported." );
 
-      if (really_verbose)
+      if (really_verbose) {
         JTRACE ( "Restoring socket." ) ( id() ) ( fds[0] );
+      }
 
       jalib::JSocket sock ( _real_socket ( _sockDomain,_sockType,_sockProtocol ) );
       JASSERT ( sock.isValid() );
@@ -514,16 +560,18 @@ void dmtcp::TcpConnection::restore ( const dmtcp::vector<int>& fds, ConnectionRe
         }
       }
 
-      if (really_verbose)
+      if (really_verbose) {
         JTRACE ( "Binding socket." ) ( id() );
+      }
       errno = 0;
       JWARNING ( sock.bind ( ( sockaddr* ) &_bindAddr,_bindAddrlen ) )
         ( JASSERT_ERRNO ) ( id() )
         .Text ( "Bind failed." );
       if ( tcpType() == TCP_BIND ) break;
 
-      if (really_verbose)
+      if (really_verbose) {
         JTRACE ( "Listening socket." ) ( id() );
+      }
       errno = 0;
       JWARNING ( sock.listen ( _listenBacklog ) )
         ( JASSERT_ERRNO ) ( id() ) ( _listenBacklog )
@@ -920,7 +968,7 @@ void dmtcp::FileConnection::preCheckpoint ( const dmtcp::vector<int>& fds
 
   calculateRelativePath();
 
-  _ckptFilesDir = UniquePid::checkpointFilesDirName();
+  _ckptFilesDir = UniquePid::checkpointFilesSubDirName();
 
   // Read the current file descriptor offset
   _offset = lseek(fds[0], 0, SEEK_CUR);
@@ -994,7 +1042,7 @@ void dmtcp::FileConnection::restoreOptions ( const dmtcp::vector<int>& fds )
 }
 
 
-void dmtcp::FileConnection::restore ( const dmtcp::vector<int>& fds, 
+void dmtcp::FileConnection::restore ( const dmtcp::vector<int>& fds,
                                       ConnectionRewirer& rewirer )
 {
   struct stat buf;
@@ -1006,11 +1054,11 @@ void dmtcp::FileConnection::restore ( const dmtcp::vector<int>& fds,
 
   if (_checkpointed) {
     JASSERT (jalib::Filesystem::FileExists(_path) == false) (_path)
-      .Text("\n**** File aready exists! Checkpointed copy can't be restored.\n"
+      .Text("\n**** File already exists! Checkpointed copy can't be restored.\n"
             "****Delete the existing file and try again!");
 
 //    JWARNING (jalib::Filesystem::FileExists(_path) == false) (_path)
-//      .Text("\n**** File aready exists! Deleting it ");
+//      .Text("\n**** File already exists! Deleting it ");
 //
 //    unlink(_path.c_str());
 
@@ -1019,7 +1067,7 @@ void dmtcp::FileConnection::restore ( const dmtcp::vector<int>& fds,
   } else if (jalib::Filesystem::FileExists(_path)) {
 
     if (stat(_path.c_str() ,&buf) == 0 && S_ISREG(buf.st_mode)) {
-      if (buf.st_size > _stat.st_size && 
+      if (buf.st_size > _stat.st_size &&
           (_fcntlFlags & (O_WRONLY|O_RDWR)) != 0) {
         errno = 0;
         JASSERT ( truncate ( _path.c_str(), _stat.st_size ) ==  0 )
@@ -1087,7 +1135,7 @@ int dmtcp::FileConnection::openFile()
   int fd;
   JASSERT(WorkerState::currentState() == WorkerState::RESTARTING);
 
-  /* 
+  /*
    * This file was not checkpointed by this process so it won't be restored by
    * this process. Thus, we wait while some other process restores this file
    */
@@ -1118,7 +1166,6 @@ int dmtcp::FileConnection::openFile()
 
 void dmtcp::FileConnection::restoreFile()
 {
-  int fd;
   JASSERT(WorkerState::currentState() == WorkerState::RESTARTING);
   JASSERT(_checkpointed);
 
@@ -1163,7 +1210,8 @@ void dmtcp::FileConnection::saveFile(int fd)
     CopyFile(_path, savedFilePath);
     return;
   } else if (_type == FileConnection::FILE_DELETED) {
-    const size_t bufSize = 2 * PAGE_SIZE;
+    long page_size = sysconf(_SC_PAGESIZE);
+    const size_t bufSize = 2 * page_size;
     char *buf = (char*)JALLOC_HELPER_MALLOC(bufSize);
 
     int destFd = open(savedFilePath.c_str(), O_CREAT | O_WRONLY | O_TRUNC,
@@ -1175,11 +1223,11 @@ void dmtcp::FileConnection::saveFile(int fd)
     int readBytes, writtenBytes;
     while(1) {
       readBytes = Util::readAll(fd, buf, bufSize);
-      JASSERT(readBytes != -1) 
+      JASSERT(readBytes != -1)
         (_path) (JASSERT_ERRNO) .Text("Read Failed");
       if (readBytes == 0) break;
       writtenBytes = Util::writeAll(destFd, buf, readBytes);
-      JASSERT(writtenBytes != -1) 
+      JASSERT(writtenBytes != -1)
         (savedFilePath) (JASSERT_ERRNO) .Text("Write failed.");
     }
 
@@ -1192,19 +1240,8 @@ void dmtcp::FileConnection::saveFile(int fd)
 
 dmtcp::string dmtcp::FileConnection::getSavedFilePath(const dmtcp::string& path)
 {
-  const char *cwd_env = getenv( ENV_VAR_CHECKPOINT_DIR );
-  dmtcp::string cwd;
-  if ( cwd_env == NULL ) {
-    cwd = jalib::Filesystem::GetCWD();
-  } else {
-    cwd = cwd_env;
-  }
-
-  JASSERT (!_ckptFilesDir.empty());
-
   dmtcp::ostringstream os;
-  os << cwd 
-     << "/" << _ckptFilesDir
+  os << _ckptFilesDir
      << "/" << jalib::Filesystem::BaseName(_path) << "_" << _id.conId();
 
   return os.str();
@@ -1567,7 +1604,7 @@ static bool ptmxTestPacketMode(int masterFd) {
   tmp_buf[0] = 'x'; /* Don't set '\n'.  Could be converted to "\r\n". */
   /* Give the masterFd something to read. */
   JWARNING ((rc = write(slave_fd, tmp_buf, 1)) == 1) (rc) .Text("write failed");
-  tcdrain(slave_fd);
+  //tcdrain(slave_fd);
   _real_close(slave_fd);
 
   /* Read the 'x':  If we also see a command byte, it's packet mode */
@@ -1623,7 +1660,7 @@ static ssize_t writeOnePacket(int fd, const void *origBuf, bool isPacketMode) {
   typedef int hdr;
   int count = *(hdr *)origBuf;
   int cum_count = 0;
-  int rc;
+  int rc = 0; // Trigger JASSERT if not modified below.
   if (count == 0)
     return sizeof(hdr);  // count of zero means we're done, hdr consumed
   // FIXME:  It would be nice to restore packet mode (flow control, etc.)
@@ -1643,9 +1680,9 @@ static ssize_t writeOnePacket(int fd, const void *origBuf, bool isPacketMode) {
 static ssize_t ptmxWriteAll(int fd, const void *buf, bool isPacketMode) {
   typedef int hdr;
   ssize_t cum_count = 0;
-  int rc;
+  ssize_t rc;
   while ((rc = writeOnePacket(fd, (char *)buf+cum_count, isPacketMode))
-	 > sizeof(hdr)) {
+	 > (ssize_t)sizeof(hdr)) {
     cum_count += rc;
   }
   JASSERT (rc < 0 || rc == sizeof(hdr)) (rc) (cum_count);
@@ -1733,7 +1770,7 @@ void dmtcp::StdioConnection::restore ( const dmtcp::vector<int>& fds, Connection
       JTRACE("Skipping restore of STDIO, just inherit from parent")(fd);
       continue;
     }
-    int oldFd;
+    int oldFd = -1;
     switch(_type){
       case STDIO_IN:
         JTRACE("Restoring STDIN")(fd);
