@@ -19,13 +19,14 @@
  *  <http://www.gnu.org/licenses/>.                                         *
  ****************************************************************************/
 
+#include "constants.h"
+#include "syscallwrappers.h"
 #include "connectionmanager.h"
 
 #include  "../jalib/jfilesystem.h"
 #include  "../jalib/jconvert.h"
 #include  "../jalib/jassert.h"
 #include "protectedfds.h"
-#include "syscallwrappers.h"
 #include "util.h"
 
 #include <sys/types.h>
@@ -37,10 +38,6 @@
 #include <sys/types.h>
 #include <sys/wait.h>
 
-#ifdef RECORD_REPLAY
-#define read _real_read
-#endif
-
 static dmtcp::string _procFDPath ( int fd )
 {
   return "/proc/self/fd/" + jalib::XToString ( fd );
@@ -79,15 +76,23 @@ dmtcp::ConnectionToFds::ConnectionToFds ( KernelDeviceToConnection& source )
   _procname = jalib::Filesystem::GetProgramName();
   _hostname = jalib::Filesystem::GetCurrentHostname();
   _inhostname = jalib::Filesystem::GetCurrentHostname();
-  _pid = UniquePid::ThisProcess();
-  _ppid = UniquePid::ParentProcess();
+  _upid = UniquePid::ThisProcess();
+  _uppid = UniquePid::ParentProcess();
 
-  for ( size_t i=0; i<fds.size(); ++i )
-  {
+  for ( size_t i=0; i<fds.size(); ++i ) {
     if ( _isBadFd ( fds[i] ) ) continue;
     if ( ProtectedFDs::isProtected ( fds[i] ) ) continue;
-    Connection* con = &source.retrieve ( fds[i] );
-    _table[con->id() ].push_back ( fds[i] );
+
+#ifdef IBV
+    dmtcp::string device = dmtcp::KernelDeviceToConnection::instance().fdToDevice(fds[i]);
+
+    if(!Util::strStartsWith(device, "/dev/infiniband/") &&
+       !Util::strStartsWith(device, "infinibandevent:"))
+#endif
+    {
+      Connection* con = &source.retrieve ( fds[i] );
+      _table[con->id() ].push_back ( fds[i] );
+    }
   }
 }
 
@@ -105,6 +110,16 @@ dmtcp::Connection& dmtcp::KernelDeviceToConnection::retrieve ( int fd )
   dmtcp::string device = fdToDevice ( fd );
   JASSERT ( device.length() > 0 ) ( fd ).Text ( "invalid fd" );
   iterator i = _table.find ( device );
+
+  if (i == _table.end() && Util::strStartsWith(device, "socket:[")) {
+    JWARNING ( false ) ( fd ) ( device ) ( _table.size() )
+      .Text ( "failed to find connection for fd. Assuming External Socket." );
+    TcpConnection *con = new TcpConnection(-1, -1, -1);
+    con->markExternalConnect();
+    create(fd, con);
+    i = _table.find ( device );
+  }
+
   JASSERT ( i != _table.end() ) ( fd ) ( device ) ( _table.size() ).Text ( "failed to find connection for fd" );
   return ConnectionList::instance() [i->second];
 }
@@ -162,11 +177,18 @@ dmtcp::string dmtcp::KernelDeviceToConnection::fdToDevice ( int fd, bool noOnDem
   bool isPtmx  = (device.compare("/dev/ptmx") == 0);
   bool isPts   = Util::strStartsWith(device, "/dev/pts/");
 
-  bool isBSDMaster  = (Util::strStartsWith(device, "/dev/pty") && 
+  bool isBSDMaster  = (Util::strStartsWith(device, "/dev/pty") &&
                        device.compare("/dev/pty") != 0);
   bool isBSDSlave   = (Util::strStartsWith(device, "/dev/tty") &&
                        device.compare("/dev/tty")) != 0;
+#ifdef IBV
+  bool isInfinibandDevice   = Util::strStartsWith(device, "/dev/infiniband/");
+  bool isInfinibandConnection   = Util::strStartsWith(device, "infinibandevent:");
 
+  if ( isInfinibandDevice || isInfinibandConnection ) {
+    return device;
+  } else
+#endif
   if ( isTty ) {
     dmtcp::string deviceName = "tty:" + device;
 
@@ -305,7 +327,7 @@ dmtcp::string dmtcp::KernelDeviceToConnection::fdToDevice ( int fd, bool noOnDem
       }
 
       return deviceName;
-    } else if (S_ISREG(buf.st_mode) || S_ISCHR(buf.st_mode) || 
+    } else if (S_ISREG(buf.st_mode) || S_ISCHR(buf.st_mode) ||
                S_ISDIR(buf.st_mode) || S_ISBLK(buf.st_mode)) {
       /* /dev/null is a character special file (non-regular file) */
       dmtcp::string deviceName = "file["+jalib::XToString ( fd ) +"]:" + device;
@@ -346,7 +368,7 @@ dmtcp::string dmtcp::KernelDeviceToConnection::fdToDevice ( int fd, bool noOnDem
       JASSERT(false) (device) .Text("Unimplemented file type.");
     }
   }
-  //JWARNING(false) (device) .Text("UnImplemented Connection Type.");
+  //JWARNING(false) (device) .Text("Unimplemented Connection Type.");
   return device;
 }
 
@@ -381,7 +403,7 @@ void dmtcp::KernelDeviceToConnection::erase( const ConnectionIdentifier& con )
   JTRACE("WARNING:: failed to find connection in table to erase it")(con);
 }
 
-dmtcp::string 
+dmtcp::string
 dmtcp::KernelDeviceToConnection::getDevice( const ConnectionIdentifier& con )
 {
   for(iterator i = _table.begin(); i!=_table.end(); ++i){
@@ -558,16 +580,17 @@ void dmtcp::ConnectionList::scanForPreExisting()
 void dmtcp::KernelDeviceToConnection::handlePreExistingFd ( int fd )
 {
   //this has the side effect of on-demand creating everything except sockets
-  dmtcp::string device = KernelDeviceToConnection::instance().fdToDevice ( fd, true );
+  dmtcp::string device =
+    KernelDeviceToConnection::instance().fdToDevice ( fd, true );
 
   JTRACE ( "scanning pre-existing device" ) ( fd ) ( device );
 
   //so if it doesn't exist it must be a socket
   if ( _table.find ( device ) == _table.end() )
   {
-    if ( fd <= 2 )
+    if ( Util::strStartsWith(device, "file"))
     {
-      create(fd, new StdioConnection(fd));
+      device = KernelDeviceToConnection::instance().fdToDevice (fd);
     }
     else if ( device.compare("/dev/tty") == 0 )
     {
@@ -580,10 +603,11 @@ void dmtcp::KernelDeviceToConnection::handlePreExistingFd ( int fd )
       PtyConnection *con = new PtyConnection ( device, device, type );
       create ( fd, con );
     }
-    else if ( Util::strStartsWith(device, "/dev/pts/")) 
+    else if ( Util::strStartsWith(device, "/dev/pts/"))
     {
       dmtcp::string deviceName = "pts["+jalib::XToString ( fd ) +"]:" + device;
-      JNOTE ( "Found pre-existing PTY connection, will be restored as current TTY" )
+      JNOTE ( "Found pre-existing PTY connection, "
+              "will be restored as current TTY" )
         ( fd ) ( deviceName );
 
       int type = dmtcp::PtyConnection::PTY_CTTY;
@@ -591,9 +615,14 @@ void dmtcp::KernelDeviceToConnection::handlePreExistingFd ( int fd )
       PtyConnection *con = new PtyConnection ( device, device, type );
       create ( fd, con );
     }
+    else if (fd <= 2)
+    {
+      create(fd, new StdioConnection(fd));
+    }
     else
     {
-      JNOTE ( "found pre-existing socket... will not be restored" ) ( fd ) ( device );
+      JNOTE ( "found pre-existing socket... will not be restored" )
+        ( fd ) ( device );
       TcpConnection* con = new TcpConnection ( 0, 0, 0 );
       con->markPreExisting();
       create ( fd, con );
@@ -620,7 +649,7 @@ void dmtcp::ConnectionToFds::serialize ( jalib::JBinarySerializer& o )
   JSERIALIZE_ASSERT_POINT ( "dmtcp::ConnectionToFds:" );
 
   // Current process information
-  o & _procname & _inhostname & _pid & _ppid;
+  o & _procname & _inhostname & _upid & _uppid;
 
   size_t numCons = _table.size();
   o & numCons;
@@ -629,7 +658,8 @@ void dmtcp::ConnectionToFds::serialize ( jalib::JBinarySerializer& o )
   {
     _hostname = jalib::Filesystem::GetCurrentHostname();
     o & _hostname;
-    JTRACE("Writing hostname to checkpoint file")(_hostname)(_inhostname)(_procname)(_ppid);
+    JTRACE("Writing hostname to checkpoint file")
+      (_hostname) (_inhostname) (_procname) (_uppid);
 
     // Save connections
     for ( iterator i=_table.begin(); i!=_table.end(); ++i )
@@ -854,18 +884,42 @@ bool dmtcp::PtsToSymlink::exists( dmtcp::string device )
 }
 */
 
-pid_t dmtcp::ConnectionToFds::gzip_child_pid = -1;
+pid_t dmtcp::ConnectionToFds::ext_decomp_pid = -1;
 static void close_ckpt_to_read(const int fd)
 {
     int status;
     int rc;
-    while (-1 == (rc = close(fd)) && errno == EINTR) ;
-    JASSERT (rc != -1) ("close:") (JASSERT_ERRNO);
-    if (dmtcp::ConnectionToFds::gzip_child_pid != -1) {
-      while (-1 == (rc = waitpid(dmtcp::ConnectionToFds::gzip_child_pid,
-			         &status, 0)) && errno == EINTR) ;
-      JASSERT (rc != -1) ("waitpid:") (JASSERT_ERRNO);
-      dmtcp::ConnectionToFds::gzip_child_pid = -1;
+    int pid = dmtcp::ConnectionToFds::ext_decomp_pid;
+    if( pid != -1 ) {
+      // First close fd to let decompressor know that we want to close it
+      while (-1 == (rc = close(fd)) && errno == EINTR) ;
+      JASSERT (rc != -1) ("close:") (JASSERT_ERRNO);
+
+      // Kill the decompressor process
+      // 1. Send SIGTERM to give the decompressor a chance to clean up
+      JASSERT (kill(pid, SIGTERM) != -1)("kill:") (JASSERT_ERRNO);
+      // 2. Wait 3 seconds for decompressor termination
+      rc = 0;
+      for(int i = 0; (i < 3000 && rc != pid); i++){
+        struct timespec sleepTime = {0, 1*1000*1000};
+        nanosleep(&sleepTime, NULL);
+        rc = waitpid(pid, &status, WNOHANG);
+      }
+      // 3. If the decompressor process still exists
+      if( rc != pid ){
+        if( (rc = kill(pid, SIGKILL)) == -1 && errno != ESRCH ){
+          // process exists but we failed to kill it
+          JASSERT (rc != -1)("kill:") (JASSERT_ERRNO);
+        }else{
+          // process exists and it was SIGKILL'ed. Endless wait for exit.
+          while(0 == (rc = waitpid(pid,&status, WNOHANG))){
+            struct timespec sleepTime = {0, 1*1000*1000};
+            nanosleep(&sleepTime, NULL);
+          }
+        }
+        JASSERT (rc == pid) ("waitpid:") (JASSERT_ERRNO);
+      }
+      dmtcp::ConnectionToFds::ext_decomp_pid = -1;
     }
 }
 
@@ -875,6 +929,11 @@ static void close_ckpt_to_read(const int fd)
 // Copied from mtcp/mtcp_restart.c.
 #define DMTCP_MAGIC_FIRST 'D'
 #define GZIP_FIRST 037
+#ifdef HBICT_DELTACOMP
+#define HBICT_FIRST 'H'
+#endif
+
+
 char *mtcp_executable_path(char *filename);
 static char first_char(const char *filename)
 {
@@ -884,7 +943,7 @@ static char first_char(const char *filename)
     fd = open(filename, O_RDONLY);
     JASSERT(fd >= 0)(filename).Text("ERROR: Cannot open filename");
 
-    rc = read(fd, &c, 1);
+    rc = _real_read(fd, &c, 1);
     JASSERT(rc == 1)(filename).Text("ERROR: Error reading from filename");
 
     close(fd);
@@ -896,14 +955,20 @@ static char first_char(const char *filename)
 // MTCP code in:  mtcp/mtcp_restart.c:open_ckpt_to_read()
 // A previous version tried to replace this with popen, causing a regression:
 //   (no call to pclose, and possibility of using a wrong fd).
-// Returns fd; sets dmtcp::gzip_child_pid::ConnectionToFds, if gzip compression.
+// Returns fd; sets dmtcp::ext_decomp_pid::ConnectionToFds, if checkpoint was compressed.
 static int open_ckpt_to_read(const char *filename)
 {
     int fd;
     int fds[2];
     char fc;
+    const char *decomp_path;
+    const char **decomp_args;
     const char *gzip_path = "gzip";
     static const char * gzip_args[] = { "gzip", "-d", "-", NULL };
+#ifdef HBICT_DELTACOMP
+    const char *hbict_path = "hbict";
+    static const char *hbict_args[] = { "hbict", "-r", NULL };
+#endif
     pid_t cpid;
 
     fc = first_char(filename);
@@ -912,24 +977,38 @@ static int open_ckpt_to_read(const char *filename)
 
     if(fc == DMTCP_MAGIC_FIRST) /* no compression */
         return fd;
-    else if(fc == GZIP_FIRST) /* gzip */
-    {
-        JASSERT(pipe(fds) != -1)(filename).Text("Cannote create pipe to execute gunzip to decompress checkpoint file!");
+#ifdef HBICT_DELTACOMP
+    else if (fc == GZIP_FIRST || fc == HBICT_FIRST){ /* External compression */
+#else
+    else if(fc == GZIP_FIRST){ /* gzip */
+#endif
+        if( fc == GZIP_FIRST ){
+          decomp_path = gzip_path;
+          decomp_args = gzip_args;
+        }
+#ifdef HBICT_DELTACOMP
+        else{
+          decomp_path = hbict_path;
+          decomp_args = hbict_args;
+        }
+#endif
+
+        JASSERT(pipe(fds) != -1)(filename).Text("Cannot create pipe to execute gunzip to decompress checkpoint file!");
 
         cpid = _real_fork();
 
         JASSERT(cpid != -1).Text("ERROR: Cannot fork to execute gunzip to decompress checkpoint file!");
         if(cpid > 0) /* parent process */
         {
-           JTRACE ( "created gzip child process to uncompress checkpoint file")(cpid);
-            dmtcp::ConnectionToFds::gzip_child_pid = cpid;
+           JTRACE ( "created child process to uncompress checkpoint file")(cpid);
+            dmtcp::ConnectionToFds::ext_decomp_pid = cpid;
             close(fd);
             close(fds[1]);
             return fds[0];
         }
         else /* child process */
         {
-           JTRACE ( "child process, will exec into gzip");
+           JTRACE ( "child process, will exec into external de-compressor");
             fd = dup(dup(dup(fd)));
             fds[1] = dup(fds[1]);
             close(fds[0]);
@@ -938,14 +1017,13 @@ static int open_ckpt_to_read(const char *filename)
             close(fd);
             JASSERT(dup2(fds[1], STDOUT_FILENO) == STDOUT_FILENO);
             close(fds[1]);
-            _real_execvp(gzip_path, (char **)gzip_args);
-            JASSERT(gzip_path!=NULL)(gzip_path).Text("Failed to launch gzip.");
+            _real_execvp(decomp_path, (char **)decomp_args);
+            JASSERT(decomp_path!=NULL)(decomp_path).Text("Failed to launch gzip.");
             /* should not get here */
             JASSERT(false)("ERROR: Decompression failed!  No restoration will be performed!  Cancelling now!");
             abort();
         }
-    }
-    else /* invalid magic number */
+    } else /* invalid magic number */
         JASSERT(false).Text("ERROR: Invalid magic number in this checkpoint file!");
     // NOT_REACHED
     return -1;
@@ -953,13 +1031,13 @@ static int open_ckpt_to_read(const char *filename)
 
 // See comments above for open_ckpt_to_read()
 int dmtcp::ConnectionToFds::openDmtcpCheckpointFile(const dmtcp::string& path){
-  // Function also sets dmtcp::gzip_child_pid::ConnectionToFds
+  // Function also sets dmtcp::ext_decomp_pid::ConnectionToFds
   int fd = open_ckpt_to_read( path.c_str() );
   // The rest of this function is for compatibility with original definition.
   JASSERT(fd>=0)(path).Text("Failed to open file.");
   char buf[512];
   const int len = strlen(DMTCP_FILE_HEADER);
-  JASSERT(read(fd, buf, len)==len)(path).Text("read() failed");
+  JASSERT(_real_read(fd, buf, len)==len)(path).Text("_real_read() failed");
   if(strncmp(buf, DMTCP_FILE_HEADER, len)==0){
     JTRACE("opened checkpoint file [uncompressed]")(path);
   }else{
@@ -974,7 +1052,7 @@ int dmtcp::ConnectionToFds::openDmtcpCheckpointFile(const dmtcp::string& path){
   JASSERT(fd>=0)(path).Text("Failed to open file.");
   char buf[512];
   const int len = strlen(DMTCP_FILE_HEADER);
-  JASSERT(read(fd, buf, len)==len)(path).Text("read() failed");
+  JASSERT(_real_read(fd, buf, len)==len)(path).Text("_real_read() failed");
   if(strncmp(buf, DMTCP_FILE_HEADER, len)==0){
     JTRACE("opened checkpoint file [uncompressed]")(path);
     return fd;
@@ -985,7 +1063,7 @@ int dmtcp::ConnectionToFds::openDmtcpCheckpointFile(const dmtcp::string& path){
     JASSERT(t!=NULL)(path)(cmd).Text("Failed to launch gzip.");
     JTRACE ( "created gzip child process to uncompress checkpoint file");
     fd = fileno(t);
-    JASSERT(read(fd, buf, len)==len)(cmd)(path).Text("Invalid checkpoint file");
+    JASSERT(_real_read(fd, buf, len)==len)(cmd)(path).Text("Invalid checkpoint file");
     JASSERT(strncmp(buf, DMTCP_FILE_HEADER, len)==0)(path).Text("Invalid checkpoint file");
     JTRACE("opened checkpoint file [compressed]")(path);
     return fd;
