@@ -23,14 +23,14 @@
 #include <stdlib.h>
 #include <string>
 #include <stdio.h>
-#include  "../jalib/jassert.h"
 #include <ctype.h>
+#include  "../jalib/jassert.h"
 #include  "../jalib/jfilesystem.h"
 #include  "../jalib/jconvert.h"
 #include "constants.h"
-#include "dmtcpworker.h"
 #include "dmtcpmessagetypes.h"
 #include "syscallwrappers.h"
+#include "dmtcpcoordinatorapi.h"
 #include "util.h"
 #include <errno.h>
 #include <sys/types.h>
@@ -42,14 +42,14 @@
 #include <sys/resource.h>
 #include <sys/personality.h>
 #include <string.h>
-#include <dlfcn.h>
+
+#define BINARY_NAME "dmtcp_checkpoint"
 
 int testMatlab(const char *filename);
+int testJava(char **argv);
 bool testSetuid(const char *filename);
 void testStaticallyLinked(const char *filename);
 bool testScreen(char **argv, char ***newArgv);
-void adjust_rlimit_stack();
-int elfType(const char *pathname, bool *isElf, bool *is32bitElf);
 
 // gcc-4.3.4 -Wformat=2 issues false positives for warnings unless the format
 // string has at least one format specifier with corresponding format argument.
@@ -64,13 +64,20 @@ static const char* theUsage =
   "      Port where dmtcp_coordinator is run (default: 7779)\n"
   "  --gzip, --no-gzip, (environment variable DMTCP_GZIP=[01]):\n"
   "      Enable/disable compression of checkpoint images (default: 1)\n"
+  "      WARNING:  gzip adds seconds.  Without gzip, ckpt is often < 1 s\n"
+#ifdef HBICT_DELTACOMP
+  "  --hbict, --no-hbict, (environment variable DMTCP_HBICT=[01]):\n"
+  "      Enable/disable compression of checkpoint images (default: 1)\n"
+#endif
+  "  --prefix <arg>:\n"
+  "      Prefix where DMTCP is installed on remote nodes.\n"
   "  --ckptdir, -c, (environment variable DMTCP_CHECKPOINT_DIR):\n"
   "      Directory to store checkpoint images (default: ./)\n"
   "  --tmpdir, -t, (environment variable DMTCP_TMPDIR):\n"
   "      Directory to store temporary files \n"
   "        (default: $TMDPIR/dmtcp-$USER@$HOST or /tmp/dmtcp-$USER@$HOST)\n"
   "  --join, -j:\n"
-  "      Join an existing coordinator, raise error if one already exists\n"
+  "      Join an existing coordinator, raise error if one doesn't already exist\n"
   "  --new, -n:\n"
   "      Create a new coordinator, raise error if one already exists\n"
   "  --new-coordinator:\n"
@@ -88,18 +95,17 @@ static const char* theUsage =
   "      Checkpoint open files and restore old working dir. (Default: do neither)\n"
   "  --mtcp-checkpoint-signal:\n"
   "      Signal number used internally by MTCP for checkpointing (default: 12)\n"
+  "  --with-module (environment variable DMTCP_MODULE):\n"
+  "      Colon-separated list of DMTCP modules to be preloaded with DMTCP.\n"
+  "      (Absolute pathnames are required.)\n"
   "  --quiet, -q, (or set environment variable DMTCP_QUIET = 0, 1, or 2):\n"
-  "      Skip banner and NOTE messages; if given twice, also skip WARNINGs\n\n"
-  "See http://dmtcp.sf.net/ for more information.\n"
-;
-
-static const char* theBanner =
-  "DMTCP/MTCP  Copyright (C) 2006-2010  Jason Ansel, Michael Rieker,\n"
-  "                                       Kapil Arya, and Gene Cooperman\n"
-  "This program comes with ABSOLUTELY NO WARRANTY.\n"
-  "This is free software, and you are welcome to redistribute it\n"
-  "under certain conditions; see COPYING file for details.\n"
-  "(Use flag \"-q\" to hide this message.)\n\n"
+  "      Skip banner and NOTE messages; if given twice, also skip WARNINGs\n"
+  "  --help:\n"
+  "      Print this message and exit.\n"
+  "  --version:\n"
+  "      Print version information and exit.\n"
+  "\n"
+  "See " PACKAGE_URL " for more information.\n"
 ;
 
 // FIXME:  The warnings below should be collected into a single function,
@@ -112,213 +118,150 @@ static const char* theBanner =
 
 static dmtcp::string _stderrProcPath()
 {
-  return "/proc/" + jalib::XToString ( getpid() ) + "/fd/" + jalib::XToString ( fileno ( stderr ) );
-}
-
-static void *get_libc_symbol ( const char* name )
-{
-  static void* handle = NULL;
-  if ( handle==NULL && ( handle=dlopen ( LIBC_FILENAME,RTLD_NOW ) ) == NULL )
-  {
-    fprintf ( stderr, "dmtcp: get_libc_symbol: ERROR in dlopen: %s \n",
-              dlerror() );
-    abort();
-  }
-
-  void* tmp = dlsym ( handle, name );
-  if ( tmp == NULL )
-  {
-    JASSERT_STDERR
-      << "dmtcp: get_libc_symbol: ERROR finding symbol "
-      << name << " using dlsym: " << dlerror() << " \n"
-      << "       Will fail if user-app tries to call this symbol.\n";
-    //abort();
-  }
-  return tmp;
-}
-
-#ifdef RECORD_REPLAY
-static void *get_libpthread_symbol ( const char* name )
-{
-  static void* handle = NULL;
-  if ( handle==NULL && ( handle=dlopen ( LIBPTHREAD_FILENAME, RTLD_NOW ) ) == NULL )
-  {
-    fprintf ( stderr,"dmtcp: get_libpthread_symbol: ERROR in dlopen: %s \n",
-              dlerror() );
-    abort();
-  }
-
-  void* tmp = dlsym ( handle, name );
-  if ( tmp==NULL )
-  {
-    fprintf ( stderr,"dmtcp: get_libpthread_symbol: ERROR in dlsym: %s \n"
-              dlerror() );
-    abort();
-  }
-  return tmp;
-}
-#endif // RECORD_REPLAY
-
-static void prepareDmtcpWrappers()
-{
-#ifndef ENABLE_DLOPEN
-  unsigned int wrapperOffsetArray[numLibcWrappers];
-  char *glibc_base_function_addr = NULL;
-
-# define _FIRST_BASE_ADDR(name) if (glibc_base_function_addr == NULL) \
-				  glibc_base_function_addr = (char *)&name;
-  FOREACH_GLIBC_BASE_FUNC(_FIRST_BASE_ADDR);
-
-# define _GET_OFFSET(x) do { \
-    char *addr = (char*)get_libc_symbol(#x); \
-    int offset = (addr == NULL) ? -1 : (addr - glibc_base_function_addr); \
-    if (addr == NULL) \
-      offset = -1; \
-    else \
-      offset = addr - glibc_base_function_addr; \
-    wrapperOffsetArray[enum_ ## x] = offset; \
-  } while(0);
-
-  FOREACH_GLIBC_FUNC_WRAPPER(_GET_OFFSET);
-
-  dmtcp::ostringstream os;
-  for (int i = 0; i < numLibcWrappers; i++) {
-    os << std::hex << wrapperOffsetArray[i] << ";";
-  }
-
-  setenv(ENV_VAR_LIBC_FUNC_OFFSETS, os.str().c_str(), 1);
-#ifdef RECORD_REPLAY
-  long wrapperOffsetArrayPthread[numLibpthreadWrappers];
-  char *libpthread_base_function_addr = (char*)&LIBPTHREAD_BASE_FUNC;
-
-# define _GET_OFFSET_LIBPTHREAD(x) \
-    wrapperOffsetArrayPthread[enum_ ## x] = ((char*)get_libpthread_symbol(#x) - libpthread_base_function_addr);
-
-  FOREACH_PTHREAD_FUNC_WRAPPER(_GET_OFFSET_LIBPTHREAD);
-
-  dmtcp::ostringstream os_pthread;
-  for (int i = 0; i < numLibpthreadWrappers; i++) {
-    os_pthread << wrapperOffsetArrayPthread[i] << ";";
-  }
-
-  setenv(ENV_VAR_LIBPTHREAD_FUNC_OFFSETS, os_pthread.str().c_str(), 1);
-#endif //RECORD_REPLAY
-#else
-  unsetenv(ENV_VAR_LIBC_FUNC_OFFSETS);
-#ifdef RECORD_REPLAY
-  unsetenv(ENV_VAR_LIBPTHREAD_FUNC_OFFSETS);
-#endif
-#endif
-
-#ifdef PTRACE
-  /* For the sake of dlsym wrapper. We compute the address of _real_dlsym by
-   * adding dlsym_offset to the address of dlopen after the exec into the user
-   * application. */
-  void* tmp1 = NULL;
-  void* tmp2 = NULL;
-  int tmp3;
-  static void* handle = NULL;
-  if (handle == NULL && (handle = dlopen("libdl.so", RTLD_NOW)) == NULL) {
-    fprintf(stderr, "dmtcp: get_libc_symbol: ERROR in dlopen: %s \n", dlerror());
-    abort();
-  }
-  tmp1 = (void *)&dlopen;
-  tmp2 = (void *)&dlsym;
-  tmp3 = (char *)tmp2 - (char *)tmp1;
-  char str[21] = {0} ;
-  sprintf(str, "%d", tmp3);
-  setenv(ENV_VAR_DLSYM_OFFSET, str, 0);
-  dlclose(handle);
-#endif
+  return "/proc/" + jalib::XToString(getpid()) + "/fd/"
+         + jalib::XToString(fileno(stderr));
 }
 
+static bool isSSHSlave=false;
+static bool autoStartCoordinator=true;
+static bool checkpointOpenFiles=false;
+static int allowedModes = dmtcp::DmtcpCoordinatorAPI::COORD_ANY;
 
 //shift args
 #define shift argc--,argv++
-int main ( int argc, char** argv )
+static void processArgs(int *orig_argc, char ***orig_argv)
 {
-  bool isSSHSlave=false;
-  bool autoStartCoordinator=true;
-  bool checkpointOpenFiles=false;
-  int allowedModes = dmtcp::DmtcpWorker::COORD_ANY;
+  char argc = *orig_argc;
+  char **argv = *orig_argv;
 
-  if (! getenv(ENV_VAR_QUIET))
-    setenv(ENV_VAR_QUIET, "0", 0);
+  if (argc == 1) {
+    JASSERT_STDERR << DMTCP_VERSION_AND_COPYRIGHT_INFO;
+    JASSERT_STDERR << "(For help:  " << argv[0] << " --help)\n\n";
+    exit(DMTCP_FAIL_RC);
+  }
 
   //process args
   shift;
-  while(true){
+  while (true) {
     dmtcp::string s = argc>0 ? argv[0] : "--help";
-    if((s=="--help" || s=="-h") && argc==1){
+    if ((s=="--help") && argc==1) {
       JASSERT_STDERR << theUsage;
-      //fprintf(stderr, theUsage, "");
-      return 1;
-    }else if(s=="--ssh-slave"){
+      exit(DMTCP_FAIL_RC);
+    } else if ((s=="--version") && argc==1) {
+      JASSERT_STDERR << DMTCP_VERSION_AND_COPYRIGHT_INFO;
+      exit(DMTCP_FAIL_RC);
+    } else if (s=="--ssh-slave") {
       isSSHSlave = true;
       shift;
-    }else if(s == "--no-check"){
+    } else if (s == "--no-check") {
       autoStartCoordinator = false;
       shift;
-    }else if(s == "-j" || s == "--join"){
-      allowedModes = dmtcp::DmtcpWorker::COORD_JOIN;
+    } else if (s == "-j" || s == "--join") {
+      allowedModes = dmtcp::DmtcpCoordinatorAPI::COORD_JOIN;
       shift;
-    }else if(s == "--gzip"){
+    } else if (s == "--gzip") {
       setenv(ENV_VAR_COMPRESSION, "1", 1);
       shift;
-    }else if(s == "--no-gzip"){
+    } else if (s == "--no-gzip") {
       setenv(ENV_VAR_COMPRESSION, "0", 1);
       shift;
-    }else if(s == "-n" || s == "--new"){
-      allowedModes = dmtcp::DmtcpWorker::COORD_NEW;
+    }
+#ifdef HBICT_DELTACOMP
+    else if (s == "--hbict") {
+      setenv(ENV_VAR_DELTACOMPRESSION, "1", 1);
       shift;
-    }else if(s == "--new-coordinator"){
-      allowedModes = dmtcp::DmtcpWorker::COORD_FORCE_NEW;
+    } else if (s == "--no-hbict") {
+      setenv(ENV_VAR_DELTACOMPRESSION, "0", 1);
       shift;
-    }else if(s == "-b" || s == "--batch"){
-      allowedModes = dmtcp::DmtcpWorker::COORD_BATCH;
+    }
+#endif
+    else if (s == "-n" || s == "--new") {
+      allowedModes = dmtcp::DmtcpCoordinatorAPI::COORD_NEW;
       shift;
-    }else if(s == "-i" || s == "--interval"){
-      setenv(ENV_VAR_CKPT_INTR, argv[1], 1);
+    } else if (s == "--new-coordinator") {
+      allowedModes = dmtcp::DmtcpCoordinatorAPI::COORD_FORCE_NEW;
+      shift;
+    } else if (s == "-b" || s == "--batch") {
+      allowedModes = dmtcp::DmtcpCoordinatorAPI::COORD_BATCH;
+      shift;
+    } else if (s == "-i" || s == "--interval" ||
+             (s.c_str()[0] == '-' && s.c_str()[1] == 'i' &&
+              isdigit(s.c_str()[2]) ) ) {
+      if (isdigit(s.c_str()[2])) { // if -i5, for example
+        setenv(ENV_VAR_CKPT_INTR, s.c_str()+2, 1);
+        shift;
+      } else { // else -i 5
+        setenv(ENV_VAR_CKPT_INTR, argv[1], 1);
+        shift; shift;
+      }
+    } else if (argc>1 && (s == "-h" || s == "--host")) {
+      setenv(ENV_VAR_NAME_HOST, argv[1], 1);
       shift; shift;
-    }else if(argc>1 && (s == "-h" || s == "--host")){
-      setenv(ENV_VAR_NAME_ADDR, argv[1], 1);
-      shift; shift;
-    }else if(argc>1 && (s == "-p" || s == "--port")){
+    } else if (argc>1 && (s == "-p" || s == "--port")) {
       setenv(ENV_VAR_NAME_PORT, argv[1], 1);
       shift; shift;
-    }else if(argc>1 && (s == "-c" || s == "--ckptdir")){
+    } else if (argc>1 && (s == "--prefix")) {
+      setenv(ENV_VAR_PREFIX_PATH, argv[1], 1);
+      shift; shift;
+    } else if (argc>1 && (s == "-c" || s == "--ckptdir")) {
       setenv(ENV_VAR_CHECKPOINT_DIR, argv[1], 1);
       shift; shift;
-    }else if(argc>1 && (s == "-t" || s == "--tmpdir")){
+    } else if (argc>1 && (s == "-t" || s == "--tmpdir")) {
       setenv(ENV_VAR_TMPDIR, argv[1], 1);
       shift; shift;
-    }else if(argc>1 && s == "--mtcp-checkpoint-signal"){
+    } else if (argc>1 && s == "--mtcp-checkpoint-signal") {
       setenv(ENV_VAR_SIGCKPT, argv[1], 1);
       shift; shift;
-    }else if(s == "--checkpoint-open-files"){
+    } else if (s == "--checkpoint-open-files") {
       checkpointOpenFiles = true;
       shift;
-    }else if(s == "-q" || s == "--quiet"){
+    } else if (s == "--with-module") {
+      setenv(ENV_VAR_MODULE, argv[1], 1);
+      shift; shift;
+    } else if (s == "-q" || s == "--quiet") {
       *getenv(ENV_VAR_QUIET) = *getenv(ENV_VAR_QUIET) + 1;
       // Just in case a non-standard version of setenv is being used:
       setenv(ENV_VAR_QUIET, getenv(ENV_VAR_QUIET), 1);
       shift;
-    }else if( (s.length()>2 && s.substr(0,2)=="--") ||
+    } else if ( (s.length()>2 && s.substr(0,2)=="--") ||
               (s.length()>1 && s.substr(0,1)=="-" ) ) {
       JASSERT_STDERR << "Invalid Argument\n";
       JASSERT_STDERR << theUsage;
-      return 1;
-    }else if(argc>1 && s=="--"){
+      exit(DMTCP_FAIL_RC);
+    } else if (argc>1 && s=="--") {
       shift;
       break;
-    }else{
+    } else {
       break;
     }
   }
+  *orig_argv = argv;
+}
+
+int main ( int argc, char** argv )
+{
+  initializeJalib();
+
+  if (! getenv(ENV_VAR_QUIET))
+    setenv(ENV_VAR_QUIET, "0", 0);
+
+  processArgs(&argc, &argv);
+
+  // If --ssh-slave and --prefix both are present, verify that the prefix-dir
+  // of this binary (dmtcp_checkpoint) is same as the one provided with
+  // --prefix
+  if (isSSHSlave && getenv(ENV_VAR_PREFIX_PATH) != NULL) {
+    const char *str = getenv(ENV_VAR_PREFIX_PATH);
+    dmtcp::string prefixDir = jalib::Filesystem::ResolveSymlink(str);
+    dmtcp::string programPrefixDir =
+      jalib::Filesystem::DirName(jalib::Filesystem::GetProgramDir());
+    JASSERT(prefixDir == programPrefixDir)
+      (prefixDir) (programPrefixDir);
+  }
 
   dmtcp::UniquePid::setTmpDir(getenv(ENV_VAR_TMPDIR));
-
-  jassert_quiet = *getenv(ENV_VAR_QUIET) - '0';
+  dmtcp::UniquePid::ThisProcess(true);
+  dmtcp::Util::initializeLogFile();
 
 #ifdef FORKED_CHECKPOINTING
   /* When this is robust, add --forked-checkpointing option on command-line,
@@ -330,7 +273,7 @@ int main ( int argc, char** argv )
 #endif
 
   if (jassert_quiet == 0)
-    JASSERT_STDERR << theBanner;
+    JASSERT_STDERR << DMTCP_BANNER;
 
   // This code will go away when zero-mapped pages are implemented in MTCP.
   struct rlimit rlim;
@@ -349,7 +292,7 @@ int main ( int argc, char** argv )
   // until finding region of memory segment with many zeroes.
   // Then mark as CS_ZERO_PAGES in MTCP instead of CS_RESTORE (or mark
   // entire segment as CS_ZERO_PAGES and then overwrite with CS_RESTORE
-  // region for portion to be read back fom checkpoint image.
+  // region for portion to be read back from checkpoint image.
   // For CS_ZERO_PAGES region, mmap // on restart, but don't write in zeroes.
   // Also, after checkpointing segment, munmap zero pages, and mmap them again.
   // Don't try to find all pages.  The above strategy may increase
@@ -358,36 +301,40 @@ int main ( int argc, char** argv )
   // - Gene
 
   testMatlab(argv[0]);
+  testJava(argv);  // Warn that -Xmx flag needed to limit virtual memory size
 
-  // If dmtcphijack.so is in standard search path and also has setgid access,
-  //   then LD_PRELOAD will work.  Otherwise, it will only work if the
-  //   application does not use setuid and setgid access.  So, we test
-  //   if the application does not use setuid/setgid.  (See 'man ld.so')
+  // If dmtcphijack.so is in standard search path and _also_ has setgid access,
+  //   then LD_PRELOAD will work.
+  // Otherwise, it will only work if the application does not use setuid and
+  //   setgid access.  So, we test //   if the application does not use
+  //   setuid/setgid.  (See 'man ld.so')
+  // FIXME:  ALSO DO THIS FOR execwrappers.cpp:dmtcpPrepareForExec()
+  //   Should pass dmtcphijack.so path, and let testSetuid determine
+  //     if setgid is set for it.  If so, no problem:  continue.
+  //   If not, call testScreen() and adapt 'screen' to run using
+  //     Util::patchArgvIfSetuid(argv[0], argv, &newArgv) (which shouldn't
+  //     will just modify argv[0] to point to /tmp/dmtcp-USER@HOST/screen
+  //     and other modifications:  doesn't need newArgv).
+  //   If it's not 'screen' and if no setgid for dmtcphijack.so, then testSetuid
+  //    should issue the warning, unset our LD_PRELOAD, and hope for the best.
+  //    A program like /usr/libexec/utempter/utempter (Fedora path)
+  //    is short-lived and can be safely run.  Ideally, we should
+  //    disable checkpoints while utempter is running, and enable checkpoints
+  //    when utempter finishes.  See possible model at
+  //    execwrappers.cpp:execLibProcessAndExit(), since the same applies
+  //    to running /lib/libXXX.so for running libraries as executables.
   if (testSetuid(argv[0])) {
     char **newArgv;
-    Util::patchArgvIfSetuid(argv[0], argv, &newArgv);
+    // THIS NEXT LINE IS DANGEROUS.  MOST setuid PROGRAMS CAN'T RUN UNPRIVILEGED
+    dmtcp::Util::patchArgvIfSetuid(argv[0], argv, &newArgv);
     argv = newArgv;
   };
 
-  prepareDmtcpWrappers();
-
-  if(autoStartCoordinator)
-     dmtcp::DmtcpWorker::startCoordinatorIfNeeded(allowedModes);
-
-  dmtcp::string dmtcphjk =
-    jalib::Filesystem::FindHelperUtility ( "dmtcphijack.so" );
-  dmtcp::string searchDir = jalib::Filesystem::GetProgramDir();
-
-  // Initialize JASSERT library here
-  dmtcp::ostringstream o;
-  o << dmtcp::UniquePid::getTmpDir() << "/jassertlog." << dmtcp::UniquePid(getpid());
-  JASSERT_INIT(o.str());
-
   if (argc > 0) {
     JTRACE("dmtcp_checkpoint starting new program:")(argv[0]);
   }
 
-  //setup CHECKPOINT_DIR
+  //set up CHECKPOINT_DIR
   if(getenv(ENV_VAR_CHECKPOINT_DIR) == NULL){
     const char* ckptDir = get_current_dir_name();
     if(ckptDir != NULL ){
@@ -405,22 +352,15 @@ int main ( int argc, char** argv )
 
   //TODO:
   // When stderr is a pseudo terminal for IPC between parent/child processes,
-  // this logic fails and JASSERT may write data to FD 2 (stderr)
-  // this will cause problems in programs that use FD 2 (stderr) for algorithmic things...
+  //  this logic fails and JASSERT may write data to FD 2 (stderr).
+  // This will cause problems in programs that use FD 2 (stderr) for
+  //  algorithmic things ...
   if ( stderrDevice.length() > 0
           && jalib::Filesystem::FileExists ( stderrDevice ) )
     setenv ( ENV_VAR_STDERR_PATH,stderrDevice.c_str(), 0 );
   else// if( isSSHSlave )
     setenv ( ENV_VAR_STDERR_PATH, "/dev/null", 0 );
 
-  // If dmtcp_checkpoint was called with user LD_PRELOAD, and if
-  //   if dmtcp_checkpoint survived the experience, then pass it back to user.
-  if (getenv("LD_PRELOAD"))
-    dmtcphjk = dmtcphjk + ":" + getenv("LD_PRELOAD");
-  setenv ( "LD_PRELOAD", dmtcphjk.c_str(), 1 );
-  JTRACE("getting value of LD_PRELOAD")(getenv("LD_PRELOAD"));
-  setenv ( ENV_VAR_HIJACK_LIB, dmtcphjk.c_str(), 0 );
-  setenv ( ENV_VAR_UTILITY_DIR, searchDir.c_str(), 0 );
   if ( getenv(ENV_VAR_SIGCKPT) != NULL )
     setenv ( "MTCP_SIGCKPT", getenv(ENV_VAR_SIGCKPT), 1);
   else
@@ -436,14 +376,15 @@ int main ( int argc, char** argv )
 #endif
 
   bool isElf, is32bitElf;
-  if  (Util::elfType(argv[0], &isElf, &is32bitElf) == -1) {
+  if  (dmtcp::Util::elfType(argv[0], &isElf, &is32bitElf) == -1) {
     // Couldn't read argv_buf
     // FIXME:  This could have been a symbolic link.  Don't issue an error,
     //         unless we're sure that the executable is not readable.
     JASSERT_STDERR <<
-      "*** ERROR:  Executable to run w/ DMTCP appears not to be readable.\n\n"
-      << argv[0];
-    exit(1);
+      "*** ERROR:  Executable to run w/ DMTCP appears not to be readable,\n"
+      "***         or no such executable in path.\n\n"
+      << argv[0] << "\n";
+    exit(DMTCP_FAIL_RC);
   } else {
 #if defined(__x86_64__) && !defined(CONFIG_M32)
     if (is32bitElf)
@@ -462,7 +403,41 @@ int main ( int argc, char** argv )
 // FIXME:  Unify this code with code prior to execvp in execwrappers.cpp
 //   Can use argument to dmtcpPrepareForExec() or getenv("DMTCP_...")
 //   from DmtcpWorker constructor, to distinguish the two cases.
-  adjust_rlimit_stack();
+  dmtcp::Util::adjustRlimitStack();
+
+  // FIXME: This call should be moved closer to call to execvp().
+  dmtcp::Util::prepareDlsymWrapper();
+
+  if (autoStartCoordinator)
+     dmtcp::DmtcpCoordinatorAPI::startCoordinatorIfNeeded(allowedModes);
+
+  // preloadLibs are to set LD_PRELOAD:
+  //   LD_PRELOAD=MODULE_LIBS:UTILITY_DIR/dmtcphijack.so:R_LIBSR_UTILITY_DIR/
+  dmtcp::string preloadLibs = "";
+  // FIXME:  If the colon-separated elements of ENV_VAR_MODULE are not
+  //     absolute pathnames, then they must be expanded to absolute pathnames.
+  //     Warn user if an absolute pathname is not valid.
+  if ( getenv(ENV_VAR_MODULE) != NULL ) {
+    preloadLibs += getenv(ENV_VAR_MODULE);
+    preloadLibs += ":";
+  }
+  // FindHelperUtiltiy requires ENV_VAR_UTILITY_DIR to be set
+  dmtcp::string searchDir = jalib::Filesystem::GetProgramDir();
+  setenv ( ENV_VAR_UTILITY_DIR, searchDir.c_str(), 0 );
+
+#ifdef PTRACE
+  preloadLibs += jalib::Filesystem::FindHelperUtility ( "ptracehijack.so" );
+  preloadLibs += ":";
+#endif
+
+  preloadLibs += jalib::Filesystem::FindHelperUtility ( "dmtcphijack.so" );
+  // If dmtcp_checkpoint was called with user LD_PRELOAD, and if
+  //   if dmtcp_checkpoint survived the experience, then pass it back to user.
+  if (getenv("LD_PRELOAD"))
+    preloadLibs = preloadLibs + ":" + getenv("LD_PRELOAD");
+  setenv ( "LD_PRELOAD", preloadLibs.c_str(), 1 );
+  JTRACE("getting value of LD_PRELOAD")(getenv("LD_PRELOAD"));
+  setenv ( ENV_VAR_HIJACK_LIB, preloadLibs.c_str(), 0 );
 
   //run the user program
   char **newArgv = NULL;
@@ -500,7 +475,7 @@ int testMatlab(const char *filename) {
 	    " executing.)\n\n" ;
 
   // FIXME:  should expand filename and "matlab" before checking
-  if ( strcmp(filename, "matlab") == 0 ) {
+  if ( strcmp(filename, "matlab") == 0 && getenv(ENV_VAR_QUIET) == NULL) {
     JASSERT_STDERR << theMatlabWarning;
     return -1;
   }
@@ -509,9 +484,43 @@ int testMatlab(const char *filename) {
   return 0;
 }
 
+// FIXME:  Remove this when DMTCP supports zero-mapped pages
+int testJava(char **argv) {
+  static const char* theJavaWarning =
+    "\n**** WARNING:  Sun/Oracle Java claims a large amount of memory\n"
+    "****  for its heap on startup.  As of DMTCP version 1.2.4, DMTCP _does_\n"
+    "****  handle zero-mapped virtual memory, but it may take up to a\n"
+    "****  minute.  This will be fixed to be much faster in a future\n"
+    "****  version of DMTCP.  In the meantime, if your Java supports it,\n"
+    "****  use the -Xmx flag for a smaller heap:  e.g.  java -Xmx64M javaApp\n"
+    "****  (Invoke dmtcp_checkpoint with --quiet to avoid this msg.)\n\n" ;
+
+  if (getenv(ENV_VAR_QUIET) != NULL
+      && strcmp(getenv(ENV_VAR_QUIET), "0") != 0)
+    return 0;
+  if ( strcmp(argv[0], "java") == 0 ) {
+    while (*(++argv) != NULL)
+      if (strncmp(*argv, "-Xmx", sizeof("-Xmx")-1) == 0)
+        return 0; // The user called java with -Xmx.  No need for warning.
+  }
+
+  // If user has more than 4 GB of RAM, warn them that -Xmx is faster.
+  int fd;
+  char buf[100];
+  static const char *meminfoPrefix = "MemTotal:       ";
+  if ( (fd = open("/proc/meminfo", O_RDONLY)) != -1 &&
+    read(fd, buf, sizeof(meminfoPrefix) + 16) == sizeof(meminfoPrefix) + 16 &&
+    strncmp(buf, meminfoPrefix, sizeof(meminfoPrefix)+1) == 0 &&
+    atol(buf + sizeof(meminfoPrefix)) > 17000000) /* units of kB : mem > 4 GB */
+      JASSERT_STDERR << theJavaWarning;
+  if (fd != -1)
+    close(fd);
+  return -1;
+}
+
 bool testSetuid(const char *filename)
 {
-  if (Util::isSetuid(filename) &&
+  if (dmtcp::Util::isSetuid(filename) &&
       strcmp(filename, "screen") != 0 && strstr(filename, "/screen") == NULL) {
 
     static const char* theSetuidWarning =
@@ -529,10 +538,11 @@ bool testSetuid(const char *filename)
 }
 
 void testStaticallyLinked(const char *pathname) {
-  if (Util::isStaticallyLinked(pathname)) {
+  if (dmtcp::Util::isStaticallyLinked(pathname)) {
     JASSERT_STDERR <<
-      "*** WARNING:  /lib/ld-2.10.1.so --verify " << pathname << " returns\n"
-      << "***  nonzero status.  This often means that " << pathname << " is\n"
+      "*** WARNING:  /lib/ld-linux.so --verify " << pathname << " returns\n"
+      << "***  nonzero status.  (Some distros use /lib64/ld-linux-x86-64.so .)"
+      << "*** This often means that " << pathname << " is\n"
       << "*** a statically linked target.  If so, you can confirm this with\n"
       << "*** the 'file' command.\n"
       << "***  The standard DMTCP only supports dynamically"
@@ -545,109 +555,13 @@ void testStaticallyLinked(const char *pathname) {
   return;
 }
 
-void adjust_rlimit_stack() {
-#ifdef __i386__
-  // This is needed in 32-bit Ubuntu 9.10, to fix bug with test/dmtcp5.c
-  // NOTE:  Setting personality() is cleanest way to force legacy_va_layout,
-  //   but there's currently a bug on restart in the sequence:
-  //   checkpoint -> restart -> checkpoint -> restart
-# if 0
-  { unsigned long oldPersonality = personality(0xffffffffL);
-    if ( ! (oldPersonality & ADDR_COMPAT_LAYOUT) ) {
-      // Force ADDR_COMPAT_LAYOUT for libs in high mem, to avoid vdso conflict
-      personality(oldPersonality & ADDR_COMPAT_LAYOUT);
-      JTRACE( "setting ADDR_COMPAT_LAYOUT" );
-      setenv("DMTCP_ADDR_COMPAT_LAYOUT", "temporarily is set", 1);
-    }
-  }
-# else
-  { struct rlimit rlim;
-    getrlimit(RLIMIT_STACK, &rlim);
-    if (rlim.rlim_cur != RLIM_INFINITY) {
-      char buf[100];
-      sprintf(buf, "%lu", rlim.rlim_cur); // "%llu" for BSD/Mac OS
-      JTRACE( "setting rlim_cur for RLIMIT_STACK" ) ( rlim.rlim_cur );
-      setenv("DMTCP_RLIMIT_STACK", buf, 1);
-      // Force kernel's internal compat_va_layout to 0; Force libs to high mem.
-      rlim.rlim_cur = rlim.rlim_max;
-      // FIXME: if rlim.rlim_cur != RLIM_INFINITY, then we should warn the user.
-      setrlimit(RLIMIT_STACK, &rlim);
-      // After exec, process will restore DMTCP_RLIMIT_STACK in DmtcpWorker()
-    }
-  }
-# endif
-#endif
-}
-
 // Test for 'screen' program, argvPtr is an in- and out- parameter
-bool testScreen(char **argv, char ***newArgv) 
+bool testScreen(char **argv, char ***newArgv)
 {
-  if (Util::isScreen(argv[0])) {
-    setenv("SCREENDIR", Util::getScreenDir().c_str(), 1);
-    Util::patchArgvIfSetuid(argv[0], argv, newArgv);
+  if (dmtcp::Util::isScreen(argv[0])) {
+    dmtcp::Util::setScreenDir();
+    dmtcp::Util::patchArgvIfSetuid(argv[0], argv, newArgv);
     return true;
   }
   return false;
-
-#if 0
-  struct stat st;
-  // If screen has setuid or segid bits set, ...
-  dmtcp::string pathname_base = jalib::Filesystem::BaseName((*argvPtr)[0]);
-  char pathname[1024];
-  if ((*argvPtr)[0] == NULL)
-    return -1;
-  if (Util::expandPathname((*argvPtr)[0], pathname, sizeof(pathname)) != 0)
-    return -1;
-  if ( pathname_base == "screen"
-       && stat(pathname, &st) == 0
-       && (st.st_mode & S_ISUID || st.st_mode & S_ISGID) ) {
-    dmtcp::string tmpdir = dmtcp::UniquePid::getTmpDir() + "/" + "uscreens";
-    Util::safeMkdir(tmpdir.c_str(), 0700);
-    setenv("SCREENDIR", tmpdir.c_str(), 1);
-
-    static char cmdBuf[1024];
-    char ** oldArgv = *argvPtr; // Initialize oldArgv with argument passed here
-    *(char **)(cmdBuf+sizeof(cmdBuf)-sizeof(char *)) = NULL;
-    Util::expandPathname(oldArgv[0], cmdBuf, sizeof(cmdBuf));
-#define COPY_SCREEN
-#ifdef COPY_SCREEN
-    // cp /usr/bin/screen /tmp/dmtcp-USER@HOST/screen
-    char *newArgv0 = cmdBuf + strlen(cmdBuf) + 1;
-    snprintf(newArgv0, sizeof(cmdBuf)-(newArgv0-cmdBuf), "%s/%s",
-	    dmtcp::UniquePid::getTmpDir().c_str(), pathname_base.c_str());
-    unlink(newArgv0);  // Remove any stale copy, just in case it's not right.
-    char *cpCmd = newArgv0 + strlen(newArgv0) + 1;
-    snprintf(cpCmd, sizeof(cmdBuf)-(cpCmd-cmdBuf), "cp %s %s",
-	     pathname, newArgv0);
-    Util::safeSystem(cpCmd);
-    JASSERT (access(newArgv0, X_OK) == 0) (newArgv0) (JASSERT_ERRNO);
-    (*argvPtr)[0] = newArgv0;
-    return 0;
-#else
-    // Translate: screen   to: /lib/ld-linux.so /usr/bin/screen
-    // This version is more general, but has a bug on restart:
-    //    memory layout is altered on restart, and so brk() doesn't match.
-    // Switch argvPtr from ptr to input to ptr to output now.
-    *argvPtr = (char **)(cmdBuf + strlen(cmdBuf) + 1); // ... + 1 for '\0'
-    // Use /lib64 if 64-bit O/S and not 32-bit app:
-# if defined(__x86_64__) && !defined(CONFIG_M32)
-    bool isElf, is32bitElf;
-    Util::elfType(cmdBuf, &isElf, &is32bitElf);
-    if (is32bitElf)
-      (*argvPtr)[0] = (char *)"/lib/ld-linux.so.2";
-    else
-      (*argvPtr)[0] = (char *)"/lib64/ld-linux-x86-64.so.2";
-# else
-    (*argvPtr)[0] = (char *)"/lib/ld-linux.so.2";
-# endif
-    (*argvPtr)[1] = cmdBuf;
-    for (int i = 1; oldArgv[i] != NULL; i++)
-      *argvPtr[i+1] = oldArgv[i];
-    JASSERT ((char *)cmdBuf[sizeof(cmdBuf)-sizeof(char *)] == NULL)
-      (sizeof(cmdBuf)) .Text("Expanded command longer than sizeof(cmdBuf");
-    return 0;
-#endif
-  } else
-    return -1;
-#endif
 }
