@@ -24,8 +24,10 @@
 #include <stdlib.h>
 #include <string>
 #include <stdio.h>
+#include <ctype.h>
 #include  "../jalib/jassert.h"
 #include  "../jalib/jfilesystem.h"
+#include "constants.h"
 #include "connectionmanager.h"
 #include "dmtcpworker.h"
 #include "dmtcpmessagetypes.h"
@@ -40,12 +42,14 @@
 #include <errno.h>
 #include <vector>
 
+#define BINARY_NAME "dmtcp_checkpoint"
+
 // Some global definitions
 static dmtcp::UniquePid compGroup;
 static int numPeers;
 static int coordTstamp = 0;
 
-dmtcp::string dmtcpTmpDir = "/DMTCP/UnInitialized/Tmp/Dir";
+dmtcp::string dmtcpTmpDir = "/DMTCP/Uninitialized/Tmp/Dir";
 
 using namespace dmtcp;
 
@@ -56,7 +60,7 @@ void unlockPidMapFile();
 static void runMtcpRestore ( const char* path, int offset,
                              size_t argvSize, size_t envSize);
 
-namespace
+namespace dmtcp
 {
 
 #ifdef PID_VIRTUALIZATION
@@ -122,7 +126,7 @@ namespace
 		.Text ( "checkpoint file missing" );
 
       dmtcp::SerializedWorkerInfo workerInfo;
-      _offset = _conToFd.loadFromFile(_path, &workerInfo); 
+      _offset = _conToFd.loadFromFile(_path, &workerInfo);
 
       _compGroup = workerInfo.compGroup;
       _numPeers  = workerInfo.numPeers;
@@ -189,7 +193,7 @@ namespace
 
       slidingFd.closeAll();
     }
-			
+
     int find_stdin( SlidingFdTable& slidingFd )
     {
       for ( ConnectionToFds::const_iterator i = _conToFd.begin();
@@ -213,7 +217,7 @@ namespace
       runMtcpRestore ( _path.c_str(), _offset, _argvSize, _envSize );
     }
 
-    const UniquePid& pid() const { return _conToFd.pid(); }
+    const UniquePid& upid() const { return _conToFd.upid(); }
     const dmtcp::string& procname() const { return _conToFd.procname(); }
 
 #ifdef PID_VIRTUALIZATION
@@ -225,23 +229,23 @@ namespace
     void addChild(RestoreTarget *t){ _children.push_back(t); }
 
     bool isSessionLeader(){
-      JTRACE("")(_virtualPidTable.sid()) (pid().pid());
-      if( _virtualPidTable.sid() == pid().pid() )
+      JTRACE("")(_virtualPidTable.sid()) (upid().pid());
+      if( _virtualPidTable.sid() == upid().pid() )
 	return true;
       else
 	return false;
     }
 
     bool isGroupLeader(){
-      JTRACE("")(_virtualPidTable.sid()) (pid().pid());
-      if( _virtualPidTable.gid() == pid().pid() )
+      JTRACE("")(_virtualPidTable.sid()) (upid().pid());
+      if( _virtualPidTable.gid() == upid().pid() )
 	return true;
       else
 	return false;
     }
 
     bool isForegroundProcess() {
-      JTRACE("")(_virtualPidTable.sid()) (pid().pid());
+      JTRACE("")(_virtualPidTable.sid()) (upid().pid());
       if( _virtualPidTable.fgid() == _virtualPidTable.gid() )
 	return true;
       else
@@ -270,7 +274,7 @@ namespace
       return 0;
     }
 
-    // Traverse this process subtree and setup information about sessions
+    // Traverse this process subtree and set up information about sessions
     //   and their leaders for all children.
     sidMapping &setupSessions() {
       pid_t sid = _virtualPidTable.sid();
@@ -320,7 +324,7 @@ namespace
       for(; it != _children.end(); it++){
 	(*it)->printMapping();
       }
-      JTRACE("")(pid());
+      JTRACE("")(upid());
       s_iterator sit = _smap.begin();
       for(; sit != _smap.end(); sit++){
 	JTRACE("") (sit->first) (sit->second);
@@ -366,7 +370,7 @@ namespace
 	  }else if (ptr == NULL){
             JTRACE("Cannot restore controlling terminal") (ttyname(sin));
           } else {
-	    JWARNING(false) (ttyname(sin)) 
+	    JWARNING(false) (ttyname(sin))
                     .Text("Cannot restore controlling terminal");
 	  }
 	}
@@ -384,8 +388,8 @@ namespace
 
       if( gid != fgid ){
 	if( !(pid = fork()) ){ // fork subversive process
-	  // This process moves itself to current foreground group
-	  // and then changes foreground group to what we need
+	  // This process moves itself to current foreground Group
+	  // and then changes foreground Group to what we need
 	  // so it works as a spy, saboteur or wrecker :)
 	  // -- Artem
 	  JTRACE("Change current GID to foreground GID.");
@@ -395,7 +399,7 @@ namespace
             JTRACE("CANNOT Change current GID to foreground GID")
                   (getpid()) (fgid) (_virtualPidTable.fgid()) (gid) (JASSERT_ERRNO);
           } else {
-            JWARNING(false) 
+            JWARNING(false)
                      (getpid()) (fgid) (_virtualPidTable.fgid()) (gid) (JASSERT_ERRNO)
                     .Text("CANNOT Change current GID to foreground GID");
           }
@@ -425,48 +429,51 @@ namespace
     void restoreGroup( SlidingFdTable& slidingFd )
     {
       if( isGroupLeader() ){
-	// create new group where this process becomes a leader
-	JTRACE("Create new group.");
+	// create new Group where this process becomes a leader
+	JTRACE("Create new Group.");
 	setpgid(0, 0);
 	bringToForeground(slidingFd);
       }
     }
 
-    void CreateProcess(DmtcpWorker& worker, SlidingFdTable& slidingFd)
+    void CreateProcess(DmtcpCoordinatorAPI& coordinatorAPI,
+                       SlidingFdTable& slidingFd)
     {
-      dmtcp::ostringstream o;
-      o << dmtcpTmpDir << "/jassertlog." << pid();
-      JASSERT_INIT(o.str());
-
       //change UniquePid
-      UniquePid::resetOnFork(pid());
+      UniquePid::resetOnFork(upid());
+      //UniquePid::ThisProcess(true) = _conToFd.upid();
+
+      Util::initializeLogFile(procname());
+      JTRACE("Creating process during restart") (upid()) (procname());
+
+
       VirtualPidTable &vt = _virtualPidTable;
 
       JTRACE("")(_real_getpid())(_real_getppid())(_real_getsid(0));
 
-      vt.updateMapping(pid().pid(), _real_getpid());
+      vt.updateMapping(upid().pid(), _real_getpid());
       pid_t psid = vt.sid();
 
       if( !isSessionLeader() ){
 
-	// Restore group information
+	// Restore Group information
 	restoreGroup(slidingFd);
 
 	// If process is not session leader, restore it and all children.
 	t_iterator it = _children.begin();
 	for(; it != _children.end(); it++){
-	  JTRACE ( "Forking Child Process" ) ( (*it)->pid() );
+	  JTRACE ( "Forking Child Process" ) ( (*it)->upid() );
 	  pid_t cid = forkChild();
 
 	  if ( cid == 0 )
             {
-              (*it)->CreateProcess (worker, slidingFd);
+              (*it)->CreateProcess (coordinatorAPI, slidingFd);
               JASSERT ( false ) . Text ( "Unreachable" );
             }
 	  JASSERT ( cid > 0 );
 	  VirtualPidTable::iterator vit = vt.begin();
 	  for(; vit != vt.end(); vit++){
-	    if( (*it)->pid() == vit->second ){
+	    if( (*it)->upid() == vit->second ){
 	      vt.updateMapping ( vit->first, cid );
 	      break;
 	    }
@@ -480,17 +487,17 @@ namespace
 	  s_iterator sit = (*it)->getSmap().find(psid);
 	  JTRACE("Restore processes that was created before their parent called setsid()");
 	  if( sit == (*it)->getSmap().end() ){
-	    JTRACE ( "Forking Child Process" ) ( (*it)->pid() );
+	    JTRACE ( "Forking Child Process" ) ( (*it)->upid() );
 	    pid_t cid = forkChild();
 	    if ( cid == 0 )
 	      {
-		(*it)->CreateProcess (worker, slidingFd);
+		(*it)->CreateProcess (coordinatorAPI, slidingFd);
 		JASSERT ( false ) . Text ( "Unreachable" );
 	      }
 	    JASSERT ( cid > 0 );
 	    VirtualPidTable::iterator vit = _virtualPidTable.begin();
 	    for(; vit != _virtualPidTable.end(); vit++){
-	      if( (*it)->pid() == vit->second ){
+	      if( (*it)->upid() == vit->second ){
 		_virtualPidTable.updateMapping ( vit->first, cid );
 	      }
 	    }
@@ -499,24 +506,24 @@ namespace
 
 	pid_t nsid = setsid();
 	JTRACE("change SID")(nsid);
-	
-	// Restore group information
+
+	// Restore Group information
 	restoreGroup(slidingFd);
 
 	for(t_iterator it = _children.begin(); it != _children.end(); it++) {
 	  JTRACE("Restore processes that was created after their parent called setsid()");
 	  s_iterator sit = (*it)->getSmap().find(psid);
 	  if( sit != (*it)->getSmap().end() ) {
-	    JTRACE ( "Forking Child Process" ) ( (*it)->pid() );
+	    JTRACE ( "Forking Child Process" ) ( (*it)->upid() );
 	    pid_t cid = forkChild();
 	    if ( cid == 0 ){
-	      (*it)->CreateProcess (worker, slidingFd );
+	      (*it)->CreateProcess (coordinatorAPI, slidingFd );
 	      JASSERT ( false ) . Text ( "Unreachable" );
 	    }
 	    JASSERT ( cid> 0 );
 	    VirtualPidTable::iterator vit = _virtualPidTable.begin();
 	    for(; vit != _virtualPidTable.end(); vit++) {
-	      if( (*it)->pid() == vit->second ) {
+	      if( (*it)->upid() == vit->second ) {
 		_virtualPidTable.updateMapping ( vit->first, cid );
 	      }
 	    }
@@ -524,30 +531,36 @@ namespace
 	}
 
 	for(t_iterator it = _roots.begin() ; it != _roots.end(); it++) {
-	  JTRACE ( "Forking Dependent Root Process" ) ( (*it)->pid() );
+	  JTRACE ( "Forking Dependent Root Process" ) ( (*it)->upid() );
 	  pid_t cid;
 	  if( (cid = fork()) ){
 	    waitpid(cid, NULL, 0);
 	  }else{
 	    if( fork() )
 	      exit(0);
-	    (*it)->CreateProcess(worker, slidingFd );
+	    (*it)->CreateProcess(coordinatorAPI, slidingFd );
 	    JASSERT (false) . Text( "Unreachable" );
 	  }
 	}
       }
 
       JTRACE("Child and dependent root processes forked, restoring process")
-	    (pid())(getpid())(isGroupLeader());
+	    (upid())(getpid())(isGroupLeader());
       // Save PID mapping information
-      pid_t orig = pid().pid();
+      pid_t orig = upid().pid();
       pid_t curr = _real_getpid();
       dmtcp::VirtualPidTable::InsertIntoPidMapFile(orig, curr);
 
       //Reconnect to dmtcp_coordinator
       WorkerState::setCurrentState ( WorkerState::RESTARTING );
-      worker.connectToCoordinatorWithoutHandshake();
-      worker.sendCoordinatorHandshake(procname(), _compGroup);
+
+      int tmpCoordFd = dup(PROTECTED_COORD_FD);
+      JASSERT(tmpCoordFd != -1);
+      coordinatorAPI.connectToCoordinator();
+      coordinatorAPI.sendCoordinatorHandshake(procname(), _compGroup);
+      coordinatorAPI.recvCoordinatorHandshake();
+      close(tmpCoordFd);
+
       dmtcp::string serialFile = dmtcp::UniquePid::pidTableFilename();
 
       JTRACE ( "PidTableFile: ") ( serialFile ) ( dmtcp::UniquePid::ThisProcess() );
@@ -582,7 +595,7 @@ namespace
 
 	if ( childPid == 0 ) { /* child process */
 	  if ( originalPidTable.isConflictingChildPid ( getpid() ) )
-	    _exit(1);
+	    _exit(DMTCP_FAIL_RC);
 	  else
 	    return 0;
 	}
@@ -619,7 +632,6 @@ namespace
 #endif
   };
 
-
 } // end namespace
 
 // gcc-4.3.4 -Wformat=2 issues false positives for warnings unless the format
@@ -636,7 +648,7 @@ static const char* theUsage =
   "      Directory to store temporary files \n"
   "        (default: $TMDPIR/dmtcp-$USER@$HOST or /tmp/dmtcp-$USER@$HOST)\n"
   "  --join, -j:\n"
-  "      Join an existing coordinator, raise error if one already exists\n"
+  "      Join an existing coordinator, raise error if one doesn't already exist\n"
   "  --new, -n:\n"
   "      Create a new coordinator, raise error if one already exists\n"
   "  --new-coordinator:\n"
@@ -651,17 +663,13 @@ static const char* theUsage =
   "  --no-check:\n"
   "      Skip check for valid coordinator and never start one automatically\n"
   "  --quiet, -q, (or set environment variable DMTCP_QUIET = 0, 1, or 2):\n"
-  "      Skip banner and NOTE messages; if given twice, also skip WARNINGs\n\n"
-  "See http://dmtcp.sf.net/ for more information.\n"
-;
-
-static const char* theBanner =
-  "DMTCP/MTCP  Copyright (C) 2006-2010  Jason Ansel, Michael Rieker,\n"
-  "                                       Kapil Arya, and Gene Cooperman\n"
-  "This program comes with ABSOLUTELY NO WARRANTY.\n"
-  "This is free software, and you are welcome to redistribute it\n"
-  "under certain conditions; see COPYING file for details.\n"
-  "(Use flag \"-q\" to hide this message.)\n\n"
+  "      Skip banner and NOTE messages; if given twice, also skip WARNINGs\n"
+  "  --help:\n"
+  "      Print this message and exit.\n"
+  "  --version:\n"
+  "      Print version information and exit.\n"
+  "\n"
+  "See " PACKAGE_URL " for more information.\n"
 ;
 
 //shift args
@@ -681,43 +689,81 @@ void SetupSessions();
 
 #endif
 
+static void restoreSockets(dmtcp::DmtcpCoordinatorAPI& coordinatorAPI,
+                           dmtcp::ConnectionState& ckptCoord)
+{
+  JTRACE ("restoreSockets begin");
+  jalib::JSocket& restoreSocket = coordinatorAPI.openRestoreSocket();
+
+  //reconnect to our coordinator
+  coordinatorAPI.connectToCoordinator();
+  coordinatorAPI.sendCoordinatorHandshake(jalib::Filesystem::GetProgramName(),
+                                          compGroup, numPeers,
+                                          DMT_RESTART_PROCESS);
+  coordinatorAPI.recvCoordinatorHandshake(&coordTstamp);
+  JTRACE("Connected to coordinator") (coordTstamp);
+
+  jalib::JSocket& coordinatorSocket = coordinatorAPI.coordinatorSocket();
+  // finish sockets restoration
+  ckptCoord.doReconnect(coordinatorSocket, restoreSocket);
+
+  JTRACE ("sockets restored!");
+}
+
 int main ( int argc, char** argv )
 {
   bool autoStartCoordinator=true;
   bool isRestart = true;
-  int allowedModes = dmtcp::DmtcpWorker::COORD_ANY;
+  int allowedModes = dmtcp::DmtcpCoordinatorAPI::COORD_ANY;
+
+  initializeJalib();
 
   if (! getenv(ENV_VAR_QUIET))
     setenv(ENV_VAR_QUIET, "0", 0);
 
+  if (argc == 1) {
+    JASSERT_STDERR << DMTCP_VERSION_AND_COPYRIGHT_INFO;
+    JASSERT_STDERR << "(For help:  " << argv[0] << " --help)\n\n";
+    return DMTCP_FAIL_RC;
+  }
+
   //process args
   shift;
   while(true){
     dmtcp::string s = argc>0 ? argv[0] : "--help";
-    if(s=="--help" || (s=="-h" && argc==1)){
+    if(s=="--help" && argc==1){
       JASSERT_STDERR << theUsage;
-      //fprintf(stderr, theUsage, "");
-      return 1;
+      return DMTCP_FAIL_RC;
+    } else if ((s=="--version") && argc==1){
+      JASSERT_STDERR << DMTCP_VERSION_AND_COPYRIGHT_INFO;
+      return DMTCP_FAIL_RC;
     }else if(s == "--no-check"){
       autoStartCoordinator = false;
       shift;
     }else if(s == "-j" || s == "--join"){
-      allowedModes = dmtcp::DmtcpWorker::COORD_JOIN;
+      allowedModes = dmtcp::DmtcpCoordinatorAPI::COORD_JOIN;
       shift;
     }else if(s == "-n" || s == "--new"){
-      allowedModes = dmtcp::DmtcpWorker::COORD_NEW;
+      allowedModes = dmtcp::DmtcpCoordinatorAPI::COORD_NEW;
       shift;
     }else if(s == "--new-coordinator"){
-      allowedModes = dmtcp::DmtcpWorker::COORD_FORCE_NEW;
+      allowedModes = dmtcp::DmtcpCoordinatorAPI::COORD_FORCE_NEW;
       shift;
     }else if(s == "-b" || s == "--batch"){
-      allowedModes = dmtcp::DmtcpWorker::COORD_BATCH;
+      allowedModes = dmtcp::DmtcpCoordinatorAPI::COORD_BATCH;
       shift;
-    }else if(s == "-i" || s == "--interval"){
-      setenv(ENV_VAR_CKPT_INTR, argv[1], 1);
-      shift; shift;
+    }else if(s == "-i" || s == "--interval" ||
+             (s.c_str()[0] == '-' && s.c_str()[1] == 'i' &&
+              isdigit(s.c_str()[2]) ) ){
+      if (isdigit(s.c_str()[2])) { // if -i5, for example
+        setenv(ENV_VAR_CKPT_INTR, s.c_str()+2, 1);
+        shift;
+      } else { // else -i 5
+        setenv(ENV_VAR_CKPT_INTR, argv[1], 1);
+        shift; shift;
+      }
     }else if(argc>1 && (s == "-h" || s == "--host")){
-      setenv(ENV_VAR_NAME_ADDR, argv[1], 1);
+      setenv(ENV_VAR_NAME_HOST, argv[1], 1);
       shift; shift;
     }else if(argc>1 && (s == "-p" || s == "--port")){
       setenv(ENV_VAR_NAME_PORT, argv[1], 1);
@@ -734,7 +780,7 @@ int main ( int argc, char** argv )
               (s.length()>1 && s.substr(0, 1)=="-" ) ) {
       JASSERT_STDERR << "Invalid Argument\n";
       JASSERT_STDERR << theUsage;
-      return 1;
+      return DMTCP_FAIL_RC;
     }else if(argc>1 && s=="--"){
       shift;
       break;
@@ -748,16 +794,16 @@ int main ( int argc, char** argv )
 
   jassert_quiet = *getenv(ENV_VAR_QUIET) - '0';
 
+  //make sure JASSERT initializes now, rather than during restart
+  Util::initializeLogFile();
+
   if (jassert_quiet == 0)
-    JASSERT_STDERR << theBanner;
+    JASSERT_STDERR << DMTCP_BANNER;
 
   if (autoStartCoordinator)
-    dmtcp::DmtcpWorker::startCoordinatorIfNeeded(allowedModes, isRestart);
+    dmtcp::DmtcpCoordinatorAPI::startCoordinatorIfNeeded(allowedModes,
+                                                         isRestart);
 
-  //make sure JASSERT initializes now, rather than during restart
-  dmtcp::ostringstream o;
-  o << dmtcpTmpDir << "/jassertlog." << dmtcp::UniquePid(getpid());
-  JASSERT_INIT(o.str());
   JTRACE("New dmtcp_restart process; _argc_ ckpt images") (argc);
 
   bool doAbort = false;
@@ -788,7 +834,7 @@ int main ( int argc, char** argv )
       doAbort = true;
     }
     if (doAbort) {
-      exit(1);
+      exit(DMTCP_FAIL_RC);
     }
 
     JTRACE("Will restart ckpt image _argv[0]_") (argv[0]);
@@ -798,10 +844,10 @@ int main ( int argc, char** argv )
   if (targets.size() <= 0) {
     JNOTE("ERROR: No DMTCP checkpoint image(s) found. Check Usage.");
     JASSERT_STDERR << theUsage;
-    exit(1);
+    exit(DMTCP_FAIL_RC);
   }
 
-  // Check that all targets belongs to one computation group
+  // Check that all targets belongs to one computation Group
   // If not - abort
   compGroup = targets[0]._compGroup;
   numPeers = targets[0]._numPeers;
@@ -813,7 +859,7 @@ int main ( int argc, char** argv )
 	.Text("ERROR: Restored programs belongs to different computation IDs");
     }else if( numPeers != targets[i]._numPeers ){
       JASSERT(false)(numPeers)(targets[i]._numPeers)
-	.Text("ERROR: Different numpber of processes saved in checkpoint images");
+	.Text("ERROR: Different number of processes saved in checkpoint images");
     }
   }
 
@@ -834,10 +880,10 @@ int main ( int argc, char** argv )
   JTRACE ( "Allocating fds for Connections" ) (out.str());
 
   //------------------------
-  DmtcpWorker worker ( false );
   WorkerState::setCurrentState ( WorkerState::RESTARTING );
   ConnectionState ckptCoord ( conToFd );
-  worker.restoreSockets ( ckptCoord, compGroup, numPeers, coordTstamp );
+  DmtcpCoordinatorAPI coordinatorAPI;
+  restoreSockets(coordinatorAPI, ckptCoord);
 
 #ifndef PID_VIRTUALIZATION
   int i = (int)targets.size();
@@ -850,15 +896,20 @@ int main ( int argc, char** argv )
   }
   RestoreTarget& targ = targets[i];
 
-  JTRACE("forked, restoring process")(i)(targets.size())(targ.pid())(getpid());
+  JTRACE("forked, restoring process")(i)(targets.size())(targ.upid())(getpid());
 
   //change UniquePid
-  UniquePid::resetOnFork(targ.pid());
+  UniquePid::resetOnFork(targ.upid());
 
   //Reconnect to dmtcp_coordinator
   WorkerState::setCurrentState ( WorkerState::RESTARTING );
-  worker.connectToCoordinatorWithoutHandshake();
-  worker.sendCoordinatorHandshake(targ.procname(), targ._compGroup);
+
+  int tmpCoordFd = dup(PROTECTED_COORD_FD);
+  JASSERT(tmpCoordFd != -1);
+  coordinatorAPI.connectToCoordinator();
+  coordinatorAPI.sendCoordinatorHandshake(targ.procname(), targ._compGroup);
+  coordinatorAPI.recvCoordinatorHandshake();
+  close(tmpCoordFd);
 
   //restart targets[i]
   targets[i].dupAllSockets ( slidingFd );
@@ -874,7 +925,7 @@ int main ( int argc, char** argv )
   BuildProcessTree();
 
   // Process all checkpoints to find one of them that can switch
-  // needed group to foreground.
+  // needed Group to foreground.
   ProcessGroupInfo();
   // Create session meta-information in each node of the process tree.
   // Node contains info about all sessions which exists at lower levels.
@@ -906,7 +957,7 @@ int main ( int argc, char** argv )
         if( fork() )
           _exit(0);
       }
-      roots[j].t->CreateProcess(worker, slidingFd);
+      roots[j].t->CreateProcess(coordinatorAPI, slidingFd);
       JASSERT (false) . Text( "Unreachable" );
     }
     JASSERT ( cid > 0 );
@@ -940,18 +991,18 @@ int main ( int argc, char** argv )
           pgrp_index = j;
           continue;
       }else{
-        targets[j].CreateProcess(worker, slidingFd);
-        JTRACE("Need in flat-like restore for process")(targets[j].pid());
+        targets[j].CreateProcess(coordinatorAPI, slidingFd);
+        JTRACE("Need in flat-like restore for process")(targets[j].upid());
       }
     }
   }
 
   if( pgrp_index >=0 ){
-    JTRACE("Restore first Root Target")(roots[pgrp_index].t->pid());
-    roots[pgrp_index].t->CreateProcess(worker, slidingFd);
+    JTRACE("Restore first Root Target")(roots[pgrp_index].t->upid());
+    roots[pgrp_index].t->CreateProcess(coordinatorAPI, slidingFd);
   }else if (flat_index >= 0){
-    JTRACE("Restore first Flat Target")(targets[flat_index].pid());
-    targets[flat_index].CreateProcess(worker, slidingFd );
+    JTRACE("Restore first Flat Target")(targets[flat_index].upid());
+    targets[flat_index].CreateProcess(coordinatorAPI, slidingFd );
   }else{
     // FIXME: Under what conditions will this path be exercised?
     JNOTE ("unknown type of target?") (targets[flat_index]._path);
@@ -974,9 +1025,10 @@ void BuildProcessTree()
       roots.push_back(rt);
       targets[j]._used = true;
     }else if( !targets[j]._used ){
-      // We set used flag if we use target as somebodys child. If it is used - no need to check is it roor
-      // Iterate through all targets and try to find the one who has this process
-      // as child process
+      // We set used flag if we use target as somebody's child.
+      // If it is used, then there is no need to check if it is root.
+      // Iterate through all targets and try to find the one who has
+      // this process as their child process.
       JTRACE("Process is not root of process tree: try to find if it has parent");
       bool is_root = true;
       for (size_t i = 0; i < targets.size(); i++) {
@@ -985,14 +1037,14 @@ void BuildProcessTree()
         // Search inside the child list of target[j], make sure that i != j
         for (it = virtualPidTable.begin(); (i != j) && (it != virtualPidTable.end()) ; it++) {
           UniquePid& childUniquePid = it->second;
-          JTRACE("Check child")(childUniquePid)(" parent ")(targets[i].pid())("checked ")(targets[j].pid());
-          if (childUniquePid == targets[j].pid()){
+          JTRACE("Check child")(childUniquePid)(" parent ")(targets[i].upid())("checked ")(targets[j].upid());
+          if (childUniquePid == targets[j].upid()){
             is_root = false;
             break;
           }
         }
       }
-      JTRACE("Root detection:")(is_root)(targets[j].pid());
+      JTRACE("Root detection:")(is_root)(targets[j].upid());
       if( is_root ){
         RootTarget rt;
         rt.t = &targets[j];
@@ -1012,10 +1064,10 @@ void BuildProcessTree()
 
       for ( size_t i = 0; i < targets.size(); i++ )
       {
-        if ( childUniquePid == targets[i].pid() )
+        if ( childUniquePid == targets[i].upid() )
         {
           found = 1;
-          JTRACE ( "Add child to current target" ) ( targets[j].pid() ) ( childUniquePid );
+          JTRACE ( "Add child to current target" ) ( targets[j].upid() ) ( childUniquePid );
           targets[i]._used = true;
           targets[j].addChild(&targets[i]);
         }
@@ -1042,9 +1094,9 @@ void BuildProcessTree()
  *
  */
 
-class group {
+class ProcessGroup {
 public:
-  group(){
+  ProcessGroup(){
     gid = -2;
   }
   pid_t gid;
@@ -1059,8 +1111,8 @@ public:
   }
   pid_t sid;
   pid_t fgid;
-  map<pid_t,group> groups;
-  typedef map<pid_t,group>::iterator group_it;
+  map<pid_t,ProcessGroup> groups;
+  typedef map<pid_t,ProcessGroup>::iterator group_it;
   UniquePid upid;
 };
 
@@ -1083,14 +1135,14 @@ void ProcessGroupInfo()
     //pid_t fgid = virtualPidTable.fgid();
 
     /*
-    // If group ID doesn't belong to known PIDs, indicate that fact
+    // If Group ID doesn't belong to known PIDs, indicate that fact
     //   using -1 value.
     if( !virtualPidTable.pidExists(gid) ){
     JTRACE("DROP gid")(gid);
     virtualPidTable.setgid(-1);
     gid = -1;
     }
-    // If foreground group ID not belongs to known PIDs,
+    // If foreground Group ID not belongs to known PIDs,
     //   indicate that fact using -1 value.
     if( !virtualPidTable.pidExists(fgid) ){
     JTRACE("DROP fgid")(fgid);
@@ -1104,8 +1156,8 @@ void ProcessGroupInfo()
     if( s.sid == -2 ){
       s.sid = sid;
     }
-    group &g = smap[sid].groups[gid];
-    // if this is first element of group gid
+    ProcessGroup &g = smap[sid].groups[gid];
+    // if this is first element of Group gid
     if( g.gid == -2 ){
       g.gid = gid;
     }
@@ -1118,8 +1170,10 @@ void ProcessGroupInfo()
     session &s = it->second;
     session::group_it g_it = s.groups.begin();
     pid_t fgid = -2;
+    if( s.sid == -1) // skip default bash session all processes will join
+      continue;
     for(; g_it!=s.groups.end();g_it++){
-      group &g = g_it->second;
+      ProcessGroup &g = g_it->second;
       for(size_t k=0; k<g.targets.size(); k++){
         VirtualPidTable& virtualPidTable = g.targets[k]->getVirtualPidTable();
         pid_t cfgid = virtualPidTable.fgid();
@@ -1127,17 +1181,19 @@ void ProcessGroupInfo()
           fgid = cfgid;
         }else if( fgid != -1 && cfgid != -1 && fgid != cfgid ){
           printf("Error: process from same session stores different"
-              " foreground group ID: %d, %d\n", fgid, cfgid);
+              " foreground Group ID: %d, %d\n", fgid, cfgid);
           // DEBUG PRINTOUT:
           {
             session::group_it g_it1 = s.groups.begin();
             for(; g_it1!=s.groups.end();g_it1++){
-              group &g1 = g_it1->second;
+              ProcessGroup &g1 = g_it1->second;
               for(size_t m=0; m<g1.targets.size() ;m++){
                 VirtualPidTable& virtualPidTable = g1.targets[m]->getVirtualPidTable();
                 pid_t pid = virtualPidTable.pid();
+                pid_t ppid = virtualPidTable.ppid();
+                pid_t sid = virtualPidTable.sid();
                 pid_t cfgid = virtualPidTable.fgid();
-                printf("PID=%d <--> FGID = %d\n",pid,cfgid);
+                printf("PID=%d, PPID=%d, SID=%d <--> FGID = %d\n",pid,ppid,sid,cfgid);
               }
             }
           }
@@ -1148,14 +1204,14 @@ void ProcessGroupInfo()
     }
     s.fgid = fgid;
     if( s.groups.find(s.fgid) == s.groups.end() ){
-      // foreground group is missing, don't need to change foreground groop
+      // foreground Group is missing, don't need to change foreground Group
       s.fgid = -1;
     }
 
     {
       session::group_it g_it1 = s.groups.begin();
       for(; g_it1!=s.groups.end();g_it1++){
-        group &g1 = g_it1->second;
+        ProcessGroup &g1 = g_it1->second;
         for(size_t m=0; m<g1.targets.size(); m++){
           VirtualPidTable& virtualPidTable = g1.targets[m]->getVirtualPidTable();
           pid_t pid = virtualPidTable.pid();
@@ -1174,7 +1230,7 @@ void ProcessGroupInfo()
     JTRACE("Session printout:")(s.sid)(s.fgid)(s.upid.toString().c_str());
     session::group_it g_it = s.groups.begin();
     for(; g_it!=s.groups.end();g_it++){
-      group &g = g_it->second;
+      ProcessGroup &g = g_it->second;
       JTRACE("\tGroup ID: ")(g.gid);
       /*
          for(k=0; k<g.targets.size() ;k++){
@@ -1199,7 +1255,7 @@ void SetupSessions()
       pid_t sid;
       if( (sid = (roots[i].t)->checkDependence(roots[j].t)) >= 0 ){
         // it2 depends on it1
-        JTRACE("Root target j depends on Root target i")(i)(roots[i].t->pid())(j)(roots[j].t->pid());
+        JTRACE("Root target j depends on Root target i")(i)(roots[i].t->upid())(j)(roots[j].t->upid());
         (roots[i].t)->addRoot(roots[j].t, sid);
         roots[j].indep = false;
       }
@@ -1294,7 +1350,8 @@ static void openOriginalToCurrentMappingFiles()
 static void runMtcpRestore ( const char* path, int offset,
                              size_t argvSize, size_t envSize)
 {
-  static dmtcp::string mtcprestart = jalib::Filesystem::FindHelperUtility ( "mtcp_restart" );
+  static dmtcp::string mtcprestart =
+    jalib::Filesystem::FindHelperUtility ( "mtcp_restart" );
 
   // Tell mtcp_restart process to write its debugging information to
   // PROTECTED_STDERR_FD. This way we prevent it from spitting out garbage onto
@@ -1305,11 +1362,11 @@ static void runMtcpRestore ( const char* path, int offset,
 #ifdef USE_MTCP_FD_CALLING
   int fd = ConnectionToFds::openMtcpCheckpointFile(path);
   char buf[64];
-  sprintf(buf, "%d", fd);
   char buf2[64];
+
+  sprintf(buf, "%d", fd);
   // gzip_child_pid set by openMtcpCheckpointFile() above.
   sprintf(buf2, "%d", dmtcp::ConnectionToFds::gzip_child_pid);
-
   char* newArgs[] = {
     ( char* ) mtcprestart.c_str(),
     ( char* ) "--stderr-fd",
@@ -1322,12 +1379,11 @@ static void runMtcpRestore ( const char* path, int offset,
   };
   if (dmtcp::ConnectionToFds::gzip_child_pid == -1) // If no gzip compression
     newArgs[3] = NULL;
-
   JTRACE ( "launching mtcp_restart --fd" )(fd)(path);
 #else
   char buf[64];
-  sprintf(buf, "%d", offset);
 
+  sprintf(buf, "%d", offset);
   char* newArgs[] = {
     ( char* ) mtcprestart.c_str(),
     ( char* ) "--stderr-fd",
@@ -1337,43 +1393,39 @@ static void runMtcpRestore ( const char* path, int offset,
     (char*) path,
     NULL
   };
-
   JTRACE ( "launching mtcp_restart --offset" )(path)(offset);
-
 #endif
 
+  // Create the placeholder for "MTCP_OLDPERS" environment.
+  // setenv("MTCP_OLDPERS_DUMMY", "XXXXXXXXXXXXXXXX", 1);
   // FIXME: Put an explanation of the logic below.   -- Kapil
 #define ENV_PTR(x) ((char*)(getenv(x) - strlen(x) - 1))
-
   char* dummyEnviron = NULL;
-
-  char* newEnv[] = {
-    ENV_PTR("PATH"),
-    (char*) dummyEnviron,
-    NULL
-  };
+  const int dummyEnvironIndex = 0; // index in newEnv[]
+  const int pathIndex = 1; // index in newEnv[]
+  // Eventually, newEnv = {ENV_PTR("MTCP_OLDPERS"), ENV_PTR("PATH"), NULL}
+  char* newEnv[3] = {NULL, NULL, NULL};
+  // Will put ENV_PTR("MTCP_OLDPERS") here.
+  newEnv[dummyEnvironIndex] = (char*) dummyEnviron;
+  newEnv[pathIndex] = (getenv("PATH") ? ENV_PTR("PATH") : NULL);
 
   size_t newArgsSize = 0;
   for (int i = 0; newArgs[i] != 0; i++) {
     newArgsSize += strlen(newArgs[i]) + 1;
   }
-
   size_t newEnvSize = 0;
   for (int i = 0; newEnv[i] != 0; i++) {
     newEnvSize += strlen(newEnv[i]) + 1;
   }
-
   size_t originalArgvEnvSize = argvSize + envSize;
   size_t newArgvEnvSize = newArgsSize + newEnvSize + strlen(newArgs[0]);
   size_t argvSizeDiff = originalArgvEnvSize - newArgvEnvSize;
-
   dummyEnviron = (char*) malloc(argvSizeDiff);
-  memset(dummyEnviron, '0', argvSizeDiff - 1);
+  memset(dummyEnviron, '0', (argvSizeDiff >= 1 ? argvSizeDiff - 1 : 0));
   strncpy(dummyEnviron, ENV_VAR_DMTCP_DUMMY "=0", strlen(ENV_VAR_DMTCP_DUMMY "="));
   dummyEnviron[argvSizeDiff - 1] = '\0';
 
-  newEnv[1] = dummyEnviron;
-
+  newEnv[dummyEnvironIndex] = dummyEnviron;
   JTRACE("Args/Env Sizes") (newArgsSize) (newEnvSize) (argvSize) (envSize) (argvSizeDiff);
 
   execve ( newArgs[0], newArgs, newEnv );
