@@ -32,7 +32,7 @@
 #include "virtualpidtable.h"
 #include "sysvipc.h"
 #include "syscallwrappers.h"
-#include "syslogcheckpointer.h"
+#include "syslogwrappers.h"
 #include "util.h"
 #include  "../jalib/jconvert.h"
 #include  "../jalib/jassert.h"
@@ -48,121 +48,74 @@
   const static bool dbg = false;
 #endif
 
-static pid_t forkChild ( long child_host, time_t child_time )
+static bool pthread_atfork_enabled = false;
+static time_t child_time;
+static dmtcp::DmtcpCoordinatorAPI coordinatorAPI(-1);
+
+LIB_PRIVATE void pthread_atfork_prepare()
 {
-  while ( 1 ) {
-
-    pid_t child_pid = _real_fork();
-
-    if ( child_pid == -1 ) {
-      // fork() failed
-      return child_pid;
-    } else if ( child_pid == 0 ) {
-      /* child process */
-
-      JALIB_RESET_ON_FORK ();
-#ifdef DEBUG
-      dmtcp::UniquePid child = dmtcp::UniquePid ( child_host, _real_getpid(), child_time );
-      //child should get new logfile
-      dmtcp::ostringstream o;
-      o << dmtcp::UniquePid::getTmpDir() << "/jassertlog." << child.toString();
-      JASSERT_INIT (o.str());
-#endif
-
-#ifdef PID_VIRTUALIZATION
-      if ( dmtcp::VirtualPidTable::isConflictingPid ( _real_getpid() ) ) {
-        _exit(1);
-      } else {
-        return child_pid;
-      }
-#else
-      return child_pid;
-#endif
-    } else {
-      /* Parent Process */
-#ifdef PID_VIRTUALIZATION
-      if ( dmtcp::VirtualPidTable::isConflictingPid ( child_pid ) ) {
-        JTRACE( "PID Conflict, creating new child" ) (child_pid);
-        _real_waitpid ( child_pid, NULL, 0 );
-      } else {
-        return child_pid;
-      }
-#else
-      return child_pid;
-#endif
-    }
-  }
-  return -1;
+  /* FIXME: The user process might register a fork prepare handler with
+   * pthread_atfork. That handler will be called _after_ we have acquired the
+   * wrapper-exec lock in exclusive mode. This can lead to a deadlock situation
+   * if the user process decides to do some operations that require calling a
+   * dmtcp wrapper that require the wrapper-exec lock.
+   *
+   * Also, the preparation that dmtcp needs to do for fork() should be done
+   * right before the child process is created, i.e. after all the user
+   * handlers have been invoked. Fortunately, pthread_atfork prepare handlers
+   * are called in reverse order or registration (as opposed to parent and
+   * child handlers which are called in the order of registration), thus our
+   * prepare handle will be called at the very last.
+   *
+   * FIXME: PID-conflict detection poses yet another serious problem. On a
+   * pid-conflict, _real_fork() will be called more than once, resulting in
+   * multiple calls of user-defined prepare handlers. This is undesired and can
+   * cause several issues. One solution to this problem is to call the fork
+   * system call directly whenever a tid-conflict is detected, however, it
+   * might have some other side-effects.  Another possible solution would be to
+   * have pid-virtualization module, which always assigns virtual pids, to the
+   * newly created processes, and thus avoiding the pid-conflict totally.
+   */
+  return;
 }
 
-static pid_t fork_work()
+LIB_PRIVATE void pthread_atfork_parent()
 {
-  /* Little bit cheating here: child_time should be same for both parent and
-   * child, thus we compute it before forking the child. */
-  time_t child_time = time ( NULL );
-  long child_host = dmtcp::UniquePid::ThisProcess().hostid();
+  return;
+}
+
+LIB_PRIVATE void pthread_atfork_child()
+{
+  if (!pthread_atfork_enabled) {
+    return;
+  }
+  pthread_atfork_enabled = false;
+  long host = dmtcp::UniquePid::ThisProcess().hostid();
   dmtcp::UniquePid parent = dmtcp::UniquePid::ThisProcess();
+  dmtcp::UniquePid child = dmtcp::UniquePid(host, -1, child_time);
+  dmtcp::string child_name = jalib::Filesystem::GetProgramName() + "_(forked)";
+  // Reset __thread_tid on fork. This should be the first thing to do in
+  // the child process.
+  dmtcp_reset_gettid();
+  JALIB_RESET_ON_FORK();
+  _dmtcp_remutex_on_fork();
+  dmtcp::SyslogCheckpointer::resetOnFork();
+  dmtcp::ThreadSync::resetLocks();
 
-  //pid_t child_pid = _real_fork();
-  pid_t child_pid = forkChild ( child_host, child_time );
-  if (child_pid < 0) {
-    return child_pid;
-  }
-
-
-  if ( child_pid == 0 ) {
-    child_pid = _real_getpid();
-
-    dmtcp::UniquePid child = dmtcp::UniquePid ( child_host, child_pid, child_time );
-
-    JTRACE ( "fork()ed [CHILD]" ) ( child ) ( parent );
-
-    //fix the mutex
-    _dmtcp_remutex_on_fork();
-
-    //update ThisProcess()
-    dmtcp::UniquePid::resetOnFork ( child );
+  child = dmtcp::UniquePid(host, _real_getpid(), child_time);
+  dmtcp::UniquePid::resetOnFork(child);
+  dmtcp::Util::initializeLogFile(child_name);
 
 #ifdef PID_VIRTUALIZATION
-    dmtcp::VirtualPidTable::instance().resetOnFork( );
-#endif
-
-    dmtcp::SyslogCheckpointer::resetOnFork();
-
-    //rewrite socket table
-    //         dmtcp::SocketTable::instance().onForkUpdate(parent,child);
-
-    //make new connection to coordinator
-    dmtcp::DmtcpWorker::resetOnFork();
-
-    JTRACE ( "fork() done [CHILD]" ) ( child ) ( parent );
-
-    return 0;
-  } else {
-    dmtcp::UniquePid child = dmtcp::UniquePid ( child_host, child_pid, child_time );
-
-#ifdef PID_VIRTUALIZATION
-    dmtcp::VirtualPidTable::instance().insert ( child_pid, child );
-#endif
-
-    JTRACE ( "fork()ed [PARENT] done" ) ( child );;
-
-    //         _dmtcp_lock();
-
-    //rewrite socket table
-    //         dmtcp::SocketTable::instance().onForkUpdate(parent,child);
-
-    //         _dmtcp_unlock();
-
-    //         JTRACE("fork() done [PARENT]")(child);
-
-    return child_pid;
+  if (dmtcp::VirtualPidTable::isConflictingPid(_real_getpid())) {
+    _exit(DMTCP_FAIL_RC);
   }
-}
+  dmtcp::VirtualPidTable::instance().resetOnFork();
+#endif
+
+  JTRACE("fork()ed [CHILD]") (child) (parent);
+  dmtcp::DmtcpWorker::resetOnFork(coordinatorAPI.coordinatorSocket());
 
-static void prepareForFork()
-{
-  dmtcp::KernelDeviceToConnection::instance().prepareForFork();
 }
 
 extern "C" pid_t fork()
@@ -170,19 +123,63 @@ extern "C" pid_t fork()
   /* Acquire the wrapperExeution lock to prevent checkpoint to happen while
    * processing this system call.
    */
-  WRAPPER_EXECUTION_DISABLE_CKPT();
+  WRAPPER_EXECUTION_GET_EXCL_LOCK();
+  dmtcp::KernelDeviceToConnection::instance().prepareForFork();
 
-  prepareForFork();
-  int retVal = fork_work();
+  /* Little bit cheating here: child_time should be same for both parent and
+   * child, thus we compute it before forking the child. */
+  child_time = time(NULL);
+  long host = dmtcp::UniquePid::ThisProcess().hostid();
+  dmtcp::UniquePid parent = dmtcp::UniquePid::ThisProcess();
+  dmtcp::UniquePid child = dmtcp::UniquePid(host, -1, child_time);
+  dmtcp::string child_name = jalib::Filesystem::GetProgramName() + "_(forked)";
+  pid_t child_pid;
 
-  if (retVal != 0) {
-    WRAPPER_EXECUTION_ENABLE_CKPT();
+  coordinatorAPI.createNewConnectionBeforeFork(child_name);
+
+  //Enable the pthread_atfork child call
+  pthread_atfork_enabled = true;
+  while (1) {
+    child_pid = _real_fork();
+    if (child_pid == -1) { // fork() failed
+      break;
+    }
+
+    if (child_pid == 0) { /* child process */
+      /* NOTE: Any work that needs to be done for the newly created child
+       * should be put into pthread_atfork_child() function. That function is
+       * hooked to the libc:fork() and will be called right after the new
+       * process is created and before the fork() returns.
+       *
+       * pthread_atfork_child is registered by calling pthread_atfork() from
+       * within the DmtcpWorker constructor to make sure that this is the first
+       * registered handle.
+       */
+      JTRACE("fork() done [CHILD]") (child) (parent);
+    } else { /* Parent Process */
+      coordinatorAPI.closeConnection();
+      child = dmtcp::UniquePid(host, child_pid, child_time);
+#ifdef PID_VIRTUALIZATION
+      if (dmtcp::VirtualPidTable::isConflictingPid(child_pid)) {
+        JTRACE("PID Conflict, creating new child") (child_pid);
+        _real_waitpid(child_pid, NULL, 0);
+        continue;
+      }
+      dmtcp::VirtualPidTable::instance().insert(child_pid, child);
+#endif
+
+      JTRACE("fork()ed [PARENT] done") (child);;
+    }
+    break;
   }
+  pthread_atfork_enabled = false;
 
-  return retVal;
+  if (child_pid != 0) {
+    WRAPPER_EXECUTION_RELEASE_EXCL_LOCK();
+  }
+  return child_pid;
 }
 
-
 extern "C" pid_t vfork()
 {
   JTRACE ( "vfork wrapper calling fork" );
@@ -191,13 +188,25 @@ extern "C" pid_t vfork()
   return fork();
 }
 
-static void execLibProcessAndExit(const char *path)
+// Special short-lived processes from executables like /lib/libc.so.6
+//   and man setuid/setgid executables cannot be loaded with LD_PRELOAD.
+// Since they're short-lived, we execute them while holding a lock
+//   delaying checkpointing.
+static void execShortLivedProcessAndExit(const char *path, char *const argv[])
 {
   unsetenv("LD_PRELOAD"); // /lib/ld.so won't let us preload if exec'ing lib
   const unsigned int bufSize = 100000;
   char *buf = (char*)JALLOC_HELPER_MALLOC(bufSize);
   memset(buf, 0, bufSize);
-  FILE *output = popen(path, "r");
+  FILE *output;
+  if (argv[0] == NULL) {
+    output = popen(path, "r");
+  } else {
+    dmtcp::string command = path;
+    for (int i = 1; argv[i] != NULL; i++)
+      command = command + " " + argv[i];
+    output = popen(command.c_str(), "r");
+  }
   int numRead = fread(buf, 1, bufSize, output);
   numRead++, numRead--; // suppress unused-var warning
 
@@ -205,7 +214,7 @@ static void execLibProcessAndExit(const char *path)
   // FIXME:  code currently allows wrapper to proceed without lock if
   //   it was busy because of a writer.  The unlock will then fail below.
   bool __wrapperExecutionLockAcquired = true; // needed for LOCK_UNLOCK macro
-  WRAPPER_EXECUTION_ENABLE_CKPT();
+  WRAPPER_EXECUTION_RELEASE_EXCL_LOCK();
   // We  are now the new /lib/libXXX process, and it's safe for DMTCP to ckpt us.
   printf("%s", buf); // print buf, which is what /lib/libXXX would print
   JALLOC_HELPER_FREE(buf);
@@ -222,16 +231,30 @@ static void dmtcpPrepareForExec(const char *path, char *const argv[],
 
   const char * libPrefix = "/lib/lib";
   const char * lib64Prefix = "/lib64/lib";
-  if (path != NULL && Util::strStartsWith(path, libPrefix))
-    execLibProcessAndExit(path);
-  if (path != NULL && Util::strStartsWith(path, lib64Prefix))
-    execLibProcessAndExit(path);
+  if (path != NULL && dmtcp::Util::strStartsWith(path, libPrefix))
+    execShortLivedProcessAndExit(path, argv);
+  if (path != NULL && dmtcp::Util::strStartsWith(path, lib64Prefix))
+    execShortLivedProcessAndExit(path, argv);
+  // Needed for /usr/libexec/utempter/utempter and other short-lived
+  //  setuid/setgid processes.
+  // FIXME:  USE THIS FOR ALL setuid/setgid PROCESSES EXCEPT ONES THAT
+  //         WE DIRECTLY HANDLE, LIKE 'screen'.  (Need to name special routine,
+  //         execScreenProcess() ??)
+  if (path != NULL &&
+      dmtcp::Util::strEndsWith(path, "/utempter")) {
+    JTRACE("Trying to exec: utempter")(path)(argv[0])(argv[1]);
+    execShortLivedProcessAndExit(path, argv);
+  }
 
-  if (Util::isSetuid(path)) {
-    if (Util::isScreen(path)) {
-      setenv(ENV_VAR_SCREENDIR, Util::getScreenDir().c_str(), 1);
+  // FIXME:  SEE COMMENTS IN dmtcp_checkpoint.cpp, rev. 1087; AND CHANGE THIS.
+  if (dmtcp::Util::isSetuid(path)) {
+    if (dmtcp::Util::isScreen(path)) {
+      dmtcp::Util::setScreenDir();
     }
-    Util::patchArgvIfSetuid(path, argv, newArgv);
+    // THIS NEXT LINE IS DANGEROUS.  MOST setuid PROGRAMS CAN'T RUN UNPRIVILEGED
+    dmtcp::Util::patchArgvIfSetuid(path, argv, newArgv);
+    // BUG:  dmtcp::Util::patchArgvIfSetuid() DOES NOT SET newArgv WHEN COPYING
+    //   BINARY IN CODE RE-FACTORING FROM REVISION 911.
     *filename = (*newArgv)[0];
   } else {
     *filename = (char*)path;
@@ -248,45 +271,12 @@ static void dmtcpPrepareForExec(const char *path, char *const argv[],
 #endif
 
   setenv ( ENV_VAR_SERIALFILE_INITIAL, serialFile.c_str(), 1 );
-  JTRACE ( "Preparing for Exec" ) ( path ) (*filename);
+  JTRACE ( "Will exec filename instead of path" ) ( path ) (*filename);
 
-#ifdef __i386__
-  // This is needed in 32-bit Ubuntu 9.10, to fix bug with test/dmtcp5.c
-  // NOTE:  Setting personality() is cleanest way to force legacy_va_layout,
-  //   but there's currently a bug on restart in the sequence:
-  //   checkpoint -> restart -> checkpoint -> restart
-# if 0
-  { unsigned long oldPersonality = personality(0xffffffffL);
-    if ( ! (oldPersonality & ADDR_COMPAT_LAYOUT) ) {
-      // Force ADDR_COMPAT_LAYOUT for libs in high mem, to avoid vdso conflict
-      personality(oldPersonality & ADDR_COMPAT_LAYOUT);
-      JTRACE( "setting ADDR_COMPAT_LAYOUT" );
-      setenv("DMTCP_ADDR_COMPAT_LAYOUT", "temporarily is set", 1);
-    }
-  }
-# else
-  { struct rlimit rlim;
-    getrlimit(RLIMIT_STACK, &rlim);
-    if (rlim.rlim_cur != RLIM_INFINITY) {
-      char buf[100];
-      sprintf(buf, "%lu", rlim.rlim_cur); // "%llu" for BSD/Mac OS
-      JTRACE( "setting rlim_cur for RLIMIT_STACK" ) ( rlim.rlim_cur );
-      setenv("DMTCP_RLIMIT_STACK", buf, 1);
-      // Force kernel's internal compat_va_layout to 0; Force libs to high mem.
-      rlim.rlim_cur = rlim.rlim_max;
-      // FIXME: if rlim.rlim_cur != RLIM_INFINITY, then we should warn the user.
-      setrlimit(RLIMIT_STACK, &rlim);
-      // After exec, process will restore DMTCP_RLIMIT_STACK in DmtcpWorker()
-    }
-  }
-# endif
-#endif
+  dmtcp::Util::adjustRlimitStack();
+
+  dmtcp::Util::prepareDlsymWrapper();
 
-  dmtcp::string preload (dmtcp::DmtcpWorker::ld_preload_c);
-  if (getenv("LD_PRELOAD")) {
-    preload = preload + ":" + getenv("LD_PRELOAD");
-  }
-  setenv("LD_PRELOAD", preload.c_str(), 1);
   JTRACE ( "Prepared for Exec" ) ( getenv( "LD_PRELOAD" ) );
 }
 
@@ -294,25 +284,37 @@ static void dmtcpProcessFailedExec(const char *path, char *newArgv[])
 {
   int saved_errno = errno;
 
-  if (Util::isSetuid(path)) {
-    Util::freePatchedArgv(newArgv);
+  if (dmtcp::Util::isSetuid(path)) {
+    dmtcp::Util::freePatchedArgv(newArgv);
   }
 
-  const char* str = getenv("LD_PRELOAD");
-  JASSERT(str != NULL );
-  dmtcp::string preload = getenv("LD_PRELOAD");
-  JASSERT(Util::strStartsWith(preload, dmtcp::DmtcpWorker::ld_preload_c));
+  const char *preload = getenv("LD_PRELOAD");
+  if (preload != NULL &&
+      dmtcp::Util::strStartsWith(preload, dmtcp::DmtcpWorker::ld_preload_c)) {
+    setenv("LD_PRELOAD",
+           preload + strlen(dmtcp::DmtcpWorker::ld_preload_c) + 1,
+           1);
+  }
 
-  preload.erase(0, strlen(dmtcp::DmtcpWorker::ld_preload_c) + 1);
+  unsetenv(ENV_VAR_DLSYM_OFFSET);
 
-  setenv("LD_PRELOAD", preload.c_str(), 1);
   JTRACE ( "Processed failed Exec Attempt" ) (path) ( getenv( "LD_PRELOAD" ) );
   errno = saved_errno;
 }
 
+static dmtcp::string getUpdatedLdPreload(const char* currLdPreload = NULL)
+{
+  dmtcp::string preload (dmtcp::DmtcpWorker::ld_preload_c);
+  if (currLdPreload != NULL) {
+    preload = preload + ":" + currLdPreload;
+  } else if (getenv("LD_PRELOAD") != NULL) {
+    preload = preload + ":" + getenv("LD_PRELOAD");
+  }
+  return preload;
+}
+
 static const char* ourImportantEnvs[] =
 {
-  "LD_PRELOAD",
   ENV_VARS_ALL //expands to a long list
 };
 #define ourImportantEnvsCnt ((sizeof(ourImportantEnvs))/(sizeof(const char*)))
@@ -328,103 +330,123 @@ static bool isImportantEnv ( dmtcp::string str )
   return false;
 }
 
-static dmtcp::list<dmtcp::string>& copyUserEnv ( char *const envp[] )
+static dmtcp::vector<dmtcp::string> copyUserEnv ( char *const envp[] )
 {
-  static dmtcp::list<dmtcp::string> strStorage;
-  strStorage.clear();
+  dmtcp::vector<dmtcp::string> strStorage;
 
   dmtcp::ostringstream out;
   out << "non-DMTCP env vars:\n";
   for ( ; *envp != NULL; ++envp ) {
     if ( isImportantEnv ( *envp ) ) {
-      if(dbg) 
+      if (dbg) {
         out << "     skipping: " << *envp << '\n';
+      }
       continue;
     }
-    strStorage.push_back ( *envp );
-    if(dbg) 
+    dmtcp::string e(*envp);
+    strStorage.push_back (e);
+    if(dbg) {
       out << "     addenv[user]:" << strStorage.back() << '\n';
+    }
   }
   JTRACE ( "Creating a copy of (non-DMTCP) user env vars..." ) (out.str());
 
   return strStorage;
 }
 
-static char** patchUserEnv ( dmtcp::list<dmtcp::string> &envList )
+static dmtcp::vector<const char*> patchUserEnv (dmtcp::vector<dmtcp::string>
+                                                  &envp)
 {
-  static dmtcp::vector<char*> envVect;
+  dmtcp::vector<const char*> envVect;
+  const char *userPreloadStr = NULL;
   envVect.clear();
-  
-  dmtcp::list<dmtcp::string>::iterator i;
-  for ( i = envList.begin() ; i != envList.end(); ++i ) {
-    JASSERT ( !isImportantEnv ( *i ) );
-    JASSERT ( !dbg || &(*i)[0] == (*i).c_str());
-    envVect.push_back ( (char*)i->c_str() );
+  JASSERT(envVect.size() == 0);
+
+  dmtcp::ostringstream out;
+  out << "non-DMTCP env vars:\n";
+
+  for ( size_t i = 0; i < envp.size(); i++) {
+    if ( isImportantEnv ( envp[i].c_str() ) ) {
+      if (dbg) {
+        out << "     skipping: " << envp[i] << '\n';
+      }
+      continue;
+    }
+    if (dmtcp::Util::strStartsWith(envp[i], "LD_PRELOAD=")) {
+      userPreloadStr = envp[i].c_str() + strlen("LD_PRELOAD=");
+      continue;
+    }
+
+    envVect.push_back (envp[i].c_str());
+    if(dbg) {
+      out << "     addenv[user]:" << envVect.back() << '\n';
+    }
   }
+  JTRACE ( "Creating a copy of (non-DMTCP) user env vars..." ) (out.str());
 
   //pack up our ENV into the new ENV
-  dmtcp::ostringstream out;
-  out << "DMTCP env vars:\n";
+  out.str("DMTCP env vars:\n");
   for ( size_t i=0; i<ourImportantEnvsCnt; ++i ) {
     const char* v = getenv ( ourImportantEnvs[i] );
     if ( v != NULL ) {
-      envList.push_back ( dmtcp::string ( ourImportantEnvs[i] ) + '=' + v );
-      envVect.push_back ( &envList.back() [0] );
-      if(dbg) 
-        out << "     addenv[dmtcp]:" << envList.back() << '\n';
+      envp.push_back ( dmtcp::string ( ourImportantEnvs[i] ) + '=' + v );
+      const char *ptr = envp.back().c_str();
+      JASSERT(ptr != NULL);
+      envVect.push_back(ptr);
+      if(dbg) {
+        out << "     addenv[dmtcp]:" << envVect.back() << '\n';
+      }
     }
   }
-  JTRACE ( "patching user envp..." ) ( getenv ( "LD_PRELOAD" ) ) (out.str());
 
+  dmtcp::string ldPreloadStr = "LD_PRELOAD=";
+  ldPreloadStr += getUpdatedLdPreload(userPreloadStr);
+
+  envp.push_back(ldPreloadStr);
+  envVect.push_back(envp.back().c_str());
+  if(dbg) {
+    out << "     addenv[dmtcp]:" << envVect.back() << '\n';
+  }
+
+  JTRACE ( "patching user envp..." )  (out.str());
 
   envVect.push_back ( NULL );
 
-  return &envVect[0];
+  JTRACE ( "Done patching environ" );
+  return envVect;
 }
 
-extern "C" int execve ( const char *filename, char *const argv[], char *const envp[] )
+extern "C" int execve ( const char *filename, char *const argv[],
+                        char *const envp[] )
 {
   JTRACE ( "execve() wrapper" ) ( filename );
 
   /* Acquire the wrapperExeution lock to prevent checkpoint to happen while
    * processing this system call.
    */
-  WRAPPER_EXECUTION_DISABLE_CKPT();
+  WRAPPER_EXECUTION_GET_EXCL_LOCK();
 
-  dmtcp::list<dmtcp::string> origUserEnv = copyUserEnv( envp );
+  dmtcp::vector<dmtcp::string> origUserEnv = copyUserEnv( envp );
 
   char *newFilename;
   char **newArgv;
   dmtcpPrepareForExec(filename, argv, &newFilename, &newArgv);
 
-  int retVal = _real_execve ( newFilename, newArgv, patchUserEnv ( origUserEnv ) );
+  dmtcp::vector<const char*> envVect = patchUserEnv(origUserEnv);
+
+  int retVal = _real_execve ( newFilename, newArgv, (char* const*)&envVect[0]);
 
   dmtcpProcessFailedExec(filename, newArgv);
 
-  WRAPPER_EXECUTION_ENABLE_CKPT();
+  WRAPPER_EXECUTION_RELEASE_EXCL_LOCK();
 
   return retVal;
 }
 
 extern "C" int execv ( const char *path, char *const argv[] )
 {
-  JTRACE ( "execv() wrapper" ) ( path );
-  /* Acquire the wrapperExeution lock to prevent checkpoint to happen while
-   * processing this system call.
-   */
-  WRAPPER_EXECUTION_DISABLE_CKPT();
-
-  char *newFilename;
-  char **newArgv;
-  dmtcpPrepareForExec(path, argv, &newFilename, &newArgv);
-
-  int retVal = _real_execv ( newFilename, newArgv );
-
-  dmtcpProcessFailedExec(path, newArgv);
-
-  WRAPPER_EXECUTION_ENABLE_CKPT();
-
-  return retVal;
+  JTRACE ( "execv() wrapper, calling execve with environ" ) ( path );
+  return execve(path, argv, environ);
 }
 
 extern "C" int execvp ( const char *filename, char *const argv[] )
@@ -433,17 +455,45 @@ extern "C" int execvp ( const char *filename, char *const argv[] )
   /* Acquire the wrapperExeution lock to prevent checkpoint to happen while
    * processing this system call.
    */
-  WRAPPER_EXECUTION_DISABLE_CKPT();
+  WRAPPER_EXECUTION_GET_EXCL_LOCK();
 
   char *newFilename;
   char **newArgv;
   dmtcpPrepareForExec(filename, argv, &newFilename, &newArgv);
+  setenv("LD_PRELOAD", getUpdatedLdPreload().c_str(), 1);
 
   int retVal = _real_execvp ( newFilename, newArgv );
 
   dmtcpProcessFailedExec(filename, newArgv);
 
-  WRAPPER_EXECUTION_ENABLE_CKPT();
+  WRAPPER_EXECUTION_RELEASE_EXCL_LOCK();
+
+  return retVal;
+}
+
+// This function first appeared in glibc 2.11
+extern "C" int execvpe ( const char *filename, char *const argv[],
+                         char *const envp[] )
+{
+  JTRACE ( "execvpe() wrapper" ) ( filename );
+  /* Acquire the wrapperExeution lock to prevent checkpoint to happen while
+   * processing this system call.
+   */
+  WRAPPER_EXECUTION_GET_EXCL_LOCK();
+
+  dmtcp::vector<dmtcp::string> origUserEnv = copyUserEnv( envp );
+
+  char *newFilename;
+  char **newArgv;
+  dmtcpPrepareForExec(filename, argv, &newFilename, &newArgv);
+
+  dmtcp::vector<const char*> envVect = patchUserEnv(origUserEnv);
+
+  int retVal = _real_execvpe(newFilename, newArgv, (char* const*)&envVect[0]);
+
+  dmtcpProcessFailedExec(filename, newArgv);
+
+  WRAPPER_EXECUTION_RELEASE_EXCL_LOCK();
 
   return retVal;
 }
@@ -455,26 +505,6 @@ extern "C" int fexecve ( int fd, char *const argv[], char *const envp[] )
 
   JTRACE ("fexecve() wrapper calling execve()") (fd) (buf);
   return execve(buf, argv, envp);
-
-#if 0
-  /* Acquire the wrapperExeution lock to prevent checkpoint to happen while
-   * processing this system call.
-   */
-  WRAPPER_EXECUTION_DISABLE_CKPT();
-
-  dmtcp::list<dmtcp::string> origUserEnv = copyUserEnv( envp );
-  // FIXME:  fexecve() could have fd bound to /lib/libXXX, requiring special
-  //    handling.  Because arg is NULL, we won't check for it.
-  dmtcpPrepareForExec(NULL);
-
-  int retVal = _real_fexecve ( fd, argv, patchUserEnv ( origUserEnv ) );
-
-  dmtcpProcessFailedExec(argv[0]);
-
-  WRAPPER_EXECUTION_ENABLE_CKPT();
-
-  return retVal;
-#endif
 }
 
 
