@@ -32,10 +32,12 @@
 #include <sys/socket.h>
 #include <unistd.h>
 #include <sys/syscall.h>
+#include <sys/epoll.h>
 #include <linux/version.h>
 #include <limits.h>
 #include "uniquepid.h"
 #include "dmtcpworker.h"
+#include "threadsync.h"
 #include "dmtcpmessagetypes.h"
 #include "protectedfds.h"
 #include "constants.h"
@@ -57,36 +59,6 @@ extern "C" void exit ( int status )
   for (;;); // Without this, gcc emits warning:  `noreturn' fnc does return
 }
 
-/* epoll is currently not supported by DMTCP */
-extern "C" int epoll_create(int size)
-{
-  JWARNING (false) .Text("epoll is currently not supported by DMTCP.");
-  errno = EPERM;
-  return -1;
-}
-
-extern "C" int epoll_create1(int flags)
-{
-  JWARNING (false) .Text("epoll is currently not supported by DMTCP.");
-  errno = EPERM;
-  return -1;
-}
-
-/* inotify is currently not supported by DMTCP */
-extern "C" int inotify_init()
-{
-  JWARNING (false) .Text("inotify is currently not supported by DMTCP.");
-  errno = EMFILE;
-  return -1;
-}
-
-/* inotify1 is currently not supported by DMTCP */
-extern "C" int inotify_init1(int flags)
-{
-  JWARNING (false) .Text("inotify is currently not supported by DMTCP.");
-  errno = EMFILE;
-  return -1;
-}
 
 extern "C" int socketpair ( int d, int type, int protocol, int sv[2] )
 {
@@ -130,6 +102,52 @@ extern "C" int pipe2 ( int fds[2], int flags )
 }
 #endif
 
+/* Reason for using thread_performing_dlopen_dlsym:
+ *
+ * dlsym/dlopen/dlclose make a call to calloc() internally. We do not want to
+ * checkpoint while we are in the midst of dlopen etc. as it can lead to
+ * undesired behavior. To do so, we use WRAPPER_EXECUTION_DISABLE_CKPT() at the
+ * beginning of the funtion. However, if a checkpoint request is received right
+ * after WRAPPER_EXECUTION_DISABLE_CKPT(), the ckpt-thread is queued for wrlock
+ * on the pthread-rwlock and any subsequent request for rdlock by other threads
+ * will have to wait until the ckpt-thread releases the lock. However, in this
+ * scenario, dlopen calls calloc, which then calls
+ * WRAPPER_EXECUTION_DISABLE_CKPT() and hence resulting in a deadlock.
+ *
+ * We set this variable to true, once we are inside the dlopen/dlsym/dlerror
+ * wrapper, so that the calling thread won't try to acquire the lock later on.
+ */
+
+// TODO: Integrate NEXT_FNC() with remaining wrappers in future.
+#include <dlfcn.h>
+#define NEXT_FNC(symbol) \
+  (next_fnc ? *next_fnc : \
+   *(next_fnc = (__typeof__(next_fnc))dlsym(RTLD_NEXT, #symbol)))
+extern "C"
+void *dlopen(const char *filename, int flag)
+{
+  void *ret;
+  WRAPPER_EXECUTION_DISABLE_CKPT();
+  dmtcp::ThreadSync::setThreadPerformingDlopenDlsym();
+  ret = _real_dlopen(filename, flag);
+  //ret = NEXT_FNC(dlopen)(filename, flag);
+  dmtcp::ThreadSync::unsetThreadPerformingDlopenDlsym();
+  WRAPPER_EXECUTION_ENABLE_CKPT();
+  return ret;
+}
+extern "C"
+int dlclose(void *handle)
+{
+  int ret;
+  WRAPPER_EXECUTION_DISABLE_CKPT();
+  dmtcp::ThreadSync::setThreadPerformingDlopenDlsym();
+  ret = _real_dlclose(handle);
+  //ret = NEXT_FNC(dlclose)(handle);
+  dmtcp::ThreadSync::unsetThreadPerformingDlopenDlsym();
+  WRAPPER_EXECUTION_ENABLE_CKPT();
+  return ret;
+}
+
 #ifdef PID_VIRTUALIZATION
 extern "C"
 int shmget(key_t key, size_t size, int shmflg)
@@ -138,7 +156,7 @@ int shmget(key_t key, size_t size, int shmflg)
   WRAPPER_EXECUTION_DISABLE_CKPT();
   while (true) {
     ret = _real_shmget(key, size, shmflg);
-    if (ret != -1 && 
+    if (ret != -1 &&
         dmtcp::SysVIPC::instance().isConflictingShmid(ret) == false) {
       dmtcp::SysVIPC::instance().on_shmget(key, size, shmflg, ret);
       break;
@@ -185,7 +203,7 @@ int shmctl(int shmid, int cmd, struct shmid_ds *buf)
   int currentShmid = dmtcp::SysVIPC::instance().originalToCurrentShmid(shmid);
   JASSERT(currentShmid != -1);
   int ret = _real_shmctl(currentShmid, cmd, buf);
-  // Change the creater-pid of the shm object to the original so that if
+  // Change the creator-pid of the shm object to the original so that if
   // calling thread wants to use it, pid-virtualization layer can take care of
   // the original to current conversion.
   // TODO: Need to update uid/gid fields to support uid/gid virtualization.
@@ -198,9 +216,6 @@ int shmctl(int shmid, int cmd, struct shmid_ds *buf)
 #endif // PID_VIRTUALIZATION
 
 extern "C" int __clone ( int ( *fn ) ( void *arg ), void *child_stack, int flags, void *arg, int *parent_tidptr, struct user_desc *newtls, int *child_tidptr );
-pid_t gettid();
-int tkill(int tid, int sig);
-int tgkill(int tgid, int tid, int sig);
 
 #define SYSCALL_VA_START()                                              \
   va_list ap;                                                           \
@@ -485,6 +500,7 @@ extern "C" long int syscall(long int sys_num, ... )
       break;
     }
 
+#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,9))
     case SYS_waitid:
     {
       //SYSCALL_GET_ARGS_4(idtype_t,idtype,id_t,id,siginfo_t*,infop,int,options);
@@ -492,6 +508,7 @@ extern "C" long int syscall(long int sys_num, ... )
       ret = waitid((idtype_t)idtype, id, infop, options);
       break;
     }
+#endif
     case SYS_wait4:
     {
       SYSCALL_GET_ARGS_4(pid_t,pid,__WAIT_STATUS,status,int,options,
@@ -591,6 +608,57 @@ extern "C" long int syscall(long int sys_num, ... )
   return ret;
 }
 
+// TODO: Move these to a separate pthreadwrappers.cpp file.
+extern "C" int
+pthread_cond_broadcast(pthread_cond_t *cond)
+{
+  return _real_pthread_cond_broadcast(cond);
+}
+
+extern "C" int
+pthread_cond_destroy(pthread_cond_t *cond)
+{
+  return _real_pthread_cond_destroy(cond);
+}
+
+extern "C" int
+pthread_cond_init(pthread_cond_t *cond, const pthread_condattr_t *attr)
+{
+  return _real_pthread_cond_init(cond, attr);
+}
+
+extern "C" int
+pthread_cond_signal(pthread_cond_t *cond)
+{
+  return _real_pthread_cond_signal(cond);
+}
+
+extern "C" int
+pthread_cond_timedwait(pthread_cond_t *cond, pthread_mutex_t *mutex,
+                       const struct timespec *abstime)
+{
+  return _real_pthread_cond_timedwait(cond, mutex, abstime);
+}
+
+extern "C" int
+pthread_cond_wait(pthread_cond_t *cond, pthread_mutex_t *mutex)
+{
+  return _real_pthread_cond_wait(cond, mutex);
+}
+
+
+extern "C" int
+pthread_mutex_lock(pthread_mutex_t *mutex)
+{
+  return _real_pthread_mutex_lock(mutex);
+}
+
+extern "C" int
+pthread_mutex_unlock(pthread_mutex_t *mutex)
+{
+  return _real_pthread_mutex_unlock(mutex);
+}
+
 /*
 extern "C" int
 printf (const char *format, ...)
