@@ -32,7 +32,6 @@
 #include "protectedfds.h"
 #include "connectionmanager.h"
 #include "connectionidentifier.h"
-#include "syslogcheckpointer.h"
 #include  "../jalib/jconvert.h"
 #include "constants.h"
 #include <sys/ioctl.h>
@@ -40,9 +39,6 @@
 #include <thread_db.h>
 #include <sys/procfs.h>
 
-// FIXME:  We need a better way to get MTCP_DEFAULT_SIGNAL
-#include "../../mtcp/mtcp.h" //for MTCP_DEFAULT_SIGNAL
-
 #ifdef PID_VIRTUALIZATION
 
 static pid_t originalToCurrentPid( pid_t originalPid )
@@ -51,7 +47,7 @@ static pid_t originalToCurrentPid( pid_t originalPid )
      the JASSERT log lock. Therefore, don't call JTRACE/JASSERT/JINFO/etc. in
      this function. */
   pid_t currentPid = dmtcp::VirtualPidTable::instance().originalToCurrentPid( originalPid );
-  
+
   if (currentPid == -1)
     currentPid = originalPid;
 
@@ -64,33 +60,29 @@ static pid_t currentToOriginalPid( pid_t currentPid )
      the JASSERT log lock. Therefore, don't call JTRACE/JASSERT/JINFO/etc. in
      this function. */
   pid_t originalPid = dmtcp::VirtualPidTable::instance().currentToOriginalPid( currentPid );
-  
+
   if (originalPid == -1)
     originalPid = currentPid;
 
   return originalPid;
 }
 
-pid_t gettid()
+static __thread pid_t dmtcp_thread_tid = -1;
+LIB_PRIVATE
+void dmtcp_reset_gettid() {
+  dmtcp_thread_tid = -1;
+}
+extern "C" pid_t gettid()
 {
-  WRAPPER_EXECUTION_DISABLE_CKPT();
-  /*
-   * We might want to cache the tid of all threads to avoid redundant calls
-   *  to _real_gettid() and currentToOriginalPid().
-   * To cache, we must make sure that this function is invoked by each thread
-   *  at least once prior to checkpoint.
-   * __thread can be used along with static storage class to make this cached
-   *  value specific to each thread
+  /* mtcpinterface.cpp:thread_start calls gettid() before calling
+   * DmtcpWorker::decrementUninitializedThreadCount() and so the value is
+   * cached before it is accessed by some other DMTCP code.
    */
-  pid_t currentTid = _real_gettid();
-  pid_t origTid =  currentToOriginalPid ( currentTid );
-
-  WRAPPER_EXECUTION_ENABLE_CKPT();
-
-  return origTid;
+  if (dmtcp_thread_tid == -1) {
+    dmtcp_thread_tid = _real_gettid();
+  }
+  return dmtcp_thread_tid;
 }
-
-
 
 extern "C" pid_t getpid()
 {
@@ -230,7 +222,7 @@ extern "C" pid_t setsid(void)
 extern "C" int   kill(pid_t pid, int sig)
 {
   /* FIXME: When bash receives a SIGINT signal, the signal handler is
-   * called to processs the signal. Once the processing is done, bash
+   * called to process the signal. Once the processing is done, bash
    * performs a longjmp to a much higher call frame. As a result, this
    * call frame never gets a chance to return and hence we fail to
    * perform WRAPPER_EXECUTION_ENABLE_CKPT() which results in the lock
@@ -262,7 +254,7 @@ extern "C" int   kill(pid_t pid, int sig)
    *       <RESTORE_SIGNAL_MASK>
    *     }
    *     return retVal;
-   * 
+   *
    *
    * This longjmp trouble can happen with any wrapper, whose execution my
    * end up in a call to user-code i.e. if the call frame looks sth like:
@@ -273,7 +265,7 @@ extern "C" int   kill(pid_t pid, int sig)
    *        ...
    *        user_func2(...)
    *        ...
-   * 
+   *
    * Another potential way would be to put a wrapper around longjmp() in
    * which the calling thread should release all the DMTCP-locks being
    * held at the moment. This would require us to keep a count of lock()
@@ -281,8 +273,8 @@ extern "C" int   kill(pid_t pid, int sig)
    * call, one need to make sure that an unlock() call is requested only
    * if there is a corresponding lock, because it might happen that
    * longjmp() was harmless in the sense that, it didn't cause a
-   * callframe like the one mentioned above. 
-   * 
+   * callframe like the one mentioned above.
+   *
    */
 //  WRAPPER_EXECUTION_DISABLE_CKPT();
 
@@ -295,7 +287,8 @@ extern "C" int   kill(pid_t pid, int sig)
   return retVal;
 }
 
-int   tkill(int tid, int sig)
+LIB_PRIVATE
+int tkill(int tid, int sig)
 {
   // FIXME: Check the comments in kill()
 //  WRAPPER_EXECUTION_DISABLE_CKPT();
@@ -309,7 +302,8 @@ int   tkill(int tid, int sig)
   return retVal;
 }
 
-int   tgkill(int tgid, int tid, int sig)
+LIB_PRIVATE
+int tgkill(int tgid, int tid, int sig)
 {
   // FIXME: Check the comments in kill()
 //  WRAPPER_EXECUTION_DISABLE_CKPT();
@@ -327,17 +321,16 @@ int   tgkill(int tgid, int tid, int sig)
 
 //long sys_tgkill (int tgid, int pid, int sig)
 
-#ifdef PTRACE
-
-#define TRUE 1
-#define FALSE 0
-
-extern "C" td_err_e   _dmtcp_td_thr_get_info ( const td_thrhandle_t  *th_p,
-                                               td_thrinfo_t *ti_p)
+typedef td_err_e (*td_thr_get_info_funcptr_t)(const td_thrhandle_t *,
+                                              td_thrinfo_t *);
+static td_thr_get_info_funcptr_t _td_thr_get_info_funcptr = NULL;
+static td_err_e  _dmtcp_td_thr_get_info (const td_thrhandle_t  *th_p,
+                                         td_thrinfo_t *ti_p)
 {
   td_err_e td_err;
 
-  td_err = _real_td_thr_get_info ( th_p, ti_p);
+  td_err = (*_td_thr_get_info_funcptr)(th_p, ti_p);
+
   ti_p->ti_lid  =  ( lwpid_t ) currentToOriginalPid ( ( int ) ti_p->ti_lid );
   ti_p->ti_tid =  ( thread_t ) currentToOriginalPid ( (int ) ti_p->ti_tid );
   return td_err;
@@ -349,27 +342,19 @@ extern "C" td_err_e   _dmtcp_td_thr_get_info ( const td_thrhandle_t  *th_p,
 */
 extern "C" void *dlsym ( void *handle, const char *symbol)
 {
-  if ( strcmp ( symbol, "td_thr_get_info" ) == 0 )
-    return (void *) &_dmtcp_td_thr_get_info;
+  if ( strcmp ( symbol, "td_thr_get_info" ) == 0 ) {
+    _td_thr_get_info_funcptr = (td_thr_get_info_funcptr_t) _real_dlsym(handle,
+                                                                       symbol);
+    if (_td_thr_get_info_funcptr != NULL) {
+      return (void *) &_dmtcp_td_thr_get_info;
+    } else {
+      return NULL;
+    }
+  }
   else
     return _real_dlsym ( handle, symbol );
 }
 
-extern "C" void ptrace_info_list_update_info(pid_t superior, pid_t inferior,
-                                             int singlestep_waited_on); 
-
-typedef struct ptrace_waitpid_info ( *t_mtcp_get_ptrace_waitpid_info) ( );
-extern "C" t_mtcp_get_ptrace_waitpid_info mtcp_get_ptrace_waitpid_info;
-
-typedef int ( *fill_in_pthread_t) ();
-extern "C" fill_in_pthread_t fill_in_pthread_ptr;
-
-typedef int ( *delete_thread_on_pthread_join_t) ();
-extern "C" delete_thread_on_pthread_join_t delete_thread_on_pthread_join_ptr;
-
-extern "C" sigset_t signals_set;
-#endif
-
 /*
  * TODO: Add the wrapper protection for wait() family of system calls.
  *       It wouldn't be a straight forward process, we need to take care of the
@@ -378,155 +363,65 @@ extern "C" sigset_t signals_set;
  */
 
 extern "C" pid_t wait (__WAIT_STATUS stat_loc)
-//extern "C" pid_t wait(int *stat_loc)
 {
-  pid_t retVal = _real_wait (stat_loc);
-
-  if (retVal > 0) {
-    pid_t pid = currentToOriginalPid (retVal);
-
-    if ( pid > 0 )
-      dmtcp::VirtualPidTable::instance().erase(pid);
-
-    return pid;
-  }
-  return retVal;
+  return waitpid(-1, (int*)stat_loc, 0);
 }
 
-/* UNUSED FUNCTION
-extern "C" {
-static int get_sigckpt() {
-  static char *nptr = getenv(ENV_VAR_SIGCKPT);
-  static int sigckpt = -1;
-  if (sigckpt == -1) {
-    char *endptr;
-    if (nptr == NULL)
-      sigckpt = MTCP_DEFAULT_SIGNAL;
-    else {
-      sigckpt = strtol(nptr, &endptr, 0);
-      if (endptr != '\0')
-        sigckpt = MTCP_DEFAULT_SIGNAL;
-    }
-  }
-  return sigckpt;
-}
-}
-*/
-
-extern "C"
-pid_t safe_real_waitpid(pid_t pid, int *stat_loc, int options) {
-  // Note that if the action for SIGCHLD is set to SIG_IGN, then waitpid fails
-  //  with errno set to ECHLD (as if the child process was not really our child)
-  //  We currently do not specially handle this case.
-  while (1) {
-    /* The checkpoint signal uses SA_RESTART which causes the system call to
-     * restart if it were interrupted by this signal. However, on restart, the
-     * currPid will not be correct anymore and hence the restarted waitpid()
-     * will fail with ECHILD.
-     * TODO: Note that the currPid after restart might be equal to the
-     *       current-pid of some other child process resulting in waitpid
-     *       waiting un-necessarily for getting status of that other child
-     *       process. This could be fixed in the sa_restorer function of
-     *       sigaction in mtcp/mtcp_sigaction.c.
-     */
-    pid_t currPid = originalToCurrentPid(pid);
-    pid_t retval = _real_waitpid(currPid, stat_loc, options);
-
-    WRAPPER_EXECUTION_DISABLE_CKPT();
-
-    /* Verify there was no checkpoint/restart between currpid and _real_waitpid.
-     * Note:  A checkpoint/restart after _real_waitpid does not need a verify. 
-     * TODO: There is a possible bug -- if by conincidence the currentPid
-     *       becomes equal to curPid, this will cause the if condition to go
-     *       false, which is wrong.
-     */
-    if ( retval == -1 && errno == ECHILD &&
-         currPid != originalToCurrentPid(pid) ) {
-      struct sigaction oldact;
-      if ( sigaction(SIGCHLD, NULL, &oldact) == 0 &&
-           oldact.sa_handler != SIG_IGN ) {
-        WRAPPER_EXECUTION_ENABLE_CKPT();
-        continue;
-      }
-    }
-    WRAPPER_EXECUTION_ENABLE_CKPT();
-    return retval;
-  }
-}
-
-
 extern "C" pid_t waitpid(pid_t pid, int *stat_loc, int options)
 {
-  int status;
-  pid_t originalPid;
-  pid_t retval;
-
-  if ( stat_loc == NULL )
-    stat_loc = &status;
-
-#ifdef PTRACE
-  pid_t superior = syscall(SYS_gettid);
-  pid_t inferior = pid;
-  struct ptrace_waitpid_info pwi = mtcp_get_ptrace_waitpid_info();
-
-  if (pwi.is_waitpid_local) {
-    retval = safe_real_waitpid (pid, stat_loc, options);
-  } else {
-    /* Where was status and pid saved?  Can we remove this code?  - Gene */
-    if (pwi.has_status_and_pid) {
-      *stat_loc = pwi.saved_status;
-      retval = pwi.saved_pid;
-    } else {
-// Please remove this comment and all code related to BLOCK_CKPT_ON_WAIT
-//  when satisfied waitpid wrapper work.  - Gene
-#undef BLOCK_CKPT_ON_WAIT
-#if BLOCK_CKPT_ON_WAIT
-      if (_real_pthread_sigmask(SIG_BLOCK, &signals_set, NULL) != 0) {
-        perror ("waitpid wrapper");
-        exit(-1);
-      }
-#endif
-      ptrace_info_list_update_info(superior, inferior, TRUE);
-      retval = safe_real_waitpid(pid, stat_loc, options);
-#if BLOCK_CKPT_ON_WAIT
-      if (_real_pthread_sigmask(SIG_UNBLOCK, &signals_set, NULL) != 0) {
-        perror("waitpid wrapper");
-        exit(-1);
-      }
-#endif
-    }
-  }
-#else 
-  retval = safe_real_waitpid(pid, stat_loc, options);
-#endif
-
-  if (retval > 0) {
-    originalPid = currentToOriginalPid(retval);
-    if ( WIFEXITED(*stat_loc)  || WIFSIGNALED(*stat_loc) )
-      dmtcp::VirtualPidTable::instance().erase(originalPid);
-    return originalPid;
-  } else {
-    return retval;
-  }
+  return wait4(pid, stat_loc, options, NULL);
 }
 
-extern "C" int   waitid(idtype_t idtype, id_t id, siginfo_t *infop, int options)
+extern "C" int waitid(idtype_t idtype, id_t id, siginfo_t *infop, int options)
 {
-  siginfo_t status;
+  int retval = 0;
+  struct timespec sleepTime = {0, 1000};
+  siginfo_t siginfop;
+  memset(&siginfop, 0, sizeof(siginfop));
 
-  if ( infop == NULL )
-    infop = &status;
+  /* waitid returns 0 in case of success as well as when WNOHANG is specified
+   * and we need to distinguish those two cases.man page for waitid says:
+   *   If WNOHANG was specified in options and there were no children in a
+   *   waitable  state, then waitid() returns 0 immediately and the state of
+   *   the siginfo_t structure pointed to by infop is unspecified.  To
+   *   distinguish this case from that where a child was in a  waitable state,
+   *   zero out the si_pid field before the call and check for a nonzero value
+   *   in this field after the call returns.
+   *
+   * See comments above wait4()
+   */
+  while (retval == 0) {
+    WRAPPER_EXECUTION_DISABLE_CKPT();
+    pid_t currPid = originalToCurrentPid (id);
+    retval = _real_waitid (idtype, currPid, &siginfop, options | WNOHANG);
 
-  pid_t currPd = originalToCurrentPid (id);
+    if (retval != -1) {
+      pid_t originalPid = currentToOriginalPid ( siginfop.si_pid );
+      siginfop.si_pid = originalPid;
 
-  int retval = _real_waitid (idtype, currPd, infop, options);
+      if ( siginfop.si_code == CLD_EXITED || siginfop.si_code == CLD_KILLED )
+        dmtcp::VirtualPidTable::instance().erase ( originalPid );
+    }
+    WRAPPER_EXECUTION_ENABLE_CKPT();
 
-  if (retval != -1) {
-    pid_t originalPid = currentToOriginalPid ( infop->si_pid );
-    infop->si_pid = originalPid;
+    if ((options & WNOHANG) ||
+        retval == -1 ||
+        siginfop.si_pid != 0) {
+      break;
+    } else {
+      if (sleepTime.tv_sec == 0) {
+        sleepTime.tv_nsec *= 2;
+        if (sleepTime.tv_nsec >= 1000 * 1000 * 1000) {
+          sleepTime.tv_sec++;
+          sleepTime.tv_nsec = 0;
+        }
+      }
+      nanosleep(&sleepTime, NULL);
+    }
+  }
 
-    if ( infop->si_code == CLD_EXITED || infop->si_code == CLD_KILLED )
-      dmtcp::VirtualPidTable::instance().erase ( originalPid );
+  if (retval == 0 && infop != NULL) {
+    *infop = siginfop;
   }
 
   return retval;
@@ -534,66 +429,117 @@ extern "C" int   waitid(idtype_t idtype, id_t id, siginfo_t *infop, int options)
 
 extern "C" pid_t wait3(__WAIT_STATUS status, int options, struct rusage *rusage)
 {
-  pid_t retval = _real_wait3 (status, options, rusage);
-
-  pid_t originalPid = currentToOriginalPid ( retval );
-
-  if ( originalPid > 0 )
-    dmtcp::VirtualPidTable::instance().erase(originalPid);
-
-  return originalPid;
+  return wait4(-1, status, options, rusage);
 }
 
-extern "C" pid_t wait4(pid_t pid, __WAIT_STATUS status, int options, struct rusage *rusage)
+/*
+ * wait() family and checkpoint/restart.
+ *
+ * wait() returns the _current_ pid of the child process. The
+ * pid-virtualization layer then converts it to original pid and return it to
+ * the caller.
+ *
+ * To guarantee the correctness of the current to original conversion, we need
+ * to make sure that there is no ckpt/restart in between (A) returning from
+ * wait(), and (B) performing conversion. If a ckpt happens in between, then on
+ * restart, the current pid won't be valid anymore and the conversion would be
+ * a false one.
+ *
+ * One way to avoid ckpt/restart in between state A and B is to disable ckpt
+ * before A and enable it only after performing B. The problem in doing this is
+ * the fact that wait is a blocking system call, unless WNOHANG is specified.
+ * Thus we force the WNOHANG flag even though the caller didn't want to.
+ *
+ * When WNOHANG is specified, a return value of '0' indicates that there is no
+ * child process to be waited for, in which case we sleep for a small amount of
+ * time and retry.
+ *
+ * The last bit of logic is the amount of time to sleep for. Too little, and we
+ * end up wasting CPU time; too large, and some application (for eg: strace)
+ * might not like. We try to avoid this problem by starting with a tiny sleep
+ * interval and on every failed wait(), we double the interval until we reach a
+ * max. Once it reaches a max time, we don't double it, we just use it as is.
+ *
+ * The initial sleep interval is set to 10 micro seconds and the max is set to
+ * 1 second. We hope that it would cover all the cases.
+ *
+ */
+extern "C"
+pid_t wait4(pid_t pid, __WAIT_STATUS status, int options, struct rusage *rusage)
 {
   int stat;
+  int saved_errno = errno;
+  pid_t currPid;
+  pid_t originalPid;
+  pid_t retval = 0;
+  struct timespec sleepTime = {0, 10*000};
 
-  if ( status == NULL )
+  if (status == NULL)
     status = (__WAIT_STATUS) &stat;
 
-  pid_t currPid = originalToCurrentPid (pid);
+  while (retval == 0) {
+    WRAPPER_EXECUTION_DISABLE_CKPT();
+    currPid = originalToCurrentPid(pid);
+    retval = _real_wait4(currPid, status, options | WNOHANG, rusage);
+    saved_errno = errno;
+    originalPid = currentToOriginalPid(retval);
 
-  pid_t retval = _real_wait4 ( currPid, status, options, rusage );
-
-  pid_t originalPid = currentToOriginalPid ( retval );
-
-  if ( retval > 0
-       && ( WIFEXITED ( * (int*) status )  || WIFSIGNALED ( * (int*) status ) ) )
-    dmtcp::VirtualPidTable::instance().erase ( originalPid );
+    if (retval > 0 &&
+        (WIFEXITED(*(int*)status) || WIFSIGNALED(*(int*)status))) {
+      dmtcp::VirtualPidTable::instance().erase(originalPid);
+    }
+    WRAPPER_EXECUTION_ENABLE_CKPT();
 
+    if ((options & WNOHANG) || retval != 0) {
+      break;
+    } else {
+      if (sleepTime.tv_sec == 0) {
+        sleepTime.tv_nsec *= 2;
+        if (sleepTime.tv_nsec >= 1000 * 1000 * 1000) {
+          sleepTime.tv_sec++;
+          sleepTime.tv_nsec = 0;
+        }
+      }
+      nanosleep(&sleepTime, NULL);
+    }
+  }
+  errno = saved_errno;
   return originalPid;
 }
 
-// TODO:  ioctl must use virtualized pids for request = TIOCGPGRP / TIOCSPGRP
-// These are synonyms for POSIX standard tcgetpgrp / tcsetpgrp
-extern "C" {
-int send_sigwinch = 0;
-}
-extern "C" int ioctl(int d,  unsigned long int request, ...)
-{ va_list ap;
-  int rc;
+extern "C" long ptrace (enum __ptrace_request request, ...)
+{
+  va_list ap;
+  pid_t pid;
+  void *addr;
+  void *data;
 
-  if (send_sigwinch && request == TIOCGWINSZ) {
-    send_sigwinch = 0;
-    va_list local_ap;
-    va_copy(local_ap, ap);
-    va_start(local_ap, request);
-    struct winsize * win = va_arg(local_ap, struct winsize *);
-    va_end(local_ap);
-    rc = _real_ioctl(d, request, win);  // This fills in win
-    win->ws_col--; // Lie to application, and force it to resize window,
-		   //  reset any scroll regions, etc.
-    kill(getpid(), SIGWINCH); // Tell application to look up true winsize
-			      // and resize again.
-  } else {
-    void * arg;
-    va_start(ap, request);
-    arg = va_arg(ap, void *);
-    va_end(ap);
-    rc = _real_ioctl(d, request, arg);
+  va_start(ap, request);
+  pid = va_arg(ap, pid_t);
+  addr = va_arg(ap, void *);
+  data = va_arg(ap, void *);
+  va_end(ap);
+
+  pid = dmtcp::VirtualPidTable::instance().originalToCurrentPid(pid);
+  long ptrace_ret =  _real_ptrace(request, pid, addr, data);
+
+  /*
+   * PTRACE_GETEVENTMSG (since Linux 2.5.46)
+   *          Retrieve  a message (as an unsigned long) about the ptrace event
+   *          that just happened, placing it in the location data in the
+   *          parent.  For PTRACE_EVENT_EXIT this is the child's exit status.
+   *          For PTRACE_EVENT_FORK, PTRACE_EVENT_VFORK and PTRACE_EVENT_CLONE
+   *          this is the PID  of the new process.  Since Linux 2.6.18, the PID
+   *          of the new process is also available for PTRACE_EVENT_VFORK_DONE.
+   *          (addr is ignored.)
+   */
+
+  if (ptrace_ret == 0 && request == PTRACE_GETEVENTMSG) {
+    unsigned long *ldata = (unsigned long*) data;
+    *ldata = (unsigned long) currentToOriginalPid((pid_t) *ldata);
   }
 
-  return rc;
+  return ptrace_ret;
 }
 
 /*
