@@ -34,6 +34,10 @@
 #include <sys/wait.h>
 #include "constants.h"
 #include <sys/ptrace.h>
+// This was needed for 64-bit SUSE LINUX Enterprise Server 9 (Linux 2.6.5):
+#ifndef PTRACE_GETEVENTMSG
+# include <linux/ptrace.h>
+#endif
 #include <stdarg.h>
 #include <asm/ldt.h>
 #include <stdio.h>
@@ -43,19 +47,13 @@
 #include <sys/ipc.h>
 #include <sys/shm.h>
 #include <sys/stat.h>
-#ifdef RECORD_REPLAY
 #include <sys/mman.h>
+#include <sys/epoll.h>
 #include <dirent.h>
 #include <unistd.h>
-#endif
-
-#ifdef RECORD_REPLAY
-#define SET_MMAP_NO_SYNC()   (mmap_no_sync = 1)
-#define UNSET_MMAP_NO_SYNC() (mmap_no_sync = 0)
-#define MMAP_NO_SYNC         (mmap_no_sync == 1)
-// Defined in dmtcpworker.cpp:
-__attribute__ ((visibility ("hidden"))) extern __thread int mmap_no_sync;
-#endif
+#include <pwd.h>
+#include <grp.h>
+#include <netdb.h>
 
 void _dmtcp_setup_trampolines();
 
@@ -64,8 +62,37 @@ extern "C"
 {
 #endif
 
-#ifdef PID_VIRTUALIZATION
-# define GLIBC_PID_FAMILY_WRAPPERS(MACRO)   \
+#if __GLIBC_PREREQ(2,5)
+# define READLINK_RET_TYPE ssize_t
+#else
+# define READLINK_RET_TYPE int
+#endif
+
+/* The following function are defined in pidwrappers.cpp */
+pid_t gettid();
+int tkill(int tid, int sig);
+int tgkill(int tgid, int tid, int sig);
+
+
+
+extern int dmtcp_wrappers_initializing;
+
+LIB_PRIVATE extern __thread int thread_performing_dlopen_dlsym;
+
+#define FOREACH_GLIBC_MALLOC_FAMILY_WRAPPERS(MACRO)\
+  MACRO(calloc)                             \
+  MACRO(malloc)                             \
+  MACRO(free)                               \
+  MACRO(__libc_memalign)                    \
+  MACRO(realloc)                            \
+  MACRO(mmap)                               \
+  MACRO(mmap64)                             \
+  MACRO(mremap)                             \
+  MACRO(munmap)
+
+#define FOREACH_GLIBC_WRAPPERS(MACRO)       \
+  MACRO(dlopen)                             \
+  MACRO(dlclose)                            \
   MACRO(getpid)                             \
   MACRO(getppid)                            \
   MACRO(kill)                               \
@@ -86,151 +113,51 @@ extern "C"
   MACRO(waitpid)                            \
   MACRO(waitid)                             \
   MACRO(wait3)                              \
-  MACRO(wait4)				    \
-  MACRO(ioctl)
-#else
-# define GLIBC_PID_FAMILY_WRAPPERS(MACRO)
-#endif /* PID_VIRTUALIZATION */
-
-#ifdef ENABLE_MALLOC_WRAPPER
-# define GLIBC_MALLOC_FAMILY_WRAPPERS(MACRO)\
-  MACRO(calloc)                             \
-  MACRO(malloc)                             \
-  MACRO(free)                               \
-  MACRO(__libc_memalign)                    \
-  MACRO(realloc)
-#else
-# define GLIBC_MALLOC_FAMILY_WRAPPERS(MACRO)
-#endif 
-
-#ifdef PTRACE
-# define GLIBC_PTRACE_WRAPPERS(MACRO)       \
-  MACRO(ptrace)
-#else
-# define GLIBC_PTRACE_WRAPPERS(MACRO)
-#endif
-
-#ifdef RECORD_REPLAY
-# define GLIBC_RECORD_WRAPPERS(MACRO)\
-  MACRO(access)                               \
-  MACRO(select)                               \
-  MACRO(read)                                 \
-  MACRO(readdir)                              \
-  MACRO(readdir_r)                            \
-  MACRO(write)                                \
-  MACRO(rand)                                 \
-  MACRO(srand)                                \
-  MACRO(time)                                 \
-  MACRO(getsockname)                          \
-  MACRO(getpeername)                          \
-  MACRO(fcntl)                                \
-  MACRO(dup)                                  \
-  MACRO(lseek)                                \
-  MACRO(__fxstat)                             \
-  MACRO(__fxstat64)                           \
-  MACRO(unlink)                               \
-  MACRO(pread)                                \
-  MACRO(pwrite)                               \
-  MACRO(sigset)                               \
-  MACRO(fdopen)                               \
-  MACRO(fgets)                                \
-  MACRO(putc)                                 \
-  MACRO(fputs)                                \
-  MACRO(fdatasync)                            \
-  MACRO(fsync)                                \
-  MACRO(link)                                 \
-  MACRO(getc)                                 \
-  MACRO(fgetc)                                \
-  MACRO(ungetc)                               \
-  MACRO(getline)                              \
-  MACRO(rename)                               \
-  MACRO(rewind)                               \
-  MACRO(rmdir)                                \
-  MACRO(ftell)                                \
-  MACRO(fwrite)                               \
-  MACRO(mkdir)                                \
-  MACRO(mkstemp)                              \
-  MACRO(mmap)                                 \
-  MACRO(mmap64)                               \
-  MACRO(mremap)                               \
-  MACRO(munmap)
-
-#define FOREACH_PTHREAD_FUNC_WRAPPER(MACRO)     \
-  MACRO(pthread_cond_wait)			\
-  MACRO(pthread_cond_timedwait)			\
-  MACRO(pthread_cond_signal)			\
-  MACRO(pthread_cond_broadcast)			\
-  MACRO(pthread_create)                         \
-  MACRO(pthread_detach)                         \
-  MACRO(pthread_exit)                           \
-  MACRO(pthread_join)                           \
-  MACRO(pthread_kill)                           \
-  MACRO(pthread_sigmask)                        \
-  MACRO(pthread_mutex_lock)			\
-  MACRO(pthread_mutex_trylock)			\
-  MACRO(pthread_mutex_unlock)			\
-  MACRO(pthread_rwlock_unlock)                  \
-  MACRO(pthread_rwlock_rdlock)                  \
-  MACRO(pthread_rwlock_wrlock)
-#else
-# define GLIBC_RECORD_WRAPPERS(MACRO)
-# define PTHREAD_RECORD_WRAPPERS(MACRO)
-#endif 
-
-/* First group below is candidates for glibc_base_func_addr in syscallsreal.c
- * We can't tell which ones were already re-defined by the user executable.
- * For example, /bin/dash defines isalnum in Ubuntu 9.10.
- * Note that a system call can't be a base fnc if we are already wrapping it.
- */
-#define FOREACH_GLIBC_BASE_FUNC(MACRO)      \
-  MACRO(isalnum)			    \
-  MACRO(clearerr)			    \
-  MACRO(getopt)				    \
-  MACRO(perror)				    \
-  MACRO(fscanf)
-
-#define GLIBC_SOCKET_WRAPPERS(MACRO)        \
+  MACRO(wait4)                              \
+  MACRO(ioctl)                              \
+  MACRO(ptrace)                             \
+                                            \
   MACRO(socket)                             \
   MACRO(connect)                            \
   MACRO(bind)                               \
   MACRO(listen)                             \
   MACRO(accept)                             \
-  MACRO(accept4)                             \
+  MACRO(accept4)                            \
   MACRO(setsockopt)                         \
-  MACRO(socketpair)
-
-//#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,28)) && __GLIBC_PREREQ(2,10)
-//# define GLIBC_ACCEPT4_WRAPPER(MACRO)      \
-//   MACRO(accept4)
-//#else
-//# define GLIBC_ACCEPT4_WRAPPER(MACRO)
-//#endif
-
-#define GLIBC_EXEC_WRAPPERS(MACRO)          \
+  MACRO(getsockopt)                         \
+  MACRO(socketpair)                         \
+                                            \
   MACRO(fexecve)                            \
   MACRO(execve)                             \
   MACRO(execv)                              \
   MACRO(execvp)                             \
+  MACRO(execvpe)                            \
   MACRO(execl)                              \
   MACRO(execlp)                             \
   MACRO(execle)                             \
-  MACRO(system)
-
-#define GLIBC_SIGNAL_WRAPPERS(MACRO)        \
+  MACRO(system)                             \
+                                            \
   MACRO(signal)                             \
   MACRO(sigaction)                          \
   MACRO(sigvec)                             \
                                             \
+  MACRO(sigset)                             \
   MACRO(sigblock)                           \
   MACRO(sigsetmask)                         \
   MACRO(siggetmask)                         \
   MACRO(sigprocmask)                        \
                                             \
+  MACRO(sigsuspend)                         \
+  MACRO(sighold)                            \
+  MACRO(sigignore)                          \
+  MACRO(__sigpause)                         \
+  MACRO(sigpause)                           \
+  MACRO(sigrelse)                           \
+                                            \
   MACRO(sigwait)                            \
   MACRO(sigwaitinfo)                        \
-  MACRO(sigtimedwait)
-
-#define GLIBC_MISC_WRAPPERS(MACRO)          \
+  MACRO(sigtimedwait)                       \
+                                            \
   MACRO(fork)                               \
   MACRO(__clone)                            \
   MACRO(open)                               \
@@ -250,37 +177,54 @@ extern "C"
   MACRO(ptsname_r)                          \
   MACRO(getpt)                              \
   MACRO(openlog)                            \
-  MACRO(closelog)
-//  MACRO(creat)
-//  MACRO(openat)
-
-#define GLIBC_SYS_V_IPC_WRAPPERS(MACRO)     \
+  MACRO(closelog)                           \
+                                            \
   MACRO(shmget)                             \
   MACRO(shmat)                              \
   MACRO(shmdt)                              \
-  MACRO(shmctl)
-
-/* FOREACH_GLIBC_BASE_FUNC (MACRO) must appear first. */
-#define FOREACH_GLIBC_FUNC_WRAPPER(MACRO)   \
-  FOREACH_GLIBC_BASE_FUNC(MACRO)	    \
+  MACRO(shmctl)                             \
                                             \
-  GLIBC_SOCKET_WRAPPERS(MACRO)              \
-  GLIBC_EXEC_WRAPPERS(MACRO)                \
-  GLIBC_SIGNAL_WRAPPERS(MACRO)              \
-  GLIBC_MISC_WRAPPERS(MACRO)                \
-  GLIBC_SYS_V_IPC_WRAPPERS(MACRO)           \
+  MACRO(select)                             \
+  MACRO(read)                               \
+  MACRO(write)                              \
                                             \
-  GLIBC_PID_FAMILY_WRAPPERS(MACRO)          \
-  GLIBC_MALLOC_FAMILY_WRAPPERS(MACRO)       \
-  GLIBC_PTRACE_WRAPPERS(MACRO)              \
-  GLIBC_RECORD_WRAPPERS(MACRO)
+  MACRO(epoll_create)                       \
+  MACRO(epoll_create1)                      \
+  MACRO(epoll_ctl)                          \
+  MACRO(epoll_wait)                         \
+  MACRO(epoll_pwait)                        \
+//  MACRO(creat)
+//  MACRO(openat)
 
-//GLIBC_ACCEPT4_WRAPPER(MACRO)
+#define FOREACH_LIBPTHREAD_WRAPPERS(MACRO)  \
+  MACRO(pthread_create)                     \
+  MACRO(pthread_exit)                       \
+  MACRO(pthread_tryjoin_np)                 \
+  MACRO(pthread_sigmask)                    \
+  MACRO(pthread_mutex_lock)                 \
+  MACRO(pthread_mutex_trylock)              \
+  MACRO(pthread_mutex_unlock)               \
+  MACRO(pthread_rwlock_unlock)              \
+  MACRO(pthread_rwlock_rdlock)              \
+  MACRO(pthread_rwlock_tryrdlock)           \
+  MACRO(pthread_rwlock_wrlock)              \
+  MACRO(pthread_rwlock_trywrlock)           \
+  MACRO(pthread_cond_broadcast)             \
+  MACRO(pthread_cond_destroy)               \
+  MACRO(pthread_cond_init)                  \
+  MACRO(pthread_cond_signal)                \
+  MACRO(pthread_cond_timedwait)             \
+  MACRO(pthread_cond_wait)
+
+#define FOREACH_DMTCP_WRAPPER(MACRO)            \
+  FOREACH_GLIBC_WRAPPERS(MACRO)                 \
+  FOREACH_GLIBC_MALLOC_FAMILY_WRAPPERS(MACRO)
 
 # define ENUM(x) enum_ ## x
 # define GEN_ENUM(x) ENUM(x),
   typedef enum {
-    FOREACH_GLIBC_FUNC_WRAPPER(GEN_ENUM)
+    FOREACH_DMTCP_WRAPPER(GEN_ENUM)
+    FOREACH_LIBPTHREAD_WRAPPERS(GEN_ENUM)
     numLibcWrappers
   } LibcWrapperOffset;
 
@@ -289,34 +233,32 @@ extern "C"
 
   void _dmtcp_remutex_on_fork();
 
+  LIB_PRIVATE void *_dmtcp_get_libc_dlsym_addr();
+
   int _dmtcp_unsetenv(const char *name);
+  void initialize_libc_wrappers();
+  void initialize_libpthread_wrappers();
+  void initializeJalib();
 
-#ifdef RECORD_REPLAY
-  typedef enum {
-    FOREACH_PTHREAD_FUNC_WRAPPER(GEN_ENUM)
-    numLibpthreadWrappers
-  } LibPthreadWrapperOffset;
-#endif
   int _real_socket ( int domain, int type, int protocol );
-  int _real_connect ( int sockfd,  const  struct sockaddr *serv_addr, socklen_t addrlen );
-  int _real_bind ( int sockfd,  const struct  sockaddr  *my_addr,  socklen_t addrlen );
+  int _real_connect ( int sockfd,  const  struct sockaddr *serv_addr,
+                      socklen_t addrlen );
+  int _real_bind ( int sockfd,  const struct  sockaddr  *my_addr,
+                   socklen_t addrlen );
   int _real_listen ( int sockfd, int backlog );
   int _real_accept ( int sockfd, struct sockaddr *addr, socklen_t *addrlen );
-//#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,28)) && __GLIBC_PREREQ(2,10)
-  int _real_accept4 ( int sockfd, struct sockaddr *addr, socklen_t *addrlen, int flags );
-//#endif
-#ifdef RECORD_REPLAY
-  int _real_getsockname(int sockfd, struct sockaddr *addr, socklen_t *addrlen);
-  int _real_getpeername(int sockfd, struct sockaddr *addr, socklen_t *addrlen);
-#endif
-  int _real_setsockopt ( int s, int  level,  int  optname,  const  void  *optval,
+  int _real_accept4 ( int sockfd, struct sockaddr *addr, socklen_t *addrlen,
+                      int flags );
+  int _real_setsockopt ( int s, int level, int optname, const void *optval,
                          socklen_t optlen );
-
+  int _real_getsockopt ( int s, int level, int optname, void *optval,
+                         socklen_t *optlen );
 
   int _real_fexecve ( int fd, char *const argv[], char *const envp[] );
   int _real_execve ( const char *filename, char *const argv[], char *const envp[] );
   int _real_execv ( const char *path, char *const argv[] );
   int _real_execvp ( const char *file, char *const argv[] );
+  int _real_execvpe(const char *file, char *const argv[], char *const envp[]);
 // int _real_execl(const char *path, const char *arg, ...);
 // int _real_execlp(const char *file, const char *arg, ...);
 // int _real_execle(const char *path, const char *arg, ..., char * const envp[]);
@@ -325,19 +267,16 @@ extern "C"
   pid_t _real_fork();
   int _real_clone ( int ( *fn ) ( void *arg ), void *child_stack, int flags, void *arg, int *parent_tidptr, struct user_desc *newtls, int *child_tidptr );
 
-  int _real_open(const char *pathname, int flags, mode_t mode);
-  int _real_open64(const char *pathname, int flags, mode_t mode);
+  int _real_open(const char *pathname, int flags, ...);
+  int _real_open64(const char *pathname, int flags, ...);
   FILE* _real_fopen(const char *path, const char *mode);
   FILE* _real_fopen64(const char *path, const char *mode);
   int _real_close ( int fd );
   int _real_fclose ( FILE *fp );
   void _real_exit ( int status );
 
-//we no longer wrap dup
-//#define _real_dup  dup
+#define _real_dup  dup
 #define _real_dup2 dup2
-//int _real_dup(int oldfd);
-//int _real_dup2(int oldfd, int newfd);
 
   int _real_ptsname_r ( int fd, char * buf, size_t buflen );
   int _real_getpt ( void );
@@ -347,13 +286,15 @@ extern "C"
   void _real_openlog ( const char *ident, int option, int facility );
   void _real_closelog ( void );
 
-
-  typedef void (*sighandler_t)(int);
+  // Despite what 'man signal' says, signal.h already defines sighandler_t
+  // typedef void (*sighandler_t)(int);
 
   //set the handler
   sighandler_t _real_signal(int signum, sighandler_t handler);
-  int _real_sigaction(int signum, const struct sigaction *act, struct sigaction *oldact);
-  int _real_rt_sigaction(int signum, const struct sigaction *act, struct sigaction *oldact);
+  int _real_sigaction(int signum, const struct sigaction *act,
+                      struct sigaction *oldact);
+  int _real_rt_sigaction(int signum, const struct sigaction *act,
+                         struct sigaction *oldact);
   int _real_sigvec(int sig, const struct sigvec *vec, struct sigvec *ovec);
 
   //set the mask
@@ -362,7 +303,16 @@ extern "C"
   int _real_siggetmask(void);
   int _real_sigprocmask(int how, const sigset_t *set, sigset_t *oldset);
   int _real_rt_sigprocmask(int how, const sigset_t *set, sigset_t *oldset);
-  int _real_pthread_sigmask(int how, const sigset_t *newmask, sigset_t *oldmask);
+  int _real_pthread_sigmask(int how, const sigset_t *newmask,
+                            sigset_t *oldmask);
+
+  int _real_sigsuspend(const sigset_t *mask);
+  int _real_sighold(int sig);
+  int _real_sigignore(int sig);
+  int _real__sigpause(int __sig_or_mask, int __is_sig);
+  int _real_sigpause(int sig);
+  int _real_sigrelse(int sig);
+  sighandler_t _real_sigset(int sig, sighandler_t disp);
 
   int _real_sigwait(const sigset_t *set, int *sig);
   int _real_sigwaitinfo(const sigset_t *set, siginfo_t *info);
@@ -381,6 +331,74 @@ extern "C"
   int _real_shmdt(const void *shmaddr);
   int _real_shmctl(int shmid, int cmd, struct shmid_ds *buf);
 
+  int _real_pthread_create(pthread_t *thread, const pthread_attr_t *attr,
+      void *(*start_routine)(void*), void *arg);
+  void _real_pthread_exit(void *retval) __attribute__ ((__noreturn__));
+  int _real_pthread_tryjoin_np(pthread_t thread, void **retval);
+
+  int _real_xstat(int vers, const char *path, struct stat *buf);
+  int _real_xstat64(int vers, const char *path, struct stat64 *buf);
+  int _real_lxstat(int vers, const char *path, struct stat *buf);
+  int _real_lxstat64(int vers, const char *path, struct stat64 *buf);
+  ssize_t _real_readlink(const char *path, char *buf, size_t bufsiz);
+  void * _real_dlsym ( void *handle, const char *symbol );
+
+  void *_real_dlopen(const char *filename, int flag);
+  int _real_dlclose(void *handle);
+
+  void *_real_calloc(size_t nmemb, size_t size);
+  void *_real_malloc(size_t size);
+  void  _real_free(void *ptr);
+  void *_real_realloc(void *ptr, size_t size);
+  void *_real_libc_memalign(size_t boundary, size_t size);
+  void *_real_mmap(void *addr, size_t length, int prot, int flags,
+      int fd, off_t offset);
+  void *_real_mmap64(void *addr, size_t length, int prot, int flags,
+      int fd, off64_t offset);
+#if __GLIBC_PREREQ (2,4)
+  void *_real_mremap(void *old_address, size_t old_size, size_t new_size,
+      int flags, ... /* void *new_address */ );
+#else
+  void *_real_mremap(void *old_address, size_t old_size, size_t new_size,
+      int flags);
+#endif
+  int _real_munmap(void *addr, size_t length);
+
+  ssize_t _real_read(int fd, void *buf, size_t count);
+  ssize_t _real_write(int fd, const void *buf, size_t count);
+  int _real_select(int nfds, fd_set *readfds, fd_set *writefds,
+                   fd_set *exceptfds, struct timeval *timeout);
+  int _real_dup(int oldfd);
+  int _real_dup2(int oldfd, int newfd);
+  int _real_dup3(int oldfd, int newfd, int flags);
+  off_t _real_lseek(int fd, off_t offset, int whence);
+  int _real_unlink(const char *pathname);
+
+  int _real_pthread_mutex_lock(pthread_mutex_t *mutex);
+  int _real_pthread_mutex_trylock(pthread_mutex_t *mutex);
+  int _real_pthread_mutex_unlock(pthread_mutex_t *mutex);
+  int _real_pthread_rwlock_unlock(pthread_rwlock_t *rwlock);
+  int _real_pthread_rwlock_rdlock(pthread_rwlock_t *rwlock);
+  int _real_pthread_rwlock_tryrdlock(pthread_rwlock_t *rwlock);
+  int _real_pthread_rwlock_wrlock(pthread_rwlock_t *rwlock);
+  int _real_pthread_rwlock_trywrlock(pthread_rwlock_t *rwlock);
+  int _real_pthread_cond_broadcast(pthread_cond_t *cond);
+  int _real_pthread_cond_destroy(pthread_cond_t *cond);
+  int _real_pthread_cond_init(pthread_cond_t *cond,
+                              const pthread_condattr_t *attr);
+  int _real_pthread_cond_signal(pthread_cond_t *cond);
+  int _real_pthread_cond_timedwait(pthread_cond_t *cond, pthread_mutex_t *mutex,
+                                   const struct timespec *abstime);
+  int _real_pthread_cond_wait(pthread_cond_t *cond, pthread_mutex_t *mutex);
+
+  int _real_epoll_create(int size);
+  int _real_epoll_create1(int flags);
+  int _real_epoll_ctl(int epfd, int op, int fd, struct epoll_event *event);
+  int _real_epoll_wait(int epfd, struct epoll_event *events,
+                       int maxevents, int timeout);
+  int _real_epoll_pwait(int epfd, struct epoll_event *events,
+                        int maxevents, int timeout, const sigset_t *sigmask);
+
 #ifdef PID_VIRTUALIZATION
   pid_t _real_getpid(void);
   pid_t _real_getppid(void);
@@ -403,117 +421,18 @@ extern "C"
   pid_t _real_waitpid(pid_t pid, int *stat_loc, int options);
   int   _real_waitid(idtype_t idtype, id_t id, siginfo_t *infop, int options);
 
-  pid_t _real_wait3(__WAIT_STATUS status, int options,      struct rusage *rusage);
-  pid_t _real_wait4(pid_t pid, __WAIT_STATUS status, int options,      struct rusage *rusage);
-  extern int send_sigwinch;
+  pid_t _real_wait3(__WAIT_STATUS status, int options, struct rusage *rusage);
+  pid_t _real_wait4(pid_t pid, __WAIT_STATUS status, int options,
+                    struct rusage *rusage);
+  LIB_PRIVATE extern int send_sigwinch;
   int _real_ioctl(int d,  unsigned long int request, ...) __THROW;
 
   int _real_setgid(gid_t gid);
   int _real_setuid(uid_t uid);
 
-#endif /* PID_VIRTUALIZATION */
-
-#ifdef RECORD_REPLAY
-  int _real_mkdir(const char *pathname, mode_t mode);
-  int _real_mkstemp(char *temp);
-  FILE * _real_fdopen(int fd, const char *mode);
-  char * _real_fgets(char *s, int size, FILE *stream);
-  ssize_t _real_getline(char **lineptr, size_t *n, FILE *stream);
-  int _real_getc(FILE *stream);
-  int _real_fgetc(FILE *stream);
-  int _real_ungetc(int c, FILE *stream);
-  int _real_putc(int c, FILE *stream);
-  int _real_fcntl(int fd, int cmd, ...);
-  int _real_fdatasync(int fd);
-  int _real_fsync(int fd);
-  int _real_fputs(const char *s, FILE *stream);
-//int _real_fcntl(int fd, int cmd, long arg_3);
-//int _real_fcntl(int fd, int cmd, struct flock *arg_3);
-  int _real_fxstat(int vers, int fd, struct stat *buf);
-  int _real_fxstat64(int vers, int fd, struct stat64 *buf);
-  int _real_link(const char *oldpath, const char *newpath);
-  int _real_rename(const char *oldpath, const char *newpath);
-  void _real_rewind(FILE *stream);
-  int _real_rmdir(const char *pathname);
-  long _real_ftell(FILE *stream);
-  size_t _real_fwrite(const void *ptr, size_t size, size_t nmemb, FILE *stream);
-#endif
-  int _real_xstat(int vers, const char *path, struct stat *buf);
-  int _real_xstat64(int vers, const char *path, struct stat64 *buf);
-  int _real_lxstat(int vers, const char *path, struct stat *buf);
-  int _real_lxstat64(int vers, const char *path, struct stat64 *buf);
-  ssize_t _real_readlink(const char *path, char *buf, size_t bufsiz);
-
-#ifdef PTRACE
-  void * _real_dlsym ( void *handle, const char *symbol );
   long _real_ptrace ( enum __ptrace_request request, pid_t pid, void *addr,
                     void *data);
-  td_err_e  _real_td_thr_get_info ( const td_thrhandle_t  *th_p,
-                                    td_thrinfo_t *ti_p);
-#endif
-  int _real_pthread_join(pthread_t thread, void **value_ptr);
-
-  int _real_xstat(int vers, const char *path, struct stat *buf);
-  int _real_xstat64(int vers, const char *path, struct stat64 *buf);
-  int _real_lxstat(int vers, const char *path, struct stat *buf);
-  int _real_lxstat64(int vers, const char *path, struct stat64 *buf);
-
-#ifdef ENABLE_MALLOC_WRAPPER
-  void *_real_calloc(size_t nmemb, size_t size);
-  void *_real_malloc(size_t size);
-  void  _real_free(void *ptr);
-  void *_real_realloc(void *ptr, size_t size);
-  void *_real_libc_memalign(size_t boundary, size_t size);
-#ifdef RECORD_REPLAY
-  void *_real_mmap(void *addr, size_t length, int prot, int flags,
-      int fd, off_t offset);
-  void *_real_mmap64(void *addr, size_t length, int prot, int flags,
-      int fd, off64_t offset);
-  void *_real_mremap(void *old_address, size_t old_size, size_t new_size,
-      int flags, void *new_address);
-  int _real_munmap(void *addr, size_t length);
-  void * _mmap_no_sync(void *addr, size_t length, int prot, int flags,
-      int fd, off_t offset);
-  int _munmap_no_sync(void *addr, size_t length);
-#endif
-  //int _real_vfprintf ( FILE *s, const char *format, va_list ap );
-#endif
-
-#ifdef RECORD_REPLAY  
-  int _real_pthread_mutex_lock(pthread_mutex_t *mutex);
-  int _real_pthread_mutex_trylock(pthread_mutex_t *mutex);
-  int _real_pthread_mutex_unlock(pthread_mutex_t *mutex);
-  int _real_pthread_rwlock_unlock(pthread_rwlock_t *rwlock);
-  int _real_pthread_rwlock_rdlock(pthread_rwlock_t *rwlock);
-  int _real_pthread_rwlock_wrlock(pthread_rwlock_t *rwlock);
-  int _real_pthread_cond_signal(pthread_cond_t *cond);
-  int _real_pthread_cond_broadcast(pthread_cond_t *cond);
-  int _real_pthread_cond_wait(pthread_cond_t *cond, pthread_mutex_t *mutex);
-  int _real_pthread_cond_timedwait(pthread_cond_t *cond, pthread_mutex_t *mutex,
-      const struct timespec *abstime);
-  int _real_pthread_create(pthread_t *thread, const pthread_attr_t *attr,
-      void *(*start_routine)(void*), void *arg);
-  void _real_pthread_exit(void *value_ptr);
-  int _real_pthread_detach(pthread_t thread);
-  int _real_pthread_join(pthread_t thread, void **value_ptr);
-  int _real_pthread_kill(pthread_t thread, int sig);
-  int _real_access(const char *pathname, int mode);
-  int _real_select(int nfds, fd_set *readfds, fd_set *writefds, 
-      fd_set *exceptfds, struct timeval *timeout);
-  int _real_read(int fd, void *buf, size_t count);
-  struct dirent *_real_readdir(DIR *dirp);
-  int _real_readdir_r(DIR *dirp, struct dirent *entry, struct dirent **result);
-  ssize_t        _real_write(int fd, const void *buf, size_t count);
-  int _real_rand(void);
-  void _real_srand(unsigned int seed);
-  time_t _real_time(time_t *tloc);
-  int _real_dup(int oldfd);
-  off_t _real_lseek(int fd, off_t offset, int whence);
-  int _real_unlink(const char *pathname);
-  ssize_t _real_pread(int fd, void *buf, size_t count, off_t offset);
-  ssize_t _real_pwrite(int fd, const void *buf, size_t count, off_t offset);
-  sighandler_t _real_sigset(int sig, sighandler_t disp);
-#endif
+#endif /* PID_VIRTUALIZATION */
 
 #ifdef __cplusplus
 }
