@@ -39,12 +39,12 @@ static pthread_mutex_t tblLock = PTHREAD_MUTEX_INITIALIZER;
 
 static void _do_lock_tbl()
 {
-  JASSERT(pthread_mutex_lock(&tblLock) == 0) (JASSERT_ERRNO);
+  JASSERT(_real_pthread_mutex_lock(&tblLock) == 0) (JASSERT_ERRNO);
 }
 
 static void _do_unlock_tbl()
 {
-  JASSERT(pthread_mutex_unlock(&tblLock) == 0) (JASSERT_ERRNO);
+  JASSERT(_real_pthread_mutex_unlock(&tblLock) == 0) (JASSERT_ERRNO);
 }
 
 dmtcp::VirtualPidTable::VirtualPidTable()
@@ -57,15 +57,21 @@ dmtcp::VirtualPidTable::VirtualPidTable()
   _isRootOfProcessTree = false;
   _childTable.clear();
   _tidVector.clear();
+  _pthreadJoinId.clear();
   _inferiorVector.clear();
   _pidMapTable.clear();
   _pidMapTable[_pid] = _pid;
   _do_unlock_tbl();
 }
 
+static dmtcp::VirtualPidTable *inst = NULL;
 dmtcp::VirtualPidTable& dmtcp::VirtualPidTable::instance()
 {
-  static VirtualPidTable *inst = new VirtualPidTable(); return *inst;
+  if (inst == NULL) {
+    inst = new VirtualPidTable();
+  }
+  return *inst;
+//  static VirtualPidTable *inst = new VirtualPidTable(); return *inst;
 }
 
 bool dmtcp::VirtualPidTable::isConflictingPid( pid_t pid)
@@ -79,6 +85,13 @@ bool dmtcp::VirtualPidTable::isConflictingPid( pid_t pid)
    *       there is not conflict because that mapping essentially is about the
    *       current pid.
    */
+#if 0
+  // Use a similar block to emulate tid-conflict. This can come handy in
+  // finding bugs related to tid-conflict.
+  if (pid >= 10000 && pid < 20000) {
+    return true;
+  }
+#endif
   if (pid == instance().originalToCurrentPid( pid ))
     return false;
 
@@ -160,7 +173,6 @@ void dmtcp::VirtualPidTable::restoreProcessGroupInfo()
 
 void dmtcp::VirtualPidTable::printPidMaps()
 {
-#ifdef DEBUG
   ostringstream out;
   out << "Pid mappings\n";
   out << "      original" << "  ->  " << "current" << "\n";
@@ -169,25 +181,27 @@ void dmtcp::VirtualPidTable::printPidMaps()
     pid_t currentPid  = i->second;
     out << "\t" << originalPid << "\t->   " << currentPid << "\n";
   }
-  JTRACE("Original To Current Pid Mappings:") ( out.str());
-#endif
+  JTRACE("Original To Current Pid Mappings:") (_pidMapTable.size()) (out.str());
 }
 
 void dmtcp::VirtualPidTable::resetOnFork()
 {
+  pthread_mutex_t newlock = PTHREAD_MUTEX_INITIALIZER;
+  tblLock = newlock;
   _pid = _real_getpid();
   _ppid = currentToOriginalPid ( _real_getppid() );
   _isRootOfProcessTree = false;
   _childTable.clear();
   _tidVector.clear();
+  _pthreadJoinId.clear();
   _inferiorVector.clear();
   //_pidMapTable[_pid] = _pid;
-  JTRACE("current original to current mappings:") (_pidMapTable.size());
   printPidMaps();
 }
 
 pid_t dmtcp::VirtualPidTable::originalToCurrentPid( pid_t originalPid )
-{
+{ pid_t retVal = 0;
+
   /* This code is called from MTCP while the checkpoint thread is holding
      the JASSERT log lock. Therefore, don't call JTRACE/JASSERT/JINFO/etc. in
      this function. */
@@ -198,8 +212,10 @@ pid_t dmtcp::VirtualPidTable::originalToCurrentPid( pid_t originalPid )
     return originalPid;
   }
 
+  // retVal required: in TID conflict, first and second clone call can interfere
+  retVal = i->second;
   _do_unlock_tbl();
-  return i->second;
+  return retVal;
 }
 
 pid_t dmtcp::VirtualPidTable::currentToOriginalPid( pid_t currentPid )
@@ -351,6 +367,27 @@ void dmtcp::VirtualPidTable::eraseInferior( pid_t tid )
   return;
 }
 
+bool dmtcp::VirtualPidTable::beginPthreadJoin(pthread_t thread)
+{
+  bool res = false;
+  _do_lock_tbl();
+  dmtcp::map<pthread_t, pthread_t>::iterator i = _pthreadJoinId.find(thread);
+  if (i == _pthreadJoinId.end()) {
+    _pthreadJoinId[thread] = pthread_self();
+    res = true;
+  }
+  _do_unlock_tbl();
+  return res;
+}
+
+void dmtcp::VirtualPidTable::endPthreadJoin(pthread_t thread)
+{
+  _do_lock_tbl();
+  JASSERT(pthread_equal(_pthreadJoinId[thread], pthread_self()));
+  _pthreadJoinId.erase(thread);
+  _do_unlock_tbl();
+}
+
 bool dmtcp::VirtualPidTable::pidExists( pid_t pid )
 {
   bool retVal = false;
@@ -486,7 +523,7 @@ void dmtcp::VirtualPidTable::serializePidMap ( jalib::JBinarySerializer& o )
   }
   else
   {
-    for (int i = 0; i < numMaps; i++) {
+    for (size_t i = 0; i < numMaps; i++) {
       serializePidMapEntry ( o, originalPid, currentPid );
       _pidMapTable[originalPid] = currentPid;
     }
@@ -516,16 +553,15 @@ void dmtcp::VirtualPidTable::serializeEntryCount ( jalib::JBinarySerializer& o,
 void dmtcp::VirtualPidTable::InsertIntoPidMapFile( pid_t originalPid, pid_t currentPid)
 {
 
-  dmtcp::string pidMapFile = "/proc/self/fd/" 
+  dmtcp::string pidMapFile = "/proc/self/fd/"
                              + jalib::XToString ( PROTECTED_PIDMAP_FD );
-  dmtcp::string pidMapCountFile = "/proc/self/fd/" 
+  dmtcp::string pidMapCountFile = "/proc/self/fd/"
                                   + jalib::XToString ( PROTECTED_PIDMAPCNT_FD );
 
   pidMapFile =  jalib::Filesystem::ResolveSymlink ( pidMapFile );
   pidMapCountFile =  jalib::Filesystem::ResolveSymlink ( pidMapCountFile );
-
   JASSERT ( pidMapFile.length() > 0 && pidMapCountFile.length() > 0 )
-    ( pidMapFile )( pidMapCountFile );
+    ( pidMapFile )( pidMapCountFile ).Text("Failed to resolve symlink.");
 
   // Create Serializers
   jalib::JBinarySerializeWriterRaw mapwr( pidMapFile, PROTECTED_PIDMAP_FD );
@@ -554,7 +590,7 @@ void dmtcp::VirtualPidTable::InsertIntoPidMapFile( pid_t originalPid, pid_t curr
 
 void dmtcp::VirtualPidTable::readPidMapsFromFile()
 {
-  dmtcp::string pidMapFile = "/proc/self/fd/" 
+  dmtcp::string pidMapFile = "/proc/self/fd/"
                              + jalib::XToString ( PROTECTED_PIDMAP_FD );
   pidMapFile =  jalib::Filesystem::ResolveSymlink ( pidMapFile );
   dmtcp::string pidMapCountFile = "/proc/self/fd/"
