@@ -7,10 +7,15 @@ import pty
 import socket
 import os
 import sys
+import errno
+import signal
 import resource
+import pwd
 import stat
 import re
 
+signal.alarm(1800)  # half hour
+
 if sys.version_info[0] != 2 or sys.version_info[0:2] < (2,4):
   print "test/autotest.py works only with Python 2.x for 2.x greater than 2.3"
   print "Change the beginning of test/autotest.py if you believe you can run."
@@ -32,9 +37,23 @@ DEFAULT_S=0.3
 S=DEFAULT_S
 #Appears as S*SLOW in code.  If --slow, then SLOW=5
 SLOW=1
+#In the case of gdb, even if both gdb and the inferior are running after
+#ckpt or restart, this does not guarantee that the ptrace-related work
+#(that is needed at resume or restart) is over. The ptrace related work happens
+#in the signal handler. Proceeding while still being inside the signal handler,
+#can lead to bad consquences. To play it on the safe side, PTRACE_SLEEP was
+#set at 2 seconds.  (Until this is fixed, --enable-ptrace-support will
+#remain experimental.)
+#if testconfig.PTRACE_SUPPORT == "yes":
+#  PTRACE_SLEEP=2
 
 #Max time to wait for ckpt/restart to finish (sec)
 TIMEOUT=10
+# Raise this value when /usr/lib/locale/locale-archive is 100MB.
+# This can happen on Red Hat-derived distros.
+if os.path.exists("/usr/lib/locale/locale-archive") and \
+   os.path.getsize("/usr/lib/locale/locale-archive") > 10e6:
+  TIMEOUT *= int( os.path.getsize("/usr/lib/locale/locale-archive") / 10e6 )
 
 #Interval between checks for ckpt/restart complete
 INTERVAL=0.1
@@ -45,8 +64,8 @@ BUFFER_SIZE=4096*8
 #False redirects process stderr
 VERBOSE=False
 
-#Run (most) tests with gzip enable 
-GZIP="1"
+#Run (most) tests with user default (usually with gzip enable)
+GZIP=os.getenv('DMTCP_GZIP') or "1"
 
 #Warn cant create a file of size:
 REQUIRE_MB=50
@@ -58,7 +77,7 @@ BIN="./bin/"
 args={}
 for i in sys.argv:
   args[i]=True
-  if i=="-v":
+  if i=="-v" or i=="--verbose":
     VERBOSE=True
   if i=="--stress":
     CYCLES=100000
@@ -71,7 +90,6 @@ for i in sys.argv:
 
 stats = [0, 0]
 
-# NOTE:  This might be replaced by shell=True in call to subprocess.Popen
 def xor(bool1, bool2):
   return (bool1 or bool2) and (not bool1 or not bool2)
 
@@ -123,7 +141,7 @@ if os.system("test -d bin") is not 0:
 assert os.system("test -d bin") is 0
 
 #make sure dmtcp is built
-if os.system("make -s --no-print-directory all tests") != 0:
+if os.system("make -s --no-print-directory tests") != 0:
   print "`make all tests` FAILED"
   sys.exit(1)
 
@@ -146,48 +164,62 @@ class MySubprocess:
     self.stdout = os.open(os.devnull, os.O_WRONLY)
     self.stderr = os.open(os.devnull, os.O_WRONLY)
 
+def master_read(fd):
+  os.read(fd, 4096)
+  return ''
+
 #launch a child process
 # NOTE:  Can eventually migrate to Python 2.7:  subprocess.check_output
-childStdoutDevNull = False
+devnullFd = os.open(os.devnull, os.O_WRONLY)
 def launch(cmd):
-  global childStdoutDevNull
+  global devnullFd
+  global master_read
   if VERBOSE:
     print "Launching... ", cmd
   cmd = splitWithQuotes(cmd);
   # Example cmd:  dmtcp_checkpoint screen ...
   ptyMode = False
   for str in cmd:
-    if re.search("(_|/|^)(screen|script)(_|$)", str):
+    # Checkpoint image can be emacs23_x, or whatever emacs is a link to.
+    # vim can be vim.gnome, etc.
+    if re.search("(_|/|^)(screen|script|vim.*|emacs.*|pty)(_|$)", str):
       ptyMode = True
   try:
     os.stat(cmd[0])
   except:
     raise CheckFailed(cmd[0] + " not found")
   if ptyMode:
-    (pid, fd) = os.forkpty()
+    # FOR DEBUGGING:  This can mysteriously fail, causing pty.fork() to fail
+    try:
+      (fd1, fd2) = os.openpty()
+    except OSError, e:
+      print "\n\n/dev/ptmx:"; os.system("ls -l /dev/ptmx /dev/pts")
+      raise e
+    else:
+      os.close(fd1); os.close(fd2)
+    (pid, fd) = pty.fork()
     if pid == 0:
-      # replace stdout; child might otherwise block on writing to stdout
-      os.close(1)
-      os.open(os.devnull, os.O_WRONLY | os.O_APPEND)
-      os.close(2)
-      os.open(os.devnull, os.O_WRONLY | os.O_APPEND)
-      os.execvp(cmd[0], cmd)
-      # os.system( reduce(lambda x,y:x+y, cmd) )
-      # os.exit(0)
+      signal.alarm(300) # pending alarm inherited across exec, but not a fork
+      # Problem:  pty.spawn invokes fork.  alarm() will have no effect.
+      pty.spawn(cmd, master_read)
+      sys.exit(0)
     else:
       return MySubprocess(pid)
   else:
-    childStderr = subprocess.STDOUT # Mix stderr into stdout file object
     if cmd[0] == BIN+"dmtcp_coordinator":
       childStdout = subprocess.PIPE
-      childStderr = subprocess.PIPE  # Don't mix stderr in; need to read stdout
+      # Don't mix stderr in with childStdout; need to read stdout
+      if VERBOSE:
+        childStderr = None
+      else:
+        childStderr = devnullFd
     elif VERBOSE:
       childStdout=None  # Inherit child stdout from parent
+      childStderr=None  # Inherit child stderr from parent
     else:
-      if childStdoutDevNull:
-        os.close(childStdoutDevNull)
-      childStdout = os.open(os.devnull, os.O_WRONLY)
-      childStdoutDevNull = childStdout
+      childStdout = devnullFd
+      childStderr = subprocess.STDOUT # Mix stderr into stdout file object
+    # NOTE:  This might be replaced by shell=True in call to subprocess.Popen
     proc = subprocess.Popen(cmd, bufsize=BUFFER_SIZE,
 		 stdin=subprocess.PIPE, stdout=childStdout,
 		 stderr=childStderr, close_fds=True)
@@ -223,6 +255,16 @@ os.system("rm -f "+tmpfile)
 
 os.environ['DMTCP_GZIP'] = GZIP
 
+# Temporary hack until DMTCP cleans up when using --enable-ptrace-support
+def deletePtraceFiles():
+  tmpdir = os.getenv("TMPDIR", "/tmp")  # if "TMPDIR" not set, return "/tmp"
+  tmpdir += "/dmtcp-" + pwd.getpwuid(os.getuid()).pw_name + \
+            "@" + socket.gethostname()
+  os.system("cd "+tmpdir+"; "+
+	    "rm -f ptrace_shared.txt ptrace_setoptions.txt \
+	     ptrace_ckpthreads.txt ptrace_shared.txt ptrace_setoptions.txt \
+	     ptrace_ckpthreads.txt new_ptrace_shared.txt ckpt_leader_file.txt")
+
 #launch the coordinator
 coordinator = launch(BIN+"dmtcp_coordinator")
 
@@ -245,6 +287,7 @@ def SHUTDOWN():
     print "SHUTDOWN() failed"
   os.system("kill -9 %d" % coordinator.pid)
   os.system("rm -rf  %s" % ckptDir)
+  os.close(devnullFd)
 
 #make sure val is true
 def CHECK(val, msg):
@@ -253,7 +296,7 @@ def CHECK(val, msg):
 
 #wait TIMEOUT for test() to be true, or throw error
 def WAITFOR(test, msg):
-  left=TIMEOUT/INTERVAL
+  left=TIMEOUT*(S/DEFAULT_S)/INTERVAL
   while not test():
     if left <= 0:
       CHECK(False, msg())
@@ -292,12 +335,22 @@ def getStatus():
     print "STATUS: peers=%s, running=%s" % (peers,running)
   return (int(peers), (running=="yes"))
 
-#delete all files in ckpDir
+#delete all files in ckptDir
 def clearCkptDir():
+  for TRIES in range(2):  # Try twice in case ckpt_*_dmtcp.temp is renamed.
     #clear checkpoint dir
     for root, dirs, files in os.walk(ckptDir, topdown=False):
       for name in files:
-        os.remove(os.path.join(root, name))
+        try:
+          # if name.endswith(".dmtcp") :
+          #   import shutil
+          #   shutil.copy(os.path.join(root, name), "/home/kapil/dmtcp/ramfs")
+          # else:
+          #   os.remove(os.path.join(root, name))
+          os.remove(os.path.join(root, name))
+        except OSError, e:
+	  if e.errno != errno.ENOENT:  # Maybe ckpt_*_dmtcp.temp was renamed.
+	    raise e
       for name in dirs:
         os.rmdir(os.path.join(root, name))
 
@@ -305,32 +358,66 @@ def getNumCkptFiles(dir):
   return len(filter(lambda f: f.startswith("ckpt_") and f.endswith(".dmtcp"), listdir(dir)))
 
 
-#test a given list of commands to see if they checkpoint
-def runTest(name, numProcs, cmds):
+# Test a given list of commands to see if they checkpoint
+# runTest() sets up a keyboard interrupt handler, and then calls this function.
+def runTestRaw(name, numProcs, cmds):
   #the expected/correct running status
+  if testconfig.USE_M32 == "1":
+    def forall(fnc, lst):
+      return reduce(lambda x, y: x and y, map(fnc, lst))
+    if not forall(lambda x: x.startswith("./test/"), cmds):
+      return
   status=(numProcs, True)
   procs=[]
 
+  def doesStatusSatisfy(newStatus,requiredStatus):
+    if isinstance(requiredStatus[0], int):
+      statRange = [requiredStatus[0]]
+    elif isinstance(requiredStatus[0], list):
+      statRange = requiredStatus[0]
+    else:
+      raise NotImplementedError
+    return newStatus[0] in statRange and newStatus[1] == requiredStatus[1]
+
   def wfMsg(msg):
     #return function to generate error message
-    return lambda: msg+", "+str(status[0])+" expected, %d found, running=%d" % getStatus()
+    return lambda: msg+", "+str(status[0])+ \
+                   " expected, %d found, running=%d" % getStatus()
 
   def testKill():
     #kill all processes
     coordinatorCmd('k')
-    WAITFOR(lambda: getStatus()==(0, False), lambda:"coordinator kill command failed")
+    try:
+      WAITFOR(lambda: getStatus()==(0, False),
+	      lambda:"coordinator kill command failed")
+    except CheckFailed:
+      global coordinator
+      coordinatorCmd('q')
+      os.system("kill -9 %d" % coordinator.pid)
+      print "Trying to kill old coordinator, and launch new one on same port"
+      coordinator = launch(BIN+"dmtcp_coordinator")
     for x in procs:
       #cleanup proc
       try:
-        if x.stdin:
-	  x.stdin.close()
-        if x.stdout:
+        if isinstance(x.stdin,int):
+          os.close(x.stdin)
+        elif x.stdin:
+          x.stdin.close()
+        if isinstance(x.stdout,int):
+          os.close(x.stdout)
+        elif x.stdout:
           x.stdout.close()
-        if x.stderr:
+        if isinstance(x.stderr,int):
+          os.close(x.stderr)
+        elif x.stderr:
           x.stderr.close()
-        os.waitpid(x.pid, os.WNOHANG)
       except:
         None
+      try:
+        os.waitpid(x.pid, os.WNOHANG)
+      except OSError, e:
+	if e.errno != errno.ECHILD:
+	  raise e
       procs.remove(x)
 
   def testCheckpoint():
@@ -338,12 +425,15 @@ def runTest(name, numProcs, cmds):
     coordinatorCmd('c')
 
     #wait for files to appear and status to return to original
-    WAITFOR(lambda: getNumCkptFiles(ckptDir)>0 and status==getStatus(),
+    WAITFOR(lambda: getNumCkptFiles(ckptDir)>0 and \
+                    doesStatusSatisfy(getStatus(), status),
             wfMsg("checkpoint error"))
 
     #make sure the right files are there
     numFiles=getNumCkptFiles(ckptDir) # len(listdir(ckptDir))
-    CHECK(numFiles==status[0], "unexpected number of checkpoint files, %d procs, %d files" % (status[0], numFiles))
+    CHECK(doesStatusSatisfy((numFiles,True),status),
+          "unexpected number of checkpoint files, %s procs, %d files"
+          % (str(status[0]), numFiles))
 
   def testRestart():
     #build restart command
@@ -353,8 +443,10 @@ def runTest(name, numProcs, cmds):
         cmd+= " "+ckptDir+"/"+i
     #run restart and test if it worked
     procs.append(launch(cmd))
-    WAITFOR(lambda: status==getStatus(), wfMsg("restart error"))
-    clearCkptDir()
+    WAITFOR(lambda: doesStatusSatisfy(getStatus(), status),
+            wfMsg("restart error"))
+    if testconfig.HBICT_DELTACOMP == "no":
+      clearCkptDir()
 
   try:
     printFixed(name,15)
@@ -371,15 +463,20 @@ def runTest(name, numProcs, cmds):
       procs.append(launch(BIN+"dmtcp_checkpoint "+cmd))
       sleep(S*SLOW)
 
-    WAITFOR(lambda: status==getStatus(), wfMsg("user program startup error"))
+    WAITFOR(lambda: doesStatusSatisfy(getStatus(), status),
+            wfMsg("user program startup error"))
 
     for i in range(CYCLES):
       if i!=0 and i%2==0:
         print #newline
         printFixed("",15)
       printFixed("ckpt:")
+      # NOTE:  If this faile, it will throw an exception to CheckFailed
+      #  of this function:  testRestart
       testCheckpoint()
       printFixed("PASSED ")
+      #if testconfig.PTRACE_SUPPORT == "yes":
+      #  sleep(PTRACE_SLEEP)
       testKill()
 
       printFixed("rstr:")
@@ -387,6 +484,8 @@ def runTest(name, numProcs, cmds):
         try:
           testRestart()
           printFixed("PASSED")
+          #if testconfig.PTRACE_SUPPORT == "yes":
+          #  sleep(PTRACE_SLEEP)
           break
         except CheckFailed, e:
           if j == RETRIES-1:
@@ -410,10 +509,47 @@ def runTest(name, numProcs, cmds):
     except CheckFailed, e:
       print "CLEANUP ERROR:", e.value
       SHUTDOWN()
+      saveResultsNMI()
       sys.exit(1)
 
   clearCkptDir()
 
+def getProcessChildren(pid):
+    p = subprocess.Popen("ps --no-headers -o pid --ppid %d" % pid, shell = True,
+                         stdout = subprocess.PIPE, stderr = subprocess.PIPE)
+    stdout, stderr = p.communicate()
+    return [int(pid) for pid in stdout.split()]
+
+# If the user types ^C, then kill all child processes.
+def runTest(name, numProcs, cmds):
+  try:
+    runTestRaw(name, numProcs, cmds)
+  except KeyboardInterrupt:
+    for pid in getProcessChildren(os.getpid()):
+      try:
+        os.kill(pid, signal.SIGKILL)
+      except OSError: # This happens if pid already died.
+        pass
+
+def saveResultsNMI():
+  if testconfig.DEBUG == "yes":
+    host = socket.getfqdn()
+    if re.search("^nmi-.*.cs.wisc.edu$", host) or \
+       re.search("^nmi-.*.cs.wisconsin.edu$", host):
+      tmpdir = os.getenv("TMPDIR", "/tmp") # if "TMPDIR" not set, return "/tmp"
+      target = "./dmtcp-" + pwd.getpwuid(os.getuid()).pw_name + \
+               "@" + socket.gethostname()
+      cmd = "mkdir results; cp -pr " + tmpdir + "/" + target + \
+	       " ./dmtcp/src/dmtcphijack.so" + \
+	       " ./dmtcp/src/dmtcp_coordinator" + \
+               " ./mtcp/libmtcp.so" + \
+               " results/"
+      os.system(cmd)
+      cmd = "tar zcf ../results.tar.gz ./results; rm -rf results"
+      os.system(cmd)
+      print "\n*** results.tar.gz ("+tmpdir+"/"+target+ \
+					      ") written to DMTCP_ROOT/.. ***"
+
 print "== Tests =="
 
 #tmp port
@@ -432,7 +568,7 @@ runTest("dmtcp2",        1, ["./test/dmtcp2"])
 # dmtcp3 creates 10 threads; Keep checkpoint image small by using gzip
 # Also, it needs some extra time to startup
 S=2
-os.environ['DMTCP_GZIP'] = "1" 
+os.environ['DMTCP_GZIP'] = "1"
 runTest("dmtcp3",        1, ["./test/dmtcp3"])
 os.environ['DMTCP_GZIP'] = GZIP
 S=DEFAULT_S
@@ -452,8 +588,39 @@ resource.setrlimit(resource.RLIMIT_STACK, [newCurrLimit, oldLimit[1]])
 runTest("dmtcp5",        2, ["./test/dmtcp5"])
 resource.setrlimit(resource.RLIMIT_STACK, oldLimit)
 
+runTest("dmtcpaware1",   1, ["./test/dmtcpaware1"])
+
+PWD=os.getcwd()
+runTest("module-sleep2", 1, ["--with-module "+
+			     PWD+"/test/module/sleep1/dmtcp_sleep1hijack.so:"+
+			     PWD+"/test/module/sleep2/dmtcp_sleep2hijack.so "+
+			     "./test/dmtcp1"])
+
+runTest("module-example-db", 2, ["--with-module "+
+			    PWD+"/test/module/example-db/dmtcp_example-dbhijack.so "+
+			     "env EXAMPLE_DB_KEY=1 EXAMPLE_DB_KEY_OTHER=2 "+
+			     "./test/dmtcp1",
+			         "--with-module "+
+			    PWD+"/test/module/example-db/dmtcp_example-dbhijack.so "+
+			     "env EXAMPLE_DB_KEY=2 EXAMPLE_DB_KEY_OTHER=1 "+
+			     "./test/dmtcp1"])
+
+# Test special case:  gettimeofday can be handled within VDSO segment.
+runTest("gettimeofday",  1, ["./test/gettimeofday"])
+
+runTest("sigchild",      1, ["./test/sigchild"])
+
 runTest("shared-fd",     2, ["./test/shared-fd"])
 
+runTest("stale-fd",      2, ["./test/stale-fd"])
+
+runTest("forkexec",      2, ["./test/forkexec"])
+
+if testconfig.PID_VIRTUALIZATION == "yes":
+  runTest("waitpid",      2, ["./test/waitpid"])
+
+runTest("client-server", 2, ["./test/client-server"])
+
 # frisbee creates three processes, each with 14 MB, if no gzip is used
 os.environ['DMTCP_GZIP'] = "1"
 runTest("frisbee",       3, ["./test/frisbee "+p1+" localhost "+p2,
@@ -465,111 +632,211 @@ runTest("shared-memory", 2, ["./test/shared-memory"])
 
 runTest("sysv-shm",      2, ["./test/sysv-shm"])
 
-#runTest("stale-fd",      2, ["./test/stale-fd"])
+#Invoke this test when we drain/restore data in pty at checkpoint time.
+# runTest("pty",   2, ["./test/pty"])
 
-runTest("forkexec",      2, ["./test/forkexec"])
+old_ld_library_path = os.getenv("LD_LIBRARY_PATH")
+os.environ['LD_LIBRARY_PATH'] = os.getenv("PWD")+"/test:"+os.getenv("PWD")
+runTest("dlopen",        1, ["./test/dlopen"])
+if old_ld_library_path:
+  os.environ['LD_LIBRARY_PATH'] = old_ld_library_path
+else:
+  del os.environ['LD_LIBRARY_PATH']
 
-if testconfig.PID_VIRTUALIZATION == "yes":
-  runTest("waitpid",      2, ["./test/waitpid"])
-
-runTest("gettimeofday",  1, ["./test/gettimeofday"])
-
-if testconfig.HAS_READLINE == "yes":
-  runTest("readline",    1,  ["./test/readline"])
+runTest("pthread1",      1, ["./test/pthread1"])
+runTest("pthread2",      1, ["./test/pthread2"])
+S=3
+runTest("pthread3",      1, ["./test/pthread2 80"])
+S=DEFAULT_S
 
 os.environ['DMTCP_GZIP'] = "1"
 runTest("gzip",          1, ["./test/dmtcp1"])
 os.environ['DMTCP_GZIP'] = GZIP
 
-runTest("dmtcpaware1",   1, ["./test/dmtcpaware1"])
-
-#Invoke this test when we drain/restore data in pty at checkpoint time.
-# runTest("pty",   2, ["./test/pty"])
+if testconfig.HAS_READLINE == "yes":
+  runTest("readline",    1,  ["./test/readline"])
 
 runTest("perl",          1, ["/usr/bin/perl"])
 
-runTest("python",        1, ["/usr/bin/python"])
+if testconfig.HAS_PYTHON == "yes":
+  runTest("python",      1, ["/usr/bin/python"])
 
 if testconfig.PID_VIRTUALIZATION == "yes":
   os.environ['DMTCP_GZIP'] = "0"
-  runTest("bash",          2, ["/bin/bash --norc -c 'ls; sleep 30; ls'"])
+  runTest("bash",        2, ["/bin/bash --norc -c 'ls; sleep 30; ls'"])
   os.environ['DMTCP_GZIP'] = GZIP
 
 if testconfig.HAS_DASH == "yes":
   os.environ['DMTCP_GZIP'] = "0"
   os.unsetenv('ENV')  # Delete reference to dash initialization file
-  runTest("dash",          2, ["/bin/dash -c 'ls; sleep 30; ls'"])
+  runTest("dash",        2, ["/bin/dash -c 'ls; sleep 30; ls'"])
   os.environ['DMTCP_GZIP'] = GZIP
 
 if testconfig.HAS_TCSH == "yes":
   os.environ['DMTCP_GZIP'] = "0"
-  runTest("tcsh",          2, ["/bin/tcsh -f -c 'ls; sleep 30; ls'"])
+  runTest("tcsh",        2, ["/bin/tcsh -f -c 'ls; sleep 30; ls'"])
   os.environ['DMTCP_GZIP'] = GZIP
 
 if testconfig.HAS_ZSH == "yes":
   os.environ['DMTCP_GZIP'] = "0"
-  runTest("zsh",          2, ["/bin/zsh -f -c 'ls; sleep 30; ls'"])
+  runTest("zsh",         2, ["/bin/zsh -f -c 'ls; sleep 30; ls'"])
   os.environ['DMTCP_GZIP'] = GZIP
 
-# *** Works manually, but not yet in autotest ***
+if testconfig.HAS_VIM == "yes" and testconfig.PID_VIRTUALIZATION == "yes":
+  # Wait to checkpoint until vim finishes reading its initialization files
+  S=3
+  if sys.version_info[0:2] >= (2,6):
+    # Delete previous vim processes.  Vim behaves poorly with stale processes.
+    vimCommand = testconfig.VIM + " /etc/passwd +3" # +3 makes cmd line unique
+    def killCommand(cmdToKill):
+      ps = subprocess.Popen(['ps', '-u', os.environ['USER'], '-o', 'pid,command'],
+    		            stdout=subprocess.PIPE).communicate()[0]
+      for row in ps.split('\n')[1:]:
+        cmd = row.split(None, 1) # maxsplit=1
+        if cmd and cmd[1] == cmdToKill:
+          os.kill(int(cmd[0]), signal.SIGKILL)
+    killCommand(vimCommand)
+    runTest("vim",       1,  ["env TERM=vt100 " + vimCommand])
+    killCommand(vimCommand)
+  S=DEFAULT_S
+
+if testconfig.HAS_EMACS == "yes" and testconfig.PID_VIRTUALIZATION == "yes":
+  # Wait to checkpoint until emacs finishes reading its initialization files
+  S=4
+  if sys.version_info[0:2] >= (2,6):
+    # Under emacs23, it opens /dev/tty directly in a new fd.
+    # To avoid this, consider using emacs --batch -l EMACS-LISTP-CODE ...
+    # ... or else a better pty wrapper to capture emacs output to /dev/tty.
+    runTest("emacs",     1,  ["env TERM=vt100 /usr/bin/emacs -nw" +
+                              " --no-init-file /etc/passwd"])
+  S=DEFAULT_S
+
 if testconfig.HAS_SCRIPT == "yes" and testconfig.PID_VIRTUALIZATION == "yes":
   S=2
   if sys.version_info[0:2] >= (2,6):
-    runTest("script",      4,  ["/usr/bin/script -f" +
+    # NOTE: If 'script' fails, try raising value of S, above, to larger number.
+    #  Arguably, there is a bug in glibc, in that locale-archive can be 100 MB.
+    #  For example, in Fedora 13 (and other recent Red Hat-derived distros?),
+    #  /usr/lib/locale/locale-archive is 100 MB, and yet 'locale -a |wc' shows
+    #  only 8KB of content in ASCII.  The 100 MB of locale-archive condenses
+    #  to 25 MB _per process_ under gzip, but this can be slow at ckpt time.
+    runTest("script",    4,  ["/usr/bin/script -f" +
     			      " -c 'bash -c \"ls; sleep 30\"'" +
     			      " dmtcp-test-typescript.tmp"])
   os.system("rm -f dmtcp-test-typescript.tmp")
   S=DEFAULT_S
 
 # SHOULD HAVE screen RUN SOMETHING LIKE:  bash -c ./test/dmtcp1
-# BUT screen -s CMD works only when CMD is single word.
-# *** Works manually, but not yet in autotest ***
 if testconfig.HAS_SCREEN == "yes" and testconfig.PID_VIRTUALIZATION == "yes":
   S=1
   if sys.version_info[0:2] >= (2,6):
-    runTest("screen",      3,  [testconfig.SCREEN + " -c /dev/null -s /bin/sh"])
+    host = socket.getfqdn()
+    runTest("screen",    3,  ["env TERM=vt100 " + testconfig.SCREEN +
+                                " -c /dev/null -s /bin/sh"])
   S=DEFAULT_S
 
+if testconfig.PTRACE_SUPPORT == "yes" and \
+   (testconfig.HAS_STRACE == "yes" or testconfig.HAS_GDB == "yes"):
+  print "  Deleting files in /tmp/dmtcp-USER@HOST before ptrace tests.  (Until"
+  print "  this is fixed, --enable-ptrace-support will remain experimental.)"
+  deletePtraceFiles()
+  if testconfig.HAS_STRACE == "yes" and testconfig.PTRACE_SUPPORT == "yes":
+    S=1
+    if sys.version_info[0:2] >= (2,6):
+      runTest("strace",    2,  ["strace test/dmtcp2"])
+    S=DEFAULT_S
+
+  deletePtraceFiles()
+  if testconfig.HAS_GDB == "yes" and testconfig.PTRACE_SUPPORT == "yes":
+    os.system("echo 'run' > dmtcp-gdbinit.tmp")
+    S=2
+    if sys.version_info[0:2] >= (2,6):
+      runTest("gdb",       2,  ["gdb -n -batch -x dmtcp-gdbinit.tmp test/dmtcp1"])
+    S=DEFAULT_S
+    os.system("rm -f dmtcp-gdbinit.tmp")
+
+  deletePtraceFiles()
+  if testconfig.HAS_GDB == "yes" and testconfig.PTRACE_SUPPORT == "yes":
+    os.system("echo 'run' > dmtcp-gdbinit.tmp")
+    S=2
+    if sys.version_info[0:2] >= (2,6):
+      runTest("gdb-pthread",2,  ["gdb -n -batch -x dmtcp-gdbinit.tmp test/dmtcp3"])
+    S=DEFAULT_S
+    os.system("rm -f dmtcp-gdbinit.tmp")
+
+if testconfig.HAS_JAVAC == "yes" and testconfig.HAS_JAVA == "yes":
+  S=3
+  os.environ['CLASSPATH'] = './test'
+  if testconfig.HAS_SUN_ORACLE_JAVA == "yes":
+    runTest("java1",         1,  ["java -Xmx512M java1"])
+  else:
+    runTest("java1",         1,  ["java java1"])
+  del os.environ['CLASSPATH']
+  S=DEFAULT_S
+
+if testconfig.HAS_CILK == "yes":
+  runTest("cilk1",        1,  ["./test/cilk1 38"])
+
 # SHOULD HAVE gcl RUN LARGE FACTORIAL OR SOMETHING.
 if testconfig.HAS_GCL == "yes":
   S=1
   runTest("gcl",         1,  [testconfig.GCL])
   S=DEFAULT_S
 
+if testconfig.HAS_OPENMP == "yes":
+  runTest("openmp-1",         1,  ["./test/openmp-1"])
+  runTest("openmp-2",         1,  ["./test/openmp-2"])
+
 # SHOULD HAVE matlab RUN LARGE FACTORIAL OR SOMETHING.
 if testconfig.HAS_MATLAB == "yes":
-  S=5
+  S=3
   if sys.version_info[0:2] >= (2,6):
-    runTest("matlab -nodisplay", 1,  [testconfig.MATLAB+" -nodisplay -nojvm"])
+    runTest("matlab-nodisplay", 1,  [testconfig.MATLAB+" -nodisplay -nojvm"])
   S=DEFAULT_S
 
-if testconfig.PTRACE_SUPPORT == "yes":
-  os.system("echo 'run' > dmtcp-gdbinit.tmp")
-  S=2
-  if sys.version_info[0:2] >= (2,6):
-    runTest("gdb", 2,  ["gdb -n -batch -x dmtcp-gdbinit.tmp test/dmtcp1"])
-  S=DEFAULT_S
-  os.system("rm -f dmtcp-gdbinit.tmp")
-
 if testconfig.HAS_MPICH == "yes":
   runTest("mpd",         1, [testconfig.MPICH_MPD])
 
-  runTest("hellompi-n1", 4, [testconfig.MPICH_MPD,
-                             testconfig.MPICH_MPIEXEC+" -n 1 ./test/hellompi"])
+  runTest("hellompich-n1", 4, [testconfig.MPICH_MPD,
+                           testconfig.MPICH_MPIEXEC+" -n 1 ./test/hellompich"])
 
-  runTest("hellompi-n2", 6, [testconfig.MPICH_MPD,
-                             testconfig.MPICH_MPIEXEC+" -n 2 ./test/hellompi"])
+  runTest("hellompich-n2", 6, [testconfig.MPICH_MPD,
+                           testconfig.MPICH_MPIEXEC+" -n 2 ./test/hellompich"])
 
   runTest("mpdboot",     1, [testconfig.MPICH_MPDBOOT+" -n 1"])
 
   #os.system(testconfig.MPICH_MPDCLEANUP)
 
+# Temporarily disabling OpenMPI test as it fails on some distros (OpenSUSE 11.4)
+if testconfig.HAS_OPENMPI == "yes":
+  numProcesses = 5 + int(testconfig.USES_OPENMPI_ORTED == "yes")
+  # FIXME: Replace "[5,6]" by numProcesses when bug in configure is fixed.
+  # /usr/bin/openmpi does not work if /usr/bin is not also in user's PATH
+  oldPath = ""
+  if not os.environ.has_key('PATH'):
+    oldPath = None
+    os.environ['PATH'] = os.path.dirname(testconfig.OPENMPI_MPIRUN)
+  elif (not re.search(os.path.dirname(testconfig.OPENMPI_MPIRUN),
+                     os.environ['PATH'])):
+    oldPath = os.environ['PATH']
+    os.environ += ":" + os.path.dirname(testconfig.OPENMPI_MPIRUN)
+  runTest("openmpi", [5,6], [testconfig.OPENMPI_MPIRUN + " -np 4" +
+			     " ./test/openmpi"])
+  if oldPath:
+    os.environ['PATH'] = oldPath
+  if oldPath == None:
+    del os.environ['PATH']
+
 print "== Summary =="
 print "%s: %d of %d tests passed" % (socket.gethostname(), stats[0], stats[1])
 
+saveResultsNMI()
+
 try:
   SHUTDOWN()
 except CheckFailed, e:
   print "Error in SHUTDOWN():", e.value
 except:
   print "Error in SHUTDOWN()"
+
+sys.exit( stats[1] - stats[0] )  # Return code is number of failing tests.
